/*
 *   $RCSfile: sm_Object.c,v $  
 *   $Revision: 1.17 $  
 *   $Date: 1993/10/18 20:56:31 $      
 */ 
/**********************************************************************
* EXODUS Database Toolkit Software
* Copyright (c) 1991 Computer Sciences Department, University of
*                    Wisconsin -- Madison
* All Rights Reserved.
*
* Permission to use, copy, modify and distribute this software and its
* documentation is hereby granted, provided that both the copyright
* notice and this permission notice appear in all copies of the
* software, derivative works or modified versions, and any portions
* thereof, and that both notices appear in supporting documentation.
*
* THE COMPUTER SCIENCES DEPARTMENT OF THE UNIVERSITY OF WISCONSIN --
* MADISON ALLOWS FREE USE OF THIS SOFTWARE IN ITS "AS IS" CONDITION.  
* THE DEPARTMENT DISCLAIMS ANY LIABILITY OF ANY KIND FOR ANY DAMAGES
* WHATSOEVER RESULTING FROM THE USE OF THIS SOFTWARE.
*
* The EXODUS Project Group requests users of this software to return 
* any improvements or extensions that they make to:
*
*   EXODUS Project Group 
*     c/o David J. DeWitt and Michael J. Carey
*   Computer Sciences Department
*   University of Wisconsin -- Madison
*   Madison, WI 53706
*
*	 or exodus@cs.wisc.edu
*
* In addition, the EXODUS Project Group requests that users grant the 
* Computer Sciences Department rights to redistribute these changes.
**********************************************************************/


#include "ess.h"
#include "sysdefs.h"
#include "checking.h"
#include "io.h"
#include "page.h"
#include "list.h"
#include "pool.h"
#include "object.h"
#include "tid.h"
#include "bf.h"
#include "chunk.h"
#include "lgobject.h"
#include "error.h"
#include "trace.h"
#include "lsn.h"
#include "clog.h"
#include "trans.h"
#include "logrecs.h"
#include "loginfo.h"
#include "logaction.h"
#include "file.h"
#include "scan.h"
#include "load.h"
#include "sm_params.h"
#include "btree.h"
#include "lock.h"
#include "version_graph.h"
#include "sm_state.h"

#include "bf_extfuncs.h"
#include "clog_extfuncs.h"
#include "fi_extfuncs.h"
#include "io_extfuncs.h"
#include "lg_extfuncs.h"
#include "lg_versionfuncs.h"
#include "lm_extfuncs.h"
#include "sm_extfuncs.h"
#include "sm_intfuncs.h"
#include "sm_version_intfuncs.h"
#include "trans_funcs.h"
#include "util_funcs.h"

#include "bf_globals.h"
#include "clog_globals.h"
#include "sm_globals.h"
#include "trans_globals.h"

#include "bf_macro.h"
#include "fi_macro.h"
#include "sm_macro.h"

#if MAGIC_CHECKING IS_ENABLED
#define CHECK_USERDESC(_userDesc)					\
	if ( ((_userDesc)->magic != USERDESC_MAGIC) ||	\
    	!(userDesc->flags & USERDESC_VALID))	{	\
		SM_ERROR(TYPE_USER, esmBADUSERDESC);		\
        return esmFAILURE;							\
	}												
#else 
#define CHECK_USERDESC(_userDesc)					\
	 if( !(userDesc->flags & USERDESC_VALID) )	{	\
		SM_ERROR(TYPE_USER, esmBADUSERDESC);		\
        return esmFAILURE;							\
    }
#endif 


 int
SM_LargeObjOnPage (
	SLOTTEDPAGE		*homePage 
)
{
	TRACE(TR_SM, TR_LEVEL_1);
	/*
	 *	If the object is the first large object on the page,
	 *	mark the page in the file
	 */
	if (homePage->header.largeCount == 0) {
		if (fi_MarkPage( &(homePage->header.fid), 
						 homePage->header.pid.page) )	{
			return esmFAILURE;
		}
	}

	/*
	 *	Increment the count of the number of large objects on page
	 */
	homePage->header.largeCount++;
	return esmNOERROR;
}

int
SM_AppendForward (
	BUFGROUP			*bufGroup,
	GROUPLINK			*homeLink,
	OID					*homeOid,
	int					length,
	char				*data 
)
{
	PID					homePid;
	PID					forwardPid;
	SLOTTEDPAGE			*homePage;
	SLOTTEDPAGE			*forwardPage;
	PAGESLOT			*homeSlotPtr;
	PAGESLOT			*forwardSlotPtr;
	SLOTINDEX			homeSlot;
	SLOTINDEX			forwardSlot;
	int					newAlignLength;
	int					origAlignLength;
	OID					forwardOid;
	OID					newOid;
	SMALLOBJ			*homeObject;
	SMALLOBJ			*forwardObject;
	GROUPLINK			*forwardLink = NULL;
	int					oldSize,newSize;
	int					objectSize;
	OPERATIONINFO       forwardOpInfo,opInfo;



	TRPRINT(TR_SM, TR_LEVEL_1, ("length:%d", length));
    /*
     *	Construct the pid from the oid
     */
    homePid.volid = homeOid->diskAddr.volid;
    homePid.page = homeOid->diskAddr.page;

	homePage = (SLOTTEDPAGE *) homeLink->bufFrame;
	homeSlot = homeOid->diskAddr.slot;
	homeSlotPtr = GETSLOTPTR(homePage, homeSlot);
    homeObject = (SMALLOBJ *) GETOBJECTPTR(homePage, homeSlotPtr);
	forwardOid = *((OID *) homeObject->data);

    forwardPid.volid = forwardOid.diskAddr.volid;
    forwardPid.page = forwardOid.diskAddr.page;

	if ((forwardLink = bf_ReadPage(bufGroup, &forwardPid,
			H_SLOTTED, SLOTTED_PAGE2SIZE, EX)) == NULL)	{
		FAIL
	}

	forwardPage = (SLOTTEDPAGE *) forwardLink->bufFrame;
	if (CHECK_SLOTTED_MAGIC(forwardPage))	{
		SM_ERROR(TYPE_FATAL, esmBADSLOTTEDMAGIC);
	}

	forwardSlot = forwardOid.diskAddr.slot;
	forwardSlotPtr = GETSLOTPTR(forwardPage, forwardSlot);
    forwardObject = (SMALLOBJ *) GETOBJECTPTR(forwardPage, forwardSlotPtr);

	if (CHECKOID(forwardSlot, forwardPage,
				 forwardSlotPtr, forwardOid.diskAddr.unique))	{
		SM_ERROR(TYPE_FATAL, esmBADOID);
	}

	newSize = (int)homeObject->objectHdr.size + length;
	TRPRINT(TR_SM, TR_LEVEL_2, ("newSize:%d", newSize));

	origAlignLength = ALIGNSIZE((int)homeObject->objectHdr.size);
	TRPRINT(TR_SM, TR_LEVEL_2, ("origAlignLength:%d", origAlignLength));

	/*
	 *	round the size up to the nearest alignment value
	 */
	newAlignLength = ALIGNSIZE(newSize);
	TRPRINT(TR_SM, TR_LEVEL_2, ("newAlignLength:%d", newAlignLength));

	sm_SetAbortIfError(); /* henceforth... */

	/*
	 *	check to see if the object can fit on a small page
	 */
	if (newSize > LARGETHRESHOLD)	{
		if (SM_ConvertAppend(bufGroup, homeOid, &homePid,
							 homeLink, (int)homeObject->objectHdr.size, length,
					  		 forwardObject->data, data))	{
			FAIL
		}

		if (GET_LOGLEVEL(LOG_SPACE)) {

			/* Log the destroy of the currently forwarded object as a
					regular destroy */
			forwardOpInfo.flags = 0;
			forwardOpInfo.oid = forwardOid;
			forwardOpInfo.start = 0;
			forwardOpInfo.size = (int)forwardObject->objectHdr.size;
			forwardOpInfo.oldData = (char *) forwardObject;
			forwardOpInfo.newData = NULL;
			forwardOpInfo.actionPid.volid = forwardOid.diskAddr.volid;
			forwardOpInfo.actionPid.page  = forwardOid.diskAddr.page;
			if (GET_LOGLEVEL(LOG_DATA)) {
				forwardOpInfo.action = LOG_ACTION_DESTROY_SMALL_OBJ;
				/* Log the destroy operation */
				if(clog_LogDestroyObj(&forwardOpInfo)) {
					FAIL
				}
			} else {
				forwardOpInfo.action = LOG_ACTION_DESTROY_SMALL_OBJ_ND;
				/* Log the destroy operation */
				if(clog_LogDestroyObjND(&forwardOpInfo)) {
					FAIL
				}
			}
		}

		/*
		 *	give back the space for the forwarded object
		 */
		SM_ReturnSlot(forwardLink, forwardSlot, origAlignLength);
		forwardLink = NULL; /* unfixed by SM_ReturnSlot */
		forwardPage = NULL; /* Audited by SM_ReturnSlot */
		goto Success;
	}

	/*
	 * Dealing with a small object.
	 */

	if (GET_LOGLEVEL(LOG_SPACE)) {
		/*
		 * First set up fields for logging,
		 * then make sure there is enough free space for the new data.
		 */
		/* set up operation info to log insert to forwarded part */
		forwardOpInfo.action = LOG_ACTION_APPEND_SMALL_OBJ;
		forwardOpInfo.flags =0 ;
		forwardOpInfo.oid = forwardOid;
		forwardOpInfo.start = (int)forwardObject->objectHdr.size;
		forwardOpInfo.size = length;
		forwardOpInfo.oldData = NULL;
		forwardOpInfo.newData = data;
		forwardOpInfo.actionPid.volid = forwardOid.diskAddr.volid;
		forwardOpInfo.actionPid.page  = forwardOid.diskAddr.page;
		forwardOpInfo.currentLRC = forwardPage->header.lrc;
	}

    switch (SM_CheckFree(forwardPage, forwardLink, forwardSlot, forwardObject, length))	{

		case PAGE_NO_SPACE: {

			TRPRINT(TR_SM, TR_LEVEL_2, ("PAGE_NO_SPACE"));

			objectSize = (int)forwardObject->objectHdr.size;
			if (SM_CreateAppend(bufGroup, &(homePage->header.fid), homeOid,
					length + objectSize, forwardLink, forwardObject->data,
					data, objectSize, &newOid, (OBJHDR *) forwardObject))	{
				FAIL;
			}

			if (GET_LOGLEVEL(LOG_SPACE)) {
				/* Log the change to the forwarding OID in home object*/
				opInfo.action = LOG_ACTION_WRITE_SMALL_OBJ;
				opInfo.flags = 0;
				opInfo.start = 0;
				opInfo.size = sizeof(OID);
				opInfo.oldData = (char *) (homeObject->data);
				opInfo.newData = (char *) &newOid;
				opInfo.oid = *homeOid;
				opInfo.currentLRC = homePage->header.lrc;
				opInfo.actionPid.volid = homeOid->diskAddr.volid;
				opInfo.actionPid.page  = homeOid->diskAddr.page;
				if(clog_LogWriteObj(&opInfo)) {
					FAIL
				}
			}

			/*
			 *	get a pointer to the oid field and fill it in
			 */
			(*((OID *) homeObject->data)) = newOid;

			/*
			 *	set the sizes
			 */
            oldSize = (int)homeObject->objectHdr.size;
			homeObject->objectHdr.size = newSize;

			if (GET_LOGLEVEL(LOG_SPACE)) {
				/* Log the update to the original header */
				if(clog_LogChangeSmSize(homeOid,&(homeObject->objectHdr),
												oldSize,newSize)) {
					FAIL
				}
			}

			/* Log the destroy of the currently forwarded object as a
				regular destroy */
			if (GET_LOGLEVEL(LOG_SPACE)) {

				/* fill in the log info - using some values filled in above*/
				forwardOpInfo.flags = 0;
				forwardOpInfo.start = 0;
				forwardOpInfo.size = objectSize;
				forwardOpInfo.newData = NULL;
				forwardOpInfo.oldData = (char *) forwardObject;

				if (GET_LOGLEVEL(LOG_DATA)) {
					forwardOpInfo.action = LOG_ACTION_DESTROY_SMALL_OBJ;
					/* Log the destroy operation */
					if(clog_LogDestroyObj(&forwardOpInfo)) {
						FAIL
					}
				} else {
					forwardOpInfo.action = LOG_ACTION_DESTROY_SMALL_OBJ_ND;
					/* Log the destroy operation */
					if(clog_LogDestroyObjND(&forwardOpInfo)) {
						FAIL
					}
				}
			}

			/*
			 *	give back old forwarded space
			 */
			SM_ReturnSlot(forwardLink, forwardSlot, origAlignLength);
			forwardLink = NULL; /* unfixed by SM_ReturnSlot */
			forwardPage = NULL; /* audited by SM_ReturnSlot */
			goto Success;
			} /* end    case PAGE_NO_SPACE: */

		case PAGE_LAST_SLOT: {

			TRPRINT(TR_SM, TR_LEVEL_2, ("PAGE_LAST_SLOT"));

			/*
			 *	Reestablish pointer as it may have moved
			 */
			forwardObject =
				 (SMALLOBJ *) GETOBJECTPTR(forwardPage, forwardSlotPtr);

			/*
			 *	we can just copy in the data
			 */
			if (data != NULL) {
		 	   bcopy(data,(forwardObject->data + forwardObject->objectHdr.size), (int)length);
			} else {
		 	   bzero((forwardObject->data + forwardObject->objectHdr.size), (int)length);
			}

			/*
			 *	adjust the free start and the free count
			 */
			forwardPage->header.freeBytes -= (newAlignLength - origAlignLength);
			forwardPage->header.freeStart += (newAlignLength - origAlignLength);

			/* 
			 * Update the object's header, and adjust the 
			 * user pointers into the object.
			 */
			forwardObject->objectHdr.size = newSize;
            oldSize = (int)homeObject->objectHdr.size;
			homeObject->objectHdr.size = newSize;

			if (GET_LOGLEVEL(LOG_SPACE)) {
				/* Log the update to the original header */
				if(clog_LogChangeSmSize(homeOid,&(homeObject->objectHdr),
									oldSize,newSize)) {
					FAIL
				}

				if (GET_LOGLEVEL(LOG_DATA) && (data != NULL)) {
					/* Log the append operation */
					if(clog_LogModifyObj(&forwardOpInfo)) {
						FAIL
					}
				} else {
					forwardOpInfo.action = LOG_ACTION_APPEND_SMALL_OBJ_ND;
					forwardOpInfo.newData = NULL;
					if(clog_LogModifyObjND(&forwardOpInfo)) {
						FAIL
					}
				}
			}
			break;
			} /* end  case PAGE_LAST_SLOT */

		
		case PAGE_INT_SLOT: {

			TRPRINT(TR_SM, TR_LEVEL_2, ("PAGE_INT_SLOT"));

			/*
			 *	Reestablish pointer as it may have moved
			 */
			forwardObject = (SMALLOBJ *) GETOBJECTPTR(forwardPage, forwardSlotPtr);

			/*
			 *	copy the exisiting data to the end of the page
			 */
			bcopy((char *) forwardObject,
				  ((char *) forwardPage) + forwardPage->header.freeStart,
				(int)(sizeof(OBJHDR) + forwardObject->objectHdr.size));


			/*
			 *	adjust the offset in the slot
			 *	Get a pointer to the location of the new object,
			 */
			forwardSlotPtr->offset = forwardPage->header.freeStart;
			forwardObject = (SMALLOBJ *) GETOBJECTPTR(forwardPage, forwardSlotPtr);

			/*
			 *	tack the data onto the end
			 */
			if (data != NULL) {
			    bcopy(data, &(forwardObject->data[forwardObject->objectHdr.size]), (int) length);
			} else {
			    bzero(&(forwardObject->data[forwardObject->objectHdr.size]), (int) length);
			}
			/*
			 *	Update the original and forward  objects' headers
			 */
			forwardObject->objectHdr.size = newSize;
            oldSize = (int)homeObject->objectHdr.size;
			homeObject->objectHdr.size = newSize;

			/*
			 *	adjust the free space to account for the extra data
			 */
			forwardPage->header.freeBytes -= (newAlignLength - origAlignLength);
			forwardPage->header.freeStart += sizeof(OBJHDR) + newAlignLength;
			TRPRINT(TR_SM, TR_LEVEL_2, ("freeBytes:%d freeStart:%d",
				forwardPage->header.freeBytes, forwardPage->header.freeStart));

			if (GET_LOGLEVEL(LOG_SPACE)) {
				/* Log the update to the original header */
				if(clog_LogChangeSmSize(homeOid,&(homeObject->objectHdr),
									oldSize,newSize)) {
					FAIL
				}

				if (GET_LOGLEVEL(LOG_DATA) && (data != NULL)) {
					/* Log the append operation */
					if(clog_LogModifyObj(&forwardOpInfo)) {
						FAIL
					}
				} else {
					forwardOpInfo.action = LOG_ACTION_APPEND_SMALL_OBJ_ND;
					forwardOpInfo.newData = NULL;
					if(clog_LogModifyObjND(&forwardOpInfo)) {
						FAIL
					}
				}
			}
			break;
			} /* end   case PAGE_INT_SLOT */

		
		default:

			/*
			 *	this cannot happen
			 */
			SM_ERROR(TYPE_FATAL, esmINTERNAL);
			break;
    }

Success:
	if(forwardPage) {
		sm_AuditSlotPage(forwardPage);
	}
	if(forwardLink) {
		bf_UnfixPage(forwardLink, BF_DEFAULT, TRUE);
	}
	if(homePage) {
		sm_AuditSlotPage(homePage);
	}
    return esmNOERROR;

FailureReturn:
	if(forwardLink) {
		bf_UnfixPage(forwardLink, BF_DEFAULT, TRUE); /* dirty from the start */
	}
	return esmFAILURE;
}


 int
SM_AppendToLarge (
	FID				*fid,
	BUFGROUP		*bufGroup,
	OID				*homeOid,
	GROUPLINK		*homeLink,
	LARGEOBJHDR		*largeObject,
	int				length,
	char			*data,
	SLOTINDEX		rootSlot 
)
{
	PID				rootPid;
	PID				homePid;
	int				oldSize;
	OPERATIONINFO   opInfo;


	TRPRINT(TR_SM, TR_LEVEL_1, ("length:%d", length));

	homePid.page = homeOid->diskAddr.page;
	homePid.volid = homeOid->diskAddr.volid;

	/* lg_ functions do sm_SetAbortIfError() so we don't have to */

	/*
	 *	Check to see if this is a slotted root
	 */
	if (largeObject->objectHdr.properties & P_HEADERONPAGE)	{
		LARGEOBJROOT		*largeRoot;

		TRPRINT(TR_SM, TR_LEVEL_2, ("slotted root"));
		/*
		 *	Do the actual large append
		 */
		if (lg_AppendToObject(fid, &homePid, bufGroup, length, data, rootSlot, (largeObject->objectHdr.properties & P_VERSIONED) ))	{

			return esmFAILURE;
		}

		largeRoot = (LARGEOBJROOT *) largeObject;

		/*
		 *	check to see if the object has moved from its
		 *	slot on the page
		 */
		if (largeRoot->rootPage.header.numSlots == NIL)	{

			SLOTTEDPAGE		*pagePtr;

			TRPRINT(TR_SM, TR_LEVEL_2, ("slotted root removed"));

            /* copy the part of the slotted root for logging purposes */
            bcopy((char *) largeRoot,OldData, sizeof(LARGEOBJROOT));

			/*
			 *	the header can no longer fit on the page
			 */
			largeObject->objectHdr.properties &= ~P_HEADERONPAGE;

			/*
			 *	store the fid of the root
			 */
			largeObject->rootPid = homePid.page;

            /* Log the conversion from a Slotted root to a large obj header*/
            /* note, this call will update the size field of the header */

			if (GET_LOGLEVEL(LOG_SPACE)) {
				opInfo.action = LOG_ACTION_CONVERT_TO_LARGEHDR;
				opInfo.flags = 0;
				opInfo.oid = *homeOid;
				opInfo.start = 0;
				opInfo.size = (int) sizeof(LARGEOBJROOT);
				opInfo.oldData = OldData;
				opInfo.newData = (char *) largeObject;
				opInfo.actionPid.volid = homeOid->diskAddr.volid;
				opInfo.actionPid.page  = homeOid->diskAddr.page;
				if(clog_LogConvertToLarge(&opInfo) ) {
					FAIL
				}
			}

			pagePtr = (SLOTTEDPAGE *) homeLink->bufFrame;
			pagePtr->header.freeBytes += (sizeof(LGSLOTROOT)-ALIGNSIZE(sizeof(SHORTPID)));
		}

	} else {

		/*
		 * Dealing with a large object.
		 * Dig out the root pid, then use large object append.
		 */
		rootPid.volid = homePid.volid;
		rootPid.page = largeObject->rootPid;

		/*
		 *	Do the actual large append
		 */
		if (lg_AppendToObject(fid, &rootPid, bufGroup, length, data, NIL, (largeObject->objectHdr.properties & P_VERSIONED)))	{

			return esmFAILURE;
		}

		/* Update the root page if it has changed */
		if (largeObject->rootPid != rootPid.page) {
			if (GET_LOGLEVEL(LOG_SPACE)) {
				/* Log the change to the root PID in the home object*/
				opInfo.action = LOG_ACTION_WRITE_SMALL_OBJ;
				opInfo.flags = 0;
				opInfo.oid = *homeOid;
				opInfo.start = 0;
				opInfo.size = sizeof(SHORTPID);
				opInfo.oldData = (char *) &(largeObject->rootPid);
				opInfo.newData = (char *) &(rootPid.page);
				opInfo.actionPid.volid = homeOid->diskAddr.volid;
				opInfo.actionPid.page  = homeOid->diskAddr.page;
				if(clog_LogWriteObj(&opInfo)) {
					FAIL
				}
			}
			largeObject->rootPid = rootPid.page;
		}
	}

	oldSize = (int)largeObject->objectHdr.size;

	/*
	 *	Update the size and the root page
	 */
	largeObject->objectHdr.size += length;

	if (GET_LOGLEVEL(LOG_SPACE)) {
		/* log the size change to the root object*/
		if(clog_LogChangeSmSize(homeOid,&(largeObject->objectHdr),
					oldSize, (int)largeObject->objectHdr.size)) {
			FAIL
		}
	}
Success:
	return esmNOERROR;

FailureReturn:
	return esmFAILURE;
}




 int
SM_ConvertAppend (

	BUFGROUP		*bufGroup,
	OID				*oid,
	PID				*homePid,
	GROUPLINK		*homeLink,
	int				oldLength,
	int				newLength,
	char			*oldData,
	char			*newData 
)
{
	SLOTTEDPAGE		*homePage;
	PAGESLOT		*slotPtr;
	SMALLOBJ		*smallObject;
	LARGEOBJHDR		*largeObject;
	PID				rootPid;
	int				currentSize,dataSize;
	SLOTINDEX		objectSlot;
	int				newSize;
	int				spaceNeeded;
	BOOL  			conversionDone;
	OPERATIONINFO  	opInfo;
	/* temporary space for storing old data for logging */
    char			oldDataSpace[sizeof(SLOTTEDPAGE)];
	int				convertSize;
	
	TRACE(TR_SM, TR_LEVEL_1);

	/*
	 *	Get a pointer to the data page
	 */
	homePage = (SLOTTEDPAGE *) homeLink->bufFrame;
	sm_AuditSlotPage(homePage);

	/*
	 *	get the slot position and a pointer to the slot
	 */
	objectSlot = oid->diskAddr.slot;
	slotPtr = GETSLOTPTR(homePage, objectSlot);

    /*
     *	Get a pointer to the existing object
     */
    smallObject = (SMALLOBJ *) GETOBJECTPTR(homePage, slotPtr);

	/*
	 *	Calculate its total final size
	 */
    newSize = oldLength + newLength;
	TRPRINT(TR_SM, TR_LEVEL_2, ("newSize:%d", newSize));

	/*
	 *  No conversion has been done yet
	 */
	conversionDone = FALSE;

    /*
     *  calculate the current size of the object on the home page
     */
    /* note the logging of the delete of the forwarded part of the object
	   must be handled by the invoker */
    if (smallObject->objectHdr.properties & P_MOVED)    {
        dataSize = sizeof(OID);
    } else {
        dataSize = (int)smallObject->objectHdr.size;
    }
    currentSize = sizeof(OBJHDR) + ALIGNSIZE(dataSize);

	/*
	 *	Check to see if the object will fit in a one level tree
	 */
	if (newSize <= (LSLOTTHRESHOLD))	{

		TRPRINT(TR_SM, TR_LEVEL_2, ("creating slotted large"));

		/*
		 *  calculate space needed based on the oid
		 */
		spaceNeeded = (ALIGNSIZE(sizeof(LARGEOBJROOT)) - currentSize);
		if (spaceNeeded < 0) {
			spaceNeeded = 0;
		}

		/*
		 * Dealing with a small object.
		 * First make sure there is enough free space for the new data.
		 * Note: SM_CheckFree may compact the page and cause objects to move.
		 */
		switch (SM_CheckFree(homePage, homeLink, objectSlot, smallObject, spaceNeeded))	{

			case PAGE_NO_SPACE:	{

				TRPRINT(TR_SM, TR_LEVEL_2, ("PAGE_NO_SPACE"));
				break;
			}

			
			case PAGE_LAST_SLOT:	{

				TRPRINT(TR_SM, TR_LEVEL_2, ("PAGE_LAST_SLOT"));

				/* Reestablish pointer as it may have moved */
				smallObject = (SMALLOBJ *) GETOBJECTPTR(homePage, slotPtr);

				/*
				 *	for logging purposes, we need to log a delete of the
				 *  data portion of the original object if it is not
				 *	a moved object.
				 */
				if (!(smallObject->objectHdr.properties & P_MOVED))	{

				    /*decrement the size of the small object */
				    smallObject->objectHdr.size -= dataSize;
					sm_SetAbortIfError();

					/* make sure we are deleting the correct amount*/
					SM_ASSERT(LEVEL_3,smallObject->objectHdr.size == 0);

                    /*log the delete of the data portion of the orig sm object*/
					if (GET_LOGLEVEL(LOG_SPACE)) {
						opInfo.flags = 0;
						opInfo.oid = *oid;
						opInfo.start = 0;
						opInfo.size = dataSize;
						opInfo.newData = NULL;
						opInfo.actionPid.volid = oid->diskAddr.volid;
						opInfo.actionPid.page  = oid->diskAddr.page;

						if (GET_LOGLEVEL(LOG_DATA)) {
							opInfo.action = LOG_ACTION_DELETE_SMALL_OBJ;
							opInfo.oldData = (char *) (smallObject->data);
							if(clog_LogModifyObj(&opInfo)) {
								FAIL
							}
						} else {
							opInfo.action = LOG_ACTION_DELETE_SMALL_OBJ_ND;
							opInfo.oldData = NULL;
							if(clog_LogModifyObjND(&opInfo)) {
								FAIL
							}
						}
					}

					/* copy the entire small object to a temporary buffer */
					bcopy((char*) smallObject, oldDataSpace, 
			 			(int) dataSize + sizeof(OBJHDR));

					/*  set the olddata pointer to point to the copy*/
					oldData = ((SMALLOBJ*) oldDataSpace)->data;
			
					convertSize = sizeof(OBJHDR);
				} else {

					/*
					 *	copy the anchor part of forwarded sm object to
					 *	a temporary buffer
					 */
					bcopy((char*) smallObject, oldDataSpace,
						  (int) sizeof(OID) + sizeof(OBJHDR));

					sm_SetAbortIfError();
					convertSize = sizeof(OID) + sizeof(OBJHDR);
				}	

				/*
				 *	initialize the fields of the object header
				 */
				if (smallObject->objectHdr.properties & P_HISTORYGRAPH)
					smallObject->objectHdr.properties =
									P_LARGEOBJ|P_HEADERONPAGE|P_HISTORYGRAPH;
				else
					smallObject->objectHdr.properties = 
													P_LARGEOBJ|P_HEADERONPAGE;


		        /* mark the object as create in progress so lg code won't log
				 any changes to the slotted root before its create is logged */
        		smallObject->objectHdr.properties |= P_CREATEINPROGRESS;

				/*
				 *	create the object with the original data
				 */
				if (lg_CreateObject( &(homePage->header.fid), homePid,
					 bufGroup, oldLength, oldData, homePid, objectSlot))   {
					 FAIL
				}

        		/* unset create in progess flag for logging*/
        		smallObject->objectHdr.properties &= ~P_CREATEINPROGRESS;

                /* Log the conversion from a small object to a large root */ 
				/* for non forwarded objects, this assumes that the entire
					data portion of the orig obj has been logged as deleted */
				/* note: logical operation - will need to adjust freespace on
					page. */
				if (GET_LOGLEVEL(LOG_SPACE)) {
					opInfo.action = LOG_ACTION_CONVERT_TO_LARGEROOT;
					opInfo.flags = 0;
					opInfo.oid = *oid;
					opInfo.start = 0;
					opInfo.size = convertSize; 
					opInfo.oldData = oldDataSpace;
					opInfo.newData = (char *) smallObject;
					opInfo.actionPid.volid = oid->diskAddr.volid;
					opInfo.actionPid.page  = oid->diskAddr.page;
					if(clog_LogConvertToLarge(&opInfo)) {
						FAIL
					}
				}

				/*
				 *	Append the data to the new object
				 */
				if (lg_AppendToObject( &(homePage->header.fid), homePid,
							bufGroup, newLength, newData, objectSlot, FALSE)) {

					return esmFAILURE;
				}

				/*
				 *	set the object to the new length
				 */
				if (GET_LOGLEVEL(LOG_SPACE)) {
					if(clog_LogChangeSmSize(oid,&(smallObject->objectHdr),
									 (int)smallObject->objectHdr.size,newSize)){
						 FAIL
					 }
				}
				smallObject->objectHdr.size = newSize;

				/*
				 *	we must give back some space
				 */
				homePage->header.freeBytes += 
							(currentSize - ALIGNSIZE(sizeof(LARGEOBJROOT)));
				homePage->header.freeStart -=
							(currentSize - ALIGNSIZE(sizeof(LARGEOBJROOT)));
				TRPRINT(TR_SM, TR_LEVEL_2, ("new freeBytes:%d",
						homePage->header.freeBytes));

				conversionDone = TRUE;

				break;
			}

			
			case PAGE_INT_SLOT:	{


				TRPRINT(TR_SM, TR_LEVEL_2, ("PAGE_INT_SLOT"));

				smallObject = (SMALLOBJ *) GETOBJECTPTR(homePage, slotPtr);
				slotPtr->offset = homePage->header.freeStart;
				sm_SetAbortIfError();

				homePage->header.freeStart += ALIGNSIZE(sizeof(LARGEOBJROOT));

				if (!(smallObject->objectHdr.properties & P_MOVED))	{
                    /*log the delete of the data portion of the orig sm object*/
					if (GET_LOGLEVEL(LOG_SPACE)) {
						opInfo.action = LOG_ACTION_DELETE_SMALL_OBJ;
						opInfo.flags = 0;
						opInfo.oid = *oid;
						opInfo.start = 0;
						opInfo.size = dataSize;
						opInfo.newData = NULL;
						opInfo.actionPid.volid = oid->diskAddr.volid;
						opInfo.actionPid.page  = oid->diskAddr.page;
						if (GET_LOGLEVEL(LOG_DATA)) {
							opInfo.action = LOG_ACTION_DELETE_SMALL_OBJ;
							opInfo.oldData = (char *) (smallObject->data);
							if(clog_LogModifyObj(&opInfo)) {
								FAIL
							}
						} else {
							opInfo.action = LOG_ACTION_DELETE_SMALL_OBJ_ND;
							opInfo.oldData = NULL;
							if(clog_LogModifyObjND(&opInfo)) {
								FAIL
							}
						}
					}

				    /*decrement the size of the small object */
				    smallObject->objectHdr.size -= dataSize;

					/* make sure we are deleting the correct amount*/
					SM_ASSERT(LEVEL_3,smallObject->objectHdr.size == 0);

					/*
					 *	copy the small object header to a temporary buffer
					 */
					bcopy((char*) smallObject, oldDataSpace, sizeof(OBJHDR));
			
					convertSize = sizeof(OBJHDR);
				} else {

					/*
					 *	copy the anchor part of forwarded sm object to
					 *	a temporary buffer
					 */
					bcopy((char*) smallObject, oldDataSpace,
						  (int) sizeof(OID) + sizeof(OBJHDR));
					sm_SetAbortIfError();

					convertSize = sizeof(OID) + sizeof(OBJHDR);
				}	

				/*
				 *  Get a pointer to the location of the new object,
				 *  and copy the small old header for the small object.
				 *  No adjustment to the user pointers into the object
				 *  is necessary since all user descriptors will be
				 *  updated at the end of sm_InsertInObject().
				 */
				smallObject = (SMALLOBJ *) GETOBJECTPTR(homePage, slotPtr);
				smallObject->objectHdr = *(OBJHDR *) oldDataSpace;

		        /* mark the object as create in progress so lg code won't log
				 any changes to the slotted root before its create is logged */
        		smallObject->objectHdr.properties |= P_CREATEINPROGRESS;

				/*
				 *	create a new object at the end of the page
				 */
				if (lg_CreateObject( &(homePage->header.fid), homePid, bufGroup,
					    oldLength, oldData, homePid, objectSlot))	{
					FAIL
				}

				/*
				 *  The new object is a large object with a root on
				 *  the slotted page.  It is not moved or forwarded.
				 */
				smallObject->objectHdr.properties |= P_LARGEOBJ|P_HEADERONPAGE;
				smallObject->objectHdr.properties &= ~(P_MOVED|P_FORWARD);

        		/* unset create in progess flag for logging*/
        		smallObject->objectHdr.properties &= ~P_CREATEINPROGRESS;

                /* Log the conversion from a small object to a large root */ 
				/* note: logical operation - will need to adjust freespace on
					page. */
				if (GET_LOGLEVEL(LOG_SPACE)) {
					opInfo.action = LOG_ACTION_CONVERT_TO_LARGEROOT;
					opInfo.flags = 0;
					opInfo.oid = *oid;
					opInfo.start = 0;
					opInfo.size = convertSize;
					opInfo.oldData = oldDataSpace;
					opInfo.newData = (char *) smallObject;
					opInfo.actionPid.volid = oid->diskAddr.volid;
					opInfo.actionPid.page  = oid->diskAddr.page;
					if(clog_LogConvertToLarge(&opInfo)) {
						FAIL
					}
				}

				/* Append the data to the new object */
				if (lg_AppendToObject(&(homePage->header.fid), homePid,
									bufGroup, newLength, newData, objectSlot, FALSE))	{
					FAIL
				}

				/*
				 *	adjust the free space to account for the extra data
				 */
				homePage->header.freeBytes +=
						(currentSize - ALIGNSIZE(sizeof(LARGEOBJROOT)));
				TRPRINT(TR_SM, TR_LEVEL_2, ("freeBytes:%d freeStart:%d",
						homePage->header.freeBytes, homePage->header.freeStart));

				/*
				 *  set the object to the new length
				 */
				if(clog_LogChangeSmSize(oid,&(smallObject->objectHdr),
							(int)smallObject->objectHdr.size,newSize)) {
					FAIL
				}
				smallObject->objectHdr.size = newSize;

				conversionDone = TRUE;

				break;
			}

			default:
				/* this cannot happen */
				SM_ERROR(TYPE_WARNING, esmINTERNAL);
		}
	}

	if (!conversionDone) {

		/* if we get here, this means we will need to create a large
			object with a separate root node.  Since we know the header
			portion will fit in the spot where the current small object is,
			there is no reason to look for additional space in the page.*/

		TRPRINT(TR_SM, TR_LEVEL_2, ("creating large object on seperate"));

		if (!(smallObject->objectHdr.properties & P_MOVED))	{

		    if (dataSize > sizeof(SHORTPID)) { 

                /*
				 * log the delete of the data portion of the original sm
				 * object beyond the space required to store the pid
				 */
				if (GET_LOGLEVEL(LOG_SPACE)) {
					opInfo.flags = 0;
					opInfo.oid = *oid;
					opInfo.start = sizeof(SHORTPID);
					opInfo.size = dataSize - sizeof(SHORTPID);
					opInfo.newData = NULL;
					opInfo.actionPid.volid = oid->diskAddr.volid;
					opInfo.actionPid.page  = oid->diskAddr.page;
					if (GET_LOGLEVEL(LOG_DATA)) {
						opInfo.action = LOG_ACTION_DELETE_SMALL_OBJ;
						opInfo.oldData = (char *) (smallObject->data) +sizeof(SHORTPID);
						if(clog_LogModifyObj(&opInfo)) {
							FAIL
						}
					} else {
						opInfo.action = LOG_ACTION_DELETE_SMALL_OBJ_ND;
						opInfo.oldData = NULL;
						if(clog_LogModifyObjND(&opInfo)){
							FAIL
						}
					}
				}
		    }
	
		    /*decrement the size of the small object */
		    smallObject->objectHdr.size -= (dataSize - sizeof(SHORTPID));

		    /* copy portion of the small object that will be
											overwritten; for logging */
		    bcopy((char *)  smallObject, oldDataSpace, sizeof(LARGEOBJHDR));
			convertSize = sizeof(LARGEOBJHDR);

		} else {
			/*
			 *	copy the anchor part of forwarded sm object to
			 *	a temporary buffer
			 */
			bcopy((char*) smallObject, oldDataSpace,
						  (int) sizeof(OID) + sizeof(OBJHDR));

			convertSize = sizeof(OID) + sizeof(OBJHDR);
		}
		sm_SetAbortIfError();

		/*
		 *	Create a large object by reading the data out of the
		 *	old object on the page
		 */
		if (lg_CreateObject( &(homePage->header.fid), homePid,
				 bufGroup, oldLength, oldData, &rootPid, NIL))	{
			FAIL
		}

		/*
		 *	Append the data to the large object.
		 */
		if (lg_AppendToObject( &(homePage->header.fid), &rootPid, bufGroup,
							   newLength, newData, NIL, FALSE))	{
			FAIL
		}

		/*
		 *	Convert the small object on the home page to a large object.
		 *	Then unfix the page and return.  We don't recopy fields
		 *	that correspond in place
		 */
		largeObject = (LARGEOBJHDR *) smallObject;
		if (largeObject->objectHdr.properties & P_HISTORYGRAPH) {
			largeObject->objectHdr.properties = P_LARGEOBJ|P_HISTORYGRAPH;
		} else {
			largeObject->objectHdr.properties = P_LARGEOBJ;
		}
		largeObject->objectHdr.size = newSize;
		largeObject->rootPid = rootPid.page;

        /* Log the conversion from a small object to large obj header*/ 
		if (GET_LOGLEVEL(LOG_SPACE)) {
			opInfo.action = LOG_ACTION_CONVERT_TO_LARGEHDR;
			opInfo.flags = 0;
			opInfo.oid = *oid;
			opInfo.start = 0;
			opInfo.size = convertSize;
			opInfo.oldData = oldDataSpace;
			opInfo.newData = (char *) largeObject;
			opInfo.actionPid.volid = oid->diskAddr.volid;
			opInfo.actionPid.page  = oid->diskAddr.page;
			if(clog_LogConvertToLarge(&opInfo)){
				FAIL
			}
		}

		/*
		 *	give back the number of free bytes
		 *	must fix to account for contiguous bytes
		 */
		homePage->header.freeBytes +=
							(currentSize - ALIGNSIZE(sizeof(LARGEOBJHDR)));

		/* The small object is now a large one */
		conversionDone = TRUE;
	}

	/*
	 *  Record the fact that a large object has been created on the
	 *  page.
	 */
	if (SM_LargeObjOnPage(homePage)) {
		FAIL
	}

Success:
	return esmNOERROR;

FailureReturn:
	return esmFAILURE;
}




 int
SM_ConvertInsert (

	BUFGROUP		*bufGroup,
	OID				*oid,
	PID				*homePid,
	GROUPLINK		*homeLink,
	int				start,
	int				oldLength,
	int				newLength,
	char			*oldData,
	char			*newData 
)
{

	SLOTTEDPAGE		*homePage;
	PAGESLOT		*slotPtr;
	SMALLOBJ		*smallObject;
	LARGEOBJHDR		*largeObject;
	PID				rootPid;
	int				currentSize,dataSize,newSize;
	SLOTINDEX		objectSlot;
	int				spaceNeeded;
	BOOL 			conversionDone;
	OPERATIONINFO	opInfo;
    /* temporary space for storing old data for logging */
    char            oldDataSpace[sizeof(SLOTTEDPAGE)];
    int             convertSize;

	TRACE(TR_SM, TR_LEVEL_1);

	/*
	 *	Get a pointer to the data page
	 */
	homePage = (SLOTTEDPAGE *) homeLink->bufFrame;

	sm_AuditSlotPage(homePage);

	objectSlot = oid->diskAddr.slot;
	slotPtr = GETSLOTPTR(homePage, objectSlot);
    smallObject = (SMALLOBJ *) GETOBJECTPTR(homePage, slotPtr);

	/*
	 *	Calculate its total final size
	 */
    newSize = oldLength + newLength;
	TRPRINT(TR_SM, TR_LEVEL_2, ("newSize:%d", newSize));

	conversionDone = FALSE;

    /*
     *  calculate the current size of the object on the home page
     */
    /* note the logging of the delete of the forwarded part of the object
       must be handled by the invoker */
    if (smallObject->objectHdr.properties & P_MOVED)    {
        dataSize = sizeof(OID);
    } else {
        dataSize = (int)smallObject->objectHdr.size;
    }
    currentSize = sizeof(OBJHDR) + ALIGNSIZE(dataSize);

	/*
	 *	Check to see if the object will fit in a one level tree
	 */
	if (newSize <= (LSLOTTHRESHOLD))	{

		TRPRINT(TR_SM, TR_LEVEL_2, ("creating slotted large"));

        spaceNeeded = (ALIGNSIZE(sizeof(LARGEOBJROOT)) - currentSize);
        if (spaceNeeded < 0) {
            spaceNeeded = 0;
        }

		/*
		 * Dealing with a small object.
		 * First make sure there is enough free space for the new data.
		 */
		switch (SM_CheckFree(homePage, homeLink, objectSlot, smallObject, spaceNeeded))	{

			case PAGE_NO_SPACE:	{

				TRPRINT(TR_SM, TR_LEVEL_2, ("PAGE_NO_SPACE"));
				break;
			}
			
			case PAGE_LAST_SLOT:	{


				TRPRINT(TR_SM, TR_LEVEL_2, ("PAGE_LAST_SLOT"));

				/*
				 *	Reestablish pointer as it may have moved
				 */
				smallObject = (SMALLOBJ *) GETOBJECTPTR(homePage, slotPtr);

				/*
                 *  for logging purposes, we need to log a delete of the
                 *  data portion of the original object if it is not
                 *  a moved object.
				 */
				if (!(smallObject->objectHdr.properties & P_MOVED))	{

                    /*decrement the size of the small object */
                    smallObject->objectHdr.size -= dataSize;
					sm_SetAbortIfError();

                    /* make sure we are deleting the correct amount*/
                    SM_ASSERT(LEVEL_3,smallObject->objectHdr.size == 0);

                    /*log the delete of the data portion of the orig sm object*/
					if (GET_LOGLEVEL(LOG_SPACE)) {
						opInfo.flags = 0;
						opInfo.oid = *oid;
						opInfo.start = 0;
						opInfo.size = dataSize;
						opInfo.newData = NULL;
						opInfo.actionPid.volid = oid->diskAddr.volid;
						opInfo.actionPid.page  = oid->diskAddr.page;
						if (GET_LOGLEVEL(LOG_DATA)) {
							opInfo.action = LOG_ACTION_DELETE_SMALL_OBJ;
							opInfo.oldData = (char *) (smallObject->data);
							if(clog_LogModifyObj(&opInfo)) {
								FAIL
							}
						} else {
							opInfo.action = LOG_ACTION_DELETE_SMALL_OBJ_ND;
							opInfo.oldData = NULL;
							if(clog_LogModifyObjND(&opInfo)) {
								FAIL
							}
						}
					}

                    /*
                     *  copy the entire small object to a temporary buffer
                     */
                    bcopy((char*) smallObject, oldDataSpace,
			 								(int) dataSize + sizeof(OBJHDR));

                    /*  set the olddata pointer to point to the copy*/
                    oldData = ((SMALLOBJ*) oldDataSpace)->data;

                    convertSize = sizeof(OBJHDR);
                } else {

                    /*
                     *  copy the anchor part of forwarded sm object to
                     *  a temporary buffer
                     */
                    bcopy((char*) smallObject, oldDataSpace,
                          (int) sizeof(OID) + sizeof(OBJHDR));
					sm_SetAbortIfError();

                    convertSize = sizeof(OID) + sizeof(OBJHDR);
                }

				/*
				 *	initialize the fields of the object header
				 */
				if (smallObject->objectHdr.properties & P_HISTORYGRAPH) 
					smallObject->objectHdr.properties = 
							P_LARGEOBJ|P_HEADERONPAGE|P_HISTORYGRAPH;
				else
				   smallObject->objectHdr.properties=P_LARGEOBJ|P_HEADERONPAGE;

		        /* mark the object as create in progress so lg code won't log
				 any changes to the slotted root before its create is logged */
        		smallObject->objectHdr.properties |= P_CREATEINPROGRESS;

				/*
				 *	create the object with the original data
				 */
				if (lg_CreateObject(&(homePage->header.fid), homePid, bufGroup,
									 start, oldData, homePid, objectSlot))	{
					FAIL
				}

                /* unset create in progess flag for logging*/
                smallObject->objectHdr.properties &= ~P_CREATEINPROGRESS;

                /*
				 * Log the conversion from a small object to a large
				 * root for non forwarded objects, this assumes that
				 * the entire data portion of the orig obj has been
				 * logged as deleted 
				 *
                 * note: logical operation - will need to adjust
				 * freespace on page.
				 */
				if (GET_LOGLEVEL(LOG_SPACE)) {
					opInfo.action = LOG_ACTION_CONVERT_TO_LARGEROOT;
					opInfo.flags = 0;
					opInfo.oid = *oid;
					opInfo.start = 0;
					opInfo.size = convertSize;
					opInfo.oldData = oldDataSpace;
					opInfo.newData = (char *) smallObject;
					opInfo.actionPid.volid = oid->diskAddr.volid;
					opInfo.actionPid.page  = oid->diskAddr.page;
					if(clog_LogConvertToLarge(&opInfo)) {
						FAIL
					}
				}

				/* Append the data to the new object */
				if (lg_AppendToObject( &(homePage->header.fid), homePid,
						  bufGroup, newLength, newData, objectSlot, FALSE))	{
					FAIL
				}
	
				if ( (oldLength - start) > 0 ) {
					/*
					 *	Append remaining portion of the original data
					 */
					if (lg_AppendToObject( &(homePage->header.fid), homePid,
						 bufGroup, oldLength - start, (oldData + start),
															 objectSlot, FALSE)) {
						FAIL
					}
				}

                /*
                 *  set the object to the new length
                 */
				if (GET_LOGLEVEL(LOG_SPACE)) {
					if(clog_LogChangeSmSize(oid,&(smallObject->objectHdr),
								(int)smallObject->objectHdr.size,newSize)) {
						FAIL
					}
				}
				smallObject->objectHdr.size = newSize;

				/*
				 *	Adjust the space on the page
				 */
				homePage->header.freeBytes +=
					 (currentSize - ALIGNSIZE(sizeof(LARGEOBJROOT)));
				homePage->header.freeStart -=
					 (currentSize - ALIGNSIZE(sizeof(LARGEOBJROOT)));
				TRPRINT(TR_SM, TR_LEVEL_2, ("new freeBytes:%d",
						homePage->header.freeBytes));

				/* The conversion has been performed	*/
				conversionDone = TRUE;
				break;
			}

			
			case PAGE_INT_SLOT:	{
				TRPRINT(TR_SM, TR_LEVEL_2, ("PAGE_INT_SLOT"));
				smallObject = (SMALLOBJ *) GETOBJECTPTR(homePage, slotPtr);
				slotPtr->offset = homePage->header.freeStart;
				homePage->header.freeStart += ALIGNSIZE(sizeof(LARGEOBJROOT));
				sm_SetAbortIfError();

                if (!(smallObject->objectHdr.properties & P_MOVED)) {
                    /*log the delete of the data portion of the orig sm object*/
					if (GET_LOGLEVEL(LOG_SPACE)) {
						opInfo.flags = 0;
						opInfo.oid = *oid;
						opInfo.start = 0;
						opInfo.size = dataSize;
						opInfo.newData = NULL;
						opInfo.actionPid.volid = oid->diskAddr.volid;
						opInfo.actionPid.page  = oid->diskAddr.page;
						if (GET_LOGLEVEL(LOG_DATA)) {
							opInfo.action = LOG_ACTION_DELETE_SMALL_OBJ;
							opInfo.oldData = (char *) (smallObject->data);
							if(clog_LogModifyObj(&opInfo)) {
								FAIL
							}
						} else {
							opInfo.action = LOG_ACTION_DELETE_SMALL_OBJ_ND;
							opInfo.oldData = NULL;
							if(clog_LogModifyObjND(&opInfo)) {
								FAIL
							}
						}
					}

                    /*decrement the size of the small object */
                    smallObject->objectHdr.size -= dataSize;

                    /* make sure we are deleting the correct amount*/
                    SM_ASSERT(LEVEL_3,smallObject->objectHdr.size == 0);

                    /* copy the small object header to temp buffer */
                    bcopy((char*) smallObject, oldDataSpace, sizeof(OBJHDR));

                    convertSize = sizeof(OBJHDR);
                } else {

                    /*
                     *  copy the anchor part of forwarded sm object to
                     *  a temporary buffer
                     */
                    bcopy((char*) smallObject, oldDataSpace,
                          (int) sizeof(OID) + sizeof(OBJHDR));

                    convertSize = sizeof(OID) + sizeof(OBJHDR);
                }

				/* 
				 *	Get a pointer to the location of the new object,
				 *	and copy the small old header for the small object.
				 *	No adjustment to the user pointers into the object
				 *	is necessary since all user descriptors will be
				 *	updated at the end of sm_InsertInObject().
				 */
				smallObject = (SMALLOBJ *) GETOBJECTPTR(homePage, slotPtr);
                smallObject->objectHdr = *(OBJHDR *) oldDataSpace;

		        /* mark the object as create in progress so lg code won't log
				 any changes to the slotted root before its create is logged */
        		smallObject->objectHdr.properties |= P_CREATEINPROGRESS;

				/*
				 *	create a new object at the end of the page
				 */
				if (lg_CreateObject(&(homePage->header.fid), homePid, bufGroup,
									 start, oldData, homePid, objectSlot))	{
					return esmFAILURE;
				}

				/*
				 *	The new object is a large object with a root on
				 *	the slotted page.  It is not moved or forwarded.
				 */
				smallObject->objectHdr.properties |= P_LARGEOBJ|P_HEADERONPAGE;
				smallObject->objectHdr.properties &= ~(P_MOVED|P_FORWARD);

                /* unset create in progess flag for logging*/
                smallObject->objectHdr.properties &= ~P_CREATEINPROGRESS;

                /*
				 * Log the conversion from a small object to large root
                 * note: logical operation - will need to adjust
				 * freespace on page.
				 */
				if (GET_LOGLEVEL(LOG_SPACE)) {
					opInfo.action = LOG_ACTION_CONVERT_TO_LARGEROOT;
					opInfo.flags = 0;
					opInfo.oid = *oid;
					opInfo.start = 0;
					opInfo.size = convertSize;
					opInfo.oldData = oldDataSpace;
					opInfo.newData = (char *) smallObject;
					opInfo.actionPid.volid = oid->diskAddr.volid;
					opInfo.actionPid.page  = oid->diskAddr.page;
					if(clog_LogConvertToLarge(&opInfo)) {
						FAIL
					}
				}

				/* Append the data to the new object */
				if (lg_AppendToObject(&(homePage->header.fid),homePid,bufGroup,
								   newLength, newData, objectSlot, FALSE))	{
					FAIL
				}

				/*
				 *	Append the data to the new object
				 */
				if (lg_AppendToObject(&(homePage->header.fid), homePid,bufGroup,
				     oldLength - start, oldData + start, objectSlot, FALSE)) {
					 FAIL
				}

				/*
				 *	Add the data back to the free space
				 */
				homePage->header.freeBytes +=
							 (currentSize - ALIGNSIZE(sizeof(LARGEOBJROOT)));
				TRPRINT(TR_SM, TR_LEVEL_2, ("freeBytes:%d freeStart:%d",
				 	  homePage->header.freeBytes, homePage->header.freeStart));

				/*
				 *	set the object to the new length
				 */
				if (GET_LOGLEVEL(LOG_SPACE)) {
					if(clog_LogChangeSmSize(oid,&(smallObject->objectHdr),
								 (int)smallObject->objectHdr.size,newSize)) {
						FAIL
					}
				}
				smallObject->objectHdr.size = newSize;

				/* The conversion has been performed	*/
				conversionDone = TRUE;

				break;
			}

			default:
				SM_ERROR(TYPE_FATAL, esmINTERNAL);
		}
	}
	if (!conversionDone) {

        /*
		 * if we get here, this means we will need to create a large
         * object with a separate root node.  Since we know the header
         * portion will fit in the spot where the current small object
		 * is, there is no reason to look for additional space in
		 * the page.
		 */

		/*
		 *	allocate enough space to hold the header only, the root of
		 *	the large object will be on a separate page
		 */
		TRPRINT(TR_SM, TR_LEVEL_2, ("creating large object on seperate"));

        if (!(smallObject->objectHdr.properties & P_MOVED)) {

            if (dataSize > sizeof(SHORTPID)) {

                /*
				 * log the delete of the data portion of the original
				 * sm object beyond the space required to store the pid
				 */
				if (GET_LOGLEVEL(LOG_SPACE)) {
					opInfo.flags = 0;
					opInfo.oid = *oid;
					opInfo.start = sizeof(SHORTPID);
					opInfo.size = dataSize - sizeof(SHORTPID);
					opInfo.newData = NULL;
					opInfo.actionPid.volid = oid->diskAddr.volid;
					opInfo.actionPid.page  = oid->diskAddr.page;
					if (GET_LOGLEVEL(LOG_DATA)) {
						opInfo.action = LOG_ACTION_DELETE_SMALL_OBJ;
						opInfo.oldData = (char *) (smallObject->data) +sizeof(SHORTPID);
						if(clog_LogModifyObj(&opInfo)) {
							FAIL
						}
					} else {
						opInfo.action = LOG_ACTION_DELETE_SMALL_OBJ_ND;
						opInfo.oldData = NULL;
						if(clog_LogModifyObjND(&opInfo)) {
							FAIL
						}
					}
				}
		    }

		    /*decrement the size of the small object */
            smallObject->objectHdr.size -= (dataSize - sizeof(SHORTPID));

            /* copy portion of the small object that will be
                                            overwritten; for logging */
            bcopy((char *)  smallObject, oldDataSpace, sizeof(LARGEOBJHDR));
            convertSize = sizeof(LARGEOBJHDR);

        } else {
            /*
             *  copy the anchor part of forwarded sm object to
             *  a temporary buffer
             */
            bcopy((char*) smallObject, oldDataSpace,
                         (int) sizeof(OID) + sizeof(OBJHDR));

            convertSize = sizeof(OID) + sizeof(OBJHDR);
        }
		sm_SetAbortIfError();

		/*
		 *	Create a large object by reading the data out of the
		 *	old object on the page
		 */
		if (lg_CreateObject( &(homePage->header.fid) , homePid,
							 bufGroup, start, oldData, &rootPid, NIL))	{
			FAIL
		}

		/*
		 *	Append the data to be inserted in the large object.
		 */
		if (lg_AppendToObject( &(homePage->header.fid), &rootPid,
							   bufGroup, newLength, newData, NIL, FALSE ))	{
			FAIL
		}

		/*
		 *	append the data after the insertion point
		 */
		if (lg_AppendToObject( &(homePage->header.fid), &rootPid, bufGroup,
							   oldLength - start, oldData + start, NIL, FALSE))	{
			FAIL
		}

		/*
		 *	Convert the small object on the home page to a large object.
		 *	Then unfix the page and return.  We don't recopy fields
		 *	that correspond in place
		 */
		largeObject = (LARGEOBJHDR *) smallObject;
		if (largeObject->objectHdr.properties & P_HISTORYGRAPH) 
			largeObject->objectHdr.properties = P_LARGEOBJ | P_HISTORYGRAPH;
		else
			largeObject->objectHdr.properties = P_LARGEOBJ ;
		largeObject->objectHdr.size = newSize;
		largeObject->rootPid = rootPid.page;

        /* Log the conversion from a small object to a large obj header*/
		if (GET_LOGLEVEL(LOG_SPACE)) {
			opInfo.action = LOG_ACTION_CONVERT_TO_LARGEHDR;
			opInfo.flags = 0;
			opInfo.oid = *oid;
			opInfo.start = 0;
			opInfo.size = convertSize;
			opInfo.oldData = oldDataSpace;
			opInfo.newData = (char *) largeObject;
			opInfo.actionPid.volid = oid->diskAddr.volid;
			opInfo.actionPid.page  = oid->diskAddr.page;
			if(clog_LogConvertToLarge(&opInfo)) {
				FAIL
			}
		}

		/*
		 *	give back the number of free bytes
		 *	must fix to account for contiguous bytes
		 */
		homePage->header.freeBytes += 
							(currentSize - ALIGNSIZE(sizeof(LARGEOBJHDR)));

		/* The small object is now a large one */
		conversionDone = TRUE;
	}

	/*
	 *	Record the fact that a large object has been created on the
	 *	page.
	 */
	if (SM_LargeObjOnPage(homePage)) {
		return esmFAILURE;
	}

Success:
	return esmNOERROR;

FailureReturn:
	return esmFAILURE;
}



int
SM_CreateAppend (
	BUFGROUP				*bufGroup,
	FID						*fid,
	OID						*nearObject,
	int						length,
	GROUPLINK				*origLink, /* does not get unfixed by this func */
	char					*oldData,
	char					*newData,
	int						oldDataSize,
	OID						*newOid,
	OBJHDR					*objectHdr 
)
{
    register PAGESLOT		*slotPtr;		/* slot pointer					*/
    register SMALLOBJ		*smallObject;	/* small object pointer			*/
    register SLOTTEDPAGE	*homePage;		/* ptr to home page				*/
    SLOTINDEX				emptySlot;		/* slot used for the new obj	*/
    int						spaceNeeded;	/* space needed for the new obj */
	int						alignLength;
    BOOL					allocatePage;	/* whether to allocate new page */
    BOOL					allocateSlot;/* whether we need to allocate slot */
    PID						nearPid;		/* create the obj near this page */
    SLOTTEDPAGE				*nearPage;		/* indirect ptr to near page	*/
    PID						homePid;		/* the new obj's home page		*/
	GROUPLINK				*nearLink = NULL;
	GROUPLINK				*homeLink = NULL;
	UNIQUE					unique;
    OPERATIONINFO           opInfo;  /* operation info used for logging */


	TRPRINT(TR_SM,TR_LEVEL_1,("length:%d oldDataSize:%d", length, oldDataSize));

	/*
	 * Put the new object near the one given by nearObject.
	 */
	nearPid.volid = nearObject->diskAddr.volid;
	nearPid.page = nearObject->diskAddr.page;

	/*
	 *	get the alignLength
	 */
	alignLength = ALIGNSIZE(length);
	TRPRINT(TR_SM, TR_LEVEL_2, ("alignLength:%d", alignLength));

	/*
	 *	calculate the spaceNeeded
	 */
	spaceNeeded = sizeof(OBJHDR) + alignLength + sizeof(PAGESLOT);

	/*
	 *	call the routine to find a neighbor page that can be used
	 *  SM_TryNeigbor does not unfix origLink, but might fix
	 *  a new page as the resulting nearLink.
	 */
	if (SM_TryNeighbor(bufGroup, spaceNeeded, origLink,
				   	   &allocatePage, &nearPid, &nearLink))	{
		FAIL
	}
	/* Now we have two cases:
	 * allocatePage == TRUE:
	 *     nearLink == origLink, fixed once.
	 * allocatePage == FALSE:
	 * 		origLink != nearLink, both are fixed once.
	 */

    /*
     *	Allocate a page if necessary.
     *	If setNeighborLink is TRUE, then nearPage is valid
     *	and points to the near page whose neighbor pid link has to be set.
     */
    if (allocatePage == TRUE) {

		/* nearLink == origLink */
		nearPage = (SLOTTEDPAGE *) nearLink->bufFrame;

		/*
		 * Do the low-level allocation.
		 */
		if (io_AllocSlottedPages(1, &nearPid, &homePid, fid, FALSE, &unique, bufGroup))	{
			FAIL
		}
		TRPRINT(TR_SM, TR_LEVEL_2, ("allocated new page:%d", homePid.page));

		sm_SetAbortIfError();

		if ((homeLink = bf_FakeReadPage(bufGroup, &homePid, H_SLOTTED,
								   SLOTTED_PAGE2SIZE)) == NULL)	{
			FAIL
		}

		homePage = (SLOTTEDPAGE *) homeLink->bufFrame;

		INITPAGE(homePage, homePid, *fid, FALSE, unique);
		homePage->header.lrc = *(GET_LOGINFO->initLRC);

		nearPage->header.neighborPid = homePid;
		TRPRINT(TR_SM, TR_LEVEL_2, ("setting neighbor link"));

    } else {
		/* nearLink != origLink */
		homeLink = nearLink; /* a new page */
		homePid = nearPid;
		homePage = (SLOTTEDPAGE *) homeLink->bufFrame;
	}
	/* homeLink always points to a new page, and gets unfixed.
	 * origLink does NOT get unfixed - caller takes care of that.
	 * Don't unfix nearLink because it's equal to homeLink or
	 * origLink.
	 */

	/*
	 *	check to see if there are any stray free slots
	 */
	if (homePage->header.freeSlot == NIL)	{
		TRPRINT(TR_SM, TR_LEVEL_2, ("slot is new"));

		/*
		 *	Just grab the next slot after current last
		 */
		emptySlot = homePage->header.slotCount;
		TRPRINT(TR_SM, TR_LEVEL_2, ("using free slot:%d", emptySlot));
		slotPtr = GETSLOTPTR(homePage, emptySlot);
		homePage->header.slotCount++;

		sm_SetAbortIfError();

		allocateSlot = TRUE;

	} else {

		emptySlot = homePage->header.freeSlot;
		TRPRINT(TR_SM, TR_LEVEL_2, ("using free slot:%d", emptySlot));

		slotPtr = GETSLOTPTR(homePage, emptySlot);

		/*
		 *	get a new head of the list
		 */
		homePage->header.freeSlot = (SLOTINDEX) slotPtr->unique;
		TRPRINT(TR_SM, TR_LEVEL_2, ("new free slot:%d", homePage->header.freeSlot));
		allocateSlot = FALSE;
	}

    /*
     * If we got here, we're creating a small object.
     * Copy the object to the home page.
     */
    smallObject = (SMALLOBJ *) (((char *) homePage) + homePage->header.freeStart);
    smallObject->objectHdr.properties = P_FORWARD;
    smallObject->objectHdr.tag	      = objectHdr->tag;
    smallObject->objectHdr.size	      = length;
	TRPRINT(TR_SM, TR_LEVEL_2, ("freestart:%d length:%d",
			homePage->header.freeStart, length));


#if OBJECT_MAGIC IS_ENABLED
	smallObject->objectHdr.magic      = OBJECT_HDR_MAGIC;
#endif OBJECT_MAGIC IS_ENABLED

	/*
	 *	copy in the old data unconditionally
	 */
   	bcopy(oldData, smallObject->data, (int) oldDataSize);

    /*
     *	Only copy the new data in if the pointer is valid
     */
    if (newData != NULL)	{
    	bcopy(newData, &(smallObject->data[oldDataSize]), (int) (length - oldDataSize));
    } else {
    	bzero(&(smallObject->data[oldDataSize]), (int) (length - oldDataSize));
	}

	/*
	 *	Initialize the contents of the slot
	 */
    slotPtr->offset = homePage->header.freeStart;
	SM_ASSERT(LEVEL_1, slotPtr->offset % ALIGNON == 0);

	/*
	 *	check to see if we can use the unique on the page
	 */
	if (homePage->header.uniqueCount == 0)	{

		/*
		 *	get a new unique range
		 */
		if (io_GetUnique(&homePid, &unique, UNIQUE_ALLOC_COUNT))	{
			FAIL;
		}
		homePage->header.unique = unique;
		homePage->header.uniqueCount = UNIQUE_ALLOC_COUNT;
	}

	slotPtr->unique = homePage->header.unique++;
	homePage->header.uniqueCount--;
	TRPRINT(TR_SM, TR_LEVEL_2, ("unique:%x", slotPtr->unique));

	/*
     *	Update the home page stats to account for the
	 *	space allocated for the object
     */
    homePage->header.freeStart += sizeof(OBJHDR) + alignLength;
    homePage->header.freeBytes -= sizeof(OBJHDR) + alignLength;

	/*
	 *	Update free space if we allocated a new slot
	 */
    if (allocateSlot == TRUE) {
		homePage->header.freeBytes -= sizeof(PAGESLOT);
    }

    /*
     * Fill in the oid for return to the user
     */
    newOid->diskAddr.page	= homePid.page;
    newOid->diskAddr.slot	= (SLOTINDEX) emptySlot;
	newOid->diskAddr.unique = slotPtr->unique;
	newOid->diskAddr.volid  = homePid.volid;
	TRPRINT(TR_SM, TR_LEVEL_2, ("oid page:%d slot:%d unique:%d volid:%d",
			newOid->diskAddr.page, newOid->diskAddr.slot,
			newOid->diskAddr.unique, newOid->diskAddr.volid));


    /* for logging purposes, we log this as a create of a new object.
        The update to the original (moved) object is logged separately */

    /* Fill in the operation info */
	if (GET_LOGLEVEL(LOG_SPACE)) {

		opInfo.flags = 0;
		opInfo.oid = *newOid;
		opInfo.start = 0;
		opInfo.size = (int) smallObject->objectHdr.size;
		opInfo.oldData = NULL;
		opInfo.actionPid.volid = newOid->diskAddr.volid;
		opInfo.actionPid.page  = newOid->diskAddr.page;
		opInfo.currentLRC = homePage->header.lrc;
		opInfo.newData = (char *) smallObject;

		if (GET_LOGLEVEL(LOG_DATA)) {
			opInfo.action = LOG_ACTION_CREATE_SMALL_OBJ;
			/* Log the create operation */
			if(clog_LogCreateObj(&opInfo)) {
				FAIL
			}
		} else {
			opInfo.action = LOG_ACTION_CREATE_SMALL_OBJ_ND;
			/* Log the create operation */
			if(clog_LogCreateObjND(&opInfo)) {
				FAIL
			}
		}
	}

Success:
	if(homePage) {
		sm_AuditSlotPage(homePage);
	}
	if(homeLink) {
		bf_UnfixPage(homeLink, BF_DEFAULT, TRUE); /* dirty from the start */
	}
    return esmNOERROR;

FailureReturn:
	if(homeLink) {
		bf_UnfixPage(homeLink, BF_DEFAULT, TRUE); /* dirty from the start */
	}
	return esmFAILURE;
}

 int
SM_CreateInsert (
	BUFGROUP				*bufGroup,
	FID						*fid,
	OID						*nearObject,
	int						length,
	GROUPLINK				*origLink, /* does not free */
	int						start,
	char					*oldData,
	char					*newData,
	int						oldDataSize,
	OID						*newOid,
	OBJHDR					*objectHdr 
)
{
    PAGESLOT			*slotPtr;		/* slot pointer					*/
    SMALLOBJ			*smallObject;	/* small object pointer			*/
    SLOTTEDPAGE			*homePage;		/* ptr to home page				*/
    int					spaceNeeded;	/* space needed for the new obj */
    SLOTINDEX			emptySlot;		/* slot used for the new obj	*/
	int					alignLength;
    BOOL				allocatePage;	/* whether to allocate new page */
    BOOL				allocateSlot;	/* need to allocate slot		*/
    PID					nearPid;		/* create the obj near this page*/
    SLOTTEDPAGE			*nearPage;		/* indirect ptr to near page	*/
    PID					homePid;		/* the new obj's home page		*/
	GROUPLINK			*nearLink;
	GROUPLINK			*homeLink = NULL;
	int					newDataSize;
	UNIQUE				unique;
	OPERATIONINFO       opInfo;  /* operation info used for logging */

	TRPRINT(TR_SM,TR_LEVEL_1,("length:%d oldDataSize:%d", length, oldDataSize));

	/*
	 * Put the new object near the one given by nearObject.
	 */
	nearPid.volid = nearObject->diskAddr.volid;
	nearPid.page = nearObject->diskAddr.page;

	/*
	 *	calculate the alignlength
	 */
	alignLength = ALIGNSIZE(length);
	TRPRINT(TR_SM, TR_LEVEL_2, ("alignLength:%d", alignLength));

	/*
	 *	calculate the spaceNeeded
	 */
	spaceNeeded = sizeof(OBJHDR) + alignLength + sizeof(PAGESLOT);

	/*
	 *	call the routine to find a neighbor page that can be used
	 */
	if (SM_TryNeighbor(bufGroup, spaceNeeded, origLink,
					   &allocatePage, &nearPid, &nearLink))	{
		SM_ASSERT(LEVEL_1, (nearLink == NULL));
		FAIL
	}

	/*
	 *	check to see if we have a valid page
	 */
	if (!allocatePage)	{
		SM_ASSERT(LEVEL_1, (nearLink != origLink));
		/*
		 *	setup the home fields
		 */
		homePid = nearPid;
		homeLink = nearLink; /* is NOT origLink, is a new page */
		homePage = (SLOTTEDPAGE *) homeLink->bufFrame;

	} else {

		SM_ASSERT(LEVEL_1, (nearLink == origLink));
		/* just set nearPage for future use */
		nearPage = (SLOTTEDPAGE *) nearLink->bufFrame;

		if (io_AllocSlottedPages(1, &nearPid, &homePid, fid, FALSE,
														&unique, bufGroup))	{
			FAIL
		}
		TRPRINT(TR_SM, TR_LEVEL_2, ("allocated new page:%d", homePid.page));

		if ((homeLink = bf_FakeReadPage(bufGroup, &homePid, H_SLOTTED,
								   SLOTTED_PAGE2SIZE)) == NULL)	{
			FAIL
		}
		homePage = (SLOTTEDPAGE *) homeLink->bufFrame;

		INITPAGE(homePage, homePid, *fid, FALSE, unique);
		homePage->header.lrc = *(GET_LOGINFO->initLRC);

		nearPage->header.neighborPid = homePid;
		TRPRINT(TR_SM, TR_LEVEL_2, ("setting neighbor link"));
    }

	/*
	 *	check to see if there are any stray free slots
	 */
	if (homePage->header.freeSlot == NIL)	{
		TRPRINT(TR_SM, TR_LEVEL_2, ("slot is new"));

		/*	Just grab the next slot after current last */
		emptySlot = homePage->header.slotCount;
		TRPRINT(TR_SM, TR_LEVEL_2, ("using free slot:%d", emptySlot));

		slotPtr = GETSLOTPTR(homePage, emptySlot);
		homePage->header.slotCount++;
		allocateSlot = TRUE;
	} else {
		emptySlot = homePage->header.freeSlot;
		TRPRINT(TR_SM, TR_LEVEL_2, ("using free slot:%d", emptySlot));

		slotPtr = GETSLOTPTR(homePage, emptySlot);
		homePage->header.freeSlot = (SLOTINDEX) slotPtr->unique;
		TRPRINT(TR_SM, TR_LEVEL_2, ("new free slot:%d", homePage->header.freeSlot));
		allocateSlot = FALSE;
	}
	sm_SetAbortIfError();

    /*
     *	If we got here, we're creating a small object.
     *	Copy the object to the home page.
     */
    smallObject = (SMALLOBJ *) (((char *) homePage)+homePage->header.freeStart);
    smallObject->objectHdr.properties = P_FORWARD;
    smallObject->objectHdr.tag		  = objectHdr->tag;
    smallObject->objectHdr.size		  = length;
	TRPRINT(TR_SM, TR_LEVEL_2, ("freestart:%d length:%d",
			homePage->header.freeStart, length));


#if OBJECT_MAGIC IS_ENABLED
	smallObject->objectHdr.magic      = OBJECT_HDR_MAGIC;
#endif OBJECT_MAGIC IS_ENABLED

	/*
	 *	calculate the new data size from the total and old
	 */
	newDataSize = length - oldDataSize;

	/*
	 *	copy in the old data up to the insertion point
	 */
   	bcopy(oldData, smallObject->data, (int) start);

    /*
     *	Only copy the new data in if the pointer is valid
     */
    if (newData != NULL)	{
    	bcopy(newData, &(smallObject->data[start]), (int) newDataSize);
    } else {
    	bzero(&(smallObject->data[start]), (int) newDataSize);
	}

	/*
	 *	copy in the old data after the insertion point
	 */
   	bcopy( &(oldData[start]), &(smallObject->data[start + newDataSize]),
		   (int) (oldDataSize - start));

	/*
	 *	Initialize the contents of the slot
	 */
    slotPtr->offset = homePage->header.freeStart;
	SM_ASSERT(LEVEL_1, slotPtr->offset % ALIGNON == 0);

	/*
	 *	check to see if we can use the unique on the page
	 */
	if (homePage->header.uniqueCount == 0)	{

		/*
		 *	get a new unique range
		 */
		if (io_GetUnique(&homePid, &unique, UNIQUE_ALLOC_COUNT))	{
			FAIL
		}

		homePage->header.unique = unique;
		homePage->header.uniqueCount = UNIQUE_ALLOC_COUNT;
	}

	/*
	 *	fill in the unique field
	 */
	slotPtr->unique = homePage->header.unique++;
	homePage->header.uniqueCount--;
	TRPRINT(TR_SM, TR_LEVEL_2, ("unique:%x", slotPtr->unique));

	/*
     *	Update the home page stats to account for the
	 *	space allocated for the object
     */
    homePage->header.freeStart += sizeof(OBJHDR) + alignLength;
    homePage->header.freeBytes -= sizeof(OBJHDR) + alignLength;

	/*
	 *	Update free space if we allocated a new slot
	 */
    if (allocateSlot == TRUE) {
		homePage->header.freeBytes -= sizeof(PAGESLOT);
    }

    /*
     *	Fill in the oid for return to the user
     */
    newOid->diskAddr.page	= homePid.page;
    newOid->diskAddr.slot	= emptySlot;
	newOid->diskAddr.unique = slotPtr->unique;
	newOid->diskAddr.volid  = homePid.volid;
	TRPRINT(TR_SM, TR_LEVEL_2, ("oid page:%d slot:%d unique:%d volid:%d",
			newOid->diskAddr.page, newOid->diskAddr.slot,
			newOid->diskAddr.unique, newOid->diskAddr.volid));

    /* 
	 * for logging purposes, we log this as a create of a new object.
	 * The update to the original (moved) object is logged separately
	 */

    /* Fill in the operation info */
	if (GET_LOGLEVEL(LOG_SPACE)) {
		opInfo.flags = 0;
		opInfo.oid = *newOid;
		opInfo.start = 0;
		opInfo.size = smallObject->objectHdr.size;
		opInfo.oldData = NULL;
		opInfo.newData = (char *) smallObject;
		opInfo.actionPid.volid = newOid->diskAddr.volid;
		opInfo.actionPid.page  = newOid->diskAddr.page;
		opInfo.currentLRC = homePage->header.lrc;
		
		if (GET_LOGLEVEL(LOG_DATA)) {
			opInfo.action = LOG_ACTION_CREATE_SMALL_OBJ;
			/* Log the create operation */
			if(clog_LogCreateObj(&opInfo)) {
				FAIL
			}
		} else {
			opInfo.action = LOG_ACTION_CREATE_SMALL_OBJ_ND;
			/* Log the create operation */
			if(clog_LogCreateObjND(&opInfo)) {
				FAIL
			}
		}
	}
Success:
	sm_AuditSlotPage(homePage);
    bf_UnfixPage(homeLink, BF_DEFAULT, TRUE);
    return esmNOERROR;

FailureReturn:
	if(homeLink) {
		bf_UnfixPage(homeLink, BF_DEFAULT, FALSE);
	}
	return esmFAILURE;

}



int
SM_DeleteForward (

	BUFGROUP				*bufGroup,
	OID						*forwardOid,
	int						start,
	int						length 
)
{
    SLOTINDEX			objectSlot;		/* the object's slot			*/
    SLOTTEDPAGE			*forwardPage = NULL;/* object's home page			*/
    OBJECT				*object;		/* ptr to generic object		*/
    int					newSize;	/* size of the object after delete	*/
	int					newAlignLength;
	int					origAlignLength;
    int					bytesToMove; 	/* bytes to move for delete	*/
    char				*source;		/* the source for the byte move	*/
    char				*deletePoint;	/* the deletion point			*/
    PID					forwardPid;		/* home page of object			*/
	GROUPLINK			*forwardLink = NULL;
	BOOL				forwardLinkIsDirty = FALSE;
	PAGESLOT			*slotPtr;
    OPERATIONINFO		forwardOpInfo;



	TRPRINT(TR_SM, TR_LEVEL_1, ("start:%d length:%d", start, length));

	/*
	 *	get the pid from the oid
	 */
	forwardPid.page  = forwardOid->diskAddr.page;
	forwardPid.volid = forwardOid->diskAddr.volid;

    /*
     * Read in the page containing the object's header.
     */
    if ((forwardLink = bf_ReadPage(bufGroup, &forwardPid, H_SLOTTED,
								   SLOTTED_PAGE2SIZE, EX)) == NULL)	{
		FAIL
	}

	/*
	 *	Get a pointer to the home page
	 */
	forwardPage = (SLOTTEDPAGE *) forwardLink->bufFrame;

	/*
	 *	check the slotted magic number
	 */
	if (CHECK_SLOTTED_MAGIC(forwardPage))	{
		SM_ERROR(TYPE_FATAL, esmBADSLOTTEDMAGIC);
		FAIL
	}
    objectSlot = forwardOid->diskAddr.slot;
	slotPtr = GETSLOTPTR(forwardPage, objectSlot);

    /*
     * Make sure we have a valid oid.
     */
    if (CHECKOID(objectSlot, forwardPage, slotPtr, forwardOid->diskAddr.unique))	{
		SM_ERROR(TYPE_USER, esmBADOID);
		FAIL
	}

    object = GETOBJECTPTR(forwardPage, slotPtr);

    /*
     * Calculate the data size of the object after the delete.
     */
    newSize = (int)object->objectHdr.size - length;

    /*
     *	Dealing with a small object.
     *	Delete the bytes using bcopy, then update object's header info.
     */
    deletePoint = object->data + start;
	origAlignLength = ALIGNSIZE((int)object->objectHdr.size);
	TRPRINT(TR_SM, TR_LEVEL_2, ("origAlignLength:%d", origAlignLength));
    source		= deletePoint + length;
    bytesToMove = (int)object->objectHdr.size - start - length;

    /* set up operation info to log delete from forwarded part */
	if (GET_LOGLEVEL(LOG_SPACE)) {
		forwardOpInfo.flags =0 ;
		forwardOpInfo.oid = *forwardOid;
		forwardOpInfo.start = start;
		forwardOpInfo.size = length;
		forwardOpInfo.newData = NULL;
		forwardOpInfo.actionPid.volid = forwardOid->diskAddr.volid;
		forwardOpInfo.actionPid.page  = forwardOid->diskAddr.page;
		forwardOpInfo.currentLRC = forwardPage->header.lrc;

		if (GET_LOGLEVEL(LOG_DATA)) {
			forwardOpInfo.action = LOG_ACTION_DELETE_SMALL_OBJ;
			forwardOpInfo.oldData = deletePoint;
			/* Log the delete operation */
			if(clog_LogModifyObj(&forwardOpInfo)) {
				FAIL
			}
		} else {
			forwardOpInfo.action = LOG_ACTION_DELETE_SMALL_OBJ_ND;
			forwardOpInfo.oldData = NULL;
			/* Log the delete operation */
			if(clog_LogModifyObjND(&forwardOpInfo)) {
				FAIL
			}
		}
	}
	forwardLinkIsDirty = TRUE;
	sm_SetAbortIfError();
    /* do the delete */
    bcopy(source, deletePoint, (int) bytesToMove);
    object->objectHdr.size = newSize;

	/*
	 *	calculate the new aligned length
	 */
	newAlignLength = ALIGNSIZE(newSize);
	TRPRINT(TR_SM, TR_LEVEL_2, ("newAlignLength:%d", newAlignLength));

    /*
     * Update the page info
     */
    forwardPage->header.freeBytes += (origAlignLength - newAlignLength);
Success:
	if(forwardLink) {
		bf_UnfixPage(forwardLink, BF_DEFAULT, TRUE);
	}
    return esmNOERROR;
FailureReturn:
	if(forwardLink) {
		bf_UnfixPage(forwardLink, BF_DEFAULT, forwardLinkIsDirty);
	}
    return esmFAILURE;
}



int
SM_DestroyForward (

	BUFGROUP				*bufGroup,
	OID						*oid 
)
{
	SLOTTEDPAGE				*page;
	OBJECT					*object;
	PAGESLOT				*slotPtr;
	PID						pid;
	GROUPLINK				*link = NULL;
	SLOTINDEX				slot;
    OPERATIONINFO		    forwardOpInfo;


	TRPRINT(TR_SM, TR_LEVEL_1, ("slot:%d", oid->diskAddr.slot));

    /*
     *	Form the Pid of the target page
     */
    pid.volid = oid->diskAddr.volid;
    pid.page  = oid->diskAddr.page;

    if ((link = bf_ReadPage(bufGroup, &pid,
							H_SLOTTED, SLOTTED_PAGE2SIZE, EX)) == NULL){
		FAIL
	}
	page = (SLOTTEDPAGE *) link->bufFrame;

	if (CHECK_SLOTTED_MAGIC(page))	{
		SM_ERROR(TYPE_FATAL, esmBADSLOTTEDMAGIC);
	}

	slot = oid->diskAddr.slot;
	slotPtr = GETSLOTPTR(page, slot);
	object = GETOBJECTPTR(page, slotPtr);

    if (CHECKOID(slot, page, slotPtr, oid->diskAddr.unique))	{
		SM_ERROR(TYPE_USER, esmBADOID);
		FAIL
	}

    /* Log the delete of the currently forwarded object as a
                regular delete */

    /* fill in the log info*/
	if (GET_LOGLEVEL(LOG_SPACE)) {
		forwardOpInfo.flags = 0;
		forwardOpInfo.oid = *oid;
		forwardOpInfo.start = 0;
		forwardOpInfo.size = object->objectHdr.size;
		forwardOpInfo.oldData = (char *) object;
		forwardOpInfo.newData = NULL;
		forwardOpInfo.actionPid.volid = oid->diskAddr.volid;
		forwardOpInfo.actionPid.page  = oid->diskAddr.page;
		forwardOpInfo.currentLRC = page->header.lrc;

		if (GET_LOGLEVEL(LOG_DATA)) {
			forwardOpInfo.action = LOG_ACTION_DESTROY_SMALL_OBJ;
			/* Log the destroy operation */
			if(clog_LogDestroyObj(&forwardOpInfo)) {
				FAIL
			}
		} else {
			forwardOpInfo.action = LOG_ACTION_DESTROY_SMALL_OBJ_ND;
			/* Log the destroy operation */
			if(clog_LogDestroyObjND(&forwardOpInfo)) {
				FAIL
			}
		}
	}

	/* free the slot */
	/* note: need to use forwardOpInfo.size here because
		clog_LogDestroyObj may overwrite the object */
	SM_ReturnSlot(link, slot, forwardOpInfo.size);
	/* SM_ReturnSlot unfixes link */

Success:
    return esmNOERROR;

FailureReturn:
	if(link) {
		bf_UnfixPage(link, BF_DEFAULT, FALSE);
	}
	return esmFAILURE;
}



int
SM_InsertForward (

	BUFGROUP			*bufGroup,
	GROUPLINK			*homeLink,
	OID					*homeOid,
	int					start,
	int					length,	
	char				*data
)
{
	PID					homePid;
	PID					forwardPid;
	SLOTTEDPAGE			*homePage;
	SLOTTEDPAGE			*forwardPage= NULL;
	PAGESLOT			*homeSlotPtr;
	PAGESLOT			*forwardSlotPtr;
	SLOTINDEX			homeSlot;
	SLOTINDEX			forwardSlot;
	int					newAlignLength;
	int					origAlignLength;
	OID					forwardOid;
	OID					newOid;
	SMALLOBJ			*homeObject;
	SMALLOBJ			*forwardObject;
	GROUPLINK			*forwardLink= NULL;
	BOOL				forwardLinkIsDirty = FALSE;
	int					oldSize,newSize;
	int					objectSize;
    OPERATIONINFO       forwardOpInfo,opInfo;


	TRPRINT(TR_SM, TR_LEVEL_1, ("length:%d", length));

    /*
     *	Construct the pid from the oid
     */
    homePid.volid = homeOid->diskAddr.volid;
    homePid.page = homeOid->diskAddr.page;
	homePage = (SLOTTEDPAGE *) homeLink->bufFrame;
	sm_AuditSlotPage(homePage);

	homeSlot = homeOid->diskAddr.slot;
	homeSlotPtr = GETSLOTPTR(homePage, homeSlot);
    homeObject = (SMALLOBJ *) GETOBJECTPTR(homePage, homeSlotPtr);

	forwardOid = *((OID *) homeObject->data);
    forwardPid.volid = forwardOid.diskAddr.volid;
    forwardPid.page = forwardOid.diskAddr.page;

	if ((forwardLink = bf_ReadPage(bufGroup, &forwardPid, H_SLOTTED,
								   SLOTTED_PAGE2SIZE, EX)) == NULL)	{
		FAIL
	}

	forwardPage = (SLOTTEDPAGE *) forwardLink->bufFrame;
	if (CHECK_SLOTTED_MAGIC(forwardPage))	{
		SM_ERROR(TYPE_FATAL, esmBADSLOTTEDMAGIC);
	}
	sm_AuditSlotPage(forwardPage);

	forwardSlot = forwardOid.diskAddr.slot;
	forwardSlotPtr = GETSLOTPTR(forwardPage, forwardSlot);
    forwardObject = (SMALLOBJ *) GETOBJECTPTR(forwardPage, forwardSlotPtr);

	if (CHECKOID(forwardSlot, forwardPage,
				 forwardSlotPtr, forwardOid.diskAddr.unique))	{
		SM_ERROR(TYPE_FATAL, esmBADOID);
	}

	SM_ASSERT(LEVEL_1, 
		(forwardObject->objectHdr.size == homeObject->objectHdr.size ));

	/*
	 *	calculate the new size of the object
	 */
	newSize = (int)homeObject->objectHdr.size + length;
	TRPRINT(TR_SM, TR_LEVEL_2, ("newSize:%d", newSize));

	/*
	 *	calculate original size
	 */
	origAlignLength = ALIGNSIZE((int)homeObject->objectHdr.size);
	TRPRINT(TR_SM, TR_LEVEL_2, ("origAlignLength:%d", origAlignLength));

	/*
	 *	round the size up to the nearest alignment value
	 */
	newAlignLength = ALIGNSIZE(newSize);
	TRPRINT(TR_SM, TR_LEVEL_2, ("newAlignLength:%d", newAlignLength));

	/*
	 *	see if the object can fit on a small page
	 */
	if (newSize > LARGETHRESHOLD)	{
		if (SM_ConvertInsert(bufGroup, homeOid, &homePid,
							 homeLink, start, (int)homeObject->objectHdr.size,
							 length, forwardObject->data, data))	{
			FAIL
		}

        /* Log the destroy of the currently forwarded object as a
                regular destroy */

		if (GET_LOGLEVEL(LOG_SPACE)) {
			forwardOpInfo.flags = 0;
			forwardOpInfo.oid = forwardOid;
			forwardOpInfo.start = 0;
			forwardOpInfo.size = (int)forwardObject->objectHdr.size;
			forwardOpInfo.newData = NULL;
			forwardOpInfo.actionPid.volid = forwardOid.diskAddr.volid;
			forwardOpInfo.actionPid.page  = forwardOid.diskAddr.page;

			if (GET_LOGLEVEL(LOG_DATA)) {
				forwardOpInfo.action = LOG_ACTION_DESTROY_SMALL_OBJ;
				forwardOpInfo.oldData = (char *) forwardObject;
				if(clog_LogDestroyObj(&forwardOpInfo)) {
					FAIL
				}
			} else {
				forwardOpInfo.action = LOG_ACTION_DESTROY_SMALL_OBJ_ND;
				forwardOpInfo.oldData = NULL;
				if(clog_LogDestroyObjND(&forwardOpInfo)) {
					FAIL
				}
			}
		}

		/* give back the space for the forwarded object */
		SM_ReturnSlot(forwardLink, forwardSlot, origAlignLength);
		forwardLink = NULL; /* unfixed by SM_ReturnSlot */
		forwardPage = NULL; /* audited by SM_ReturnSlot */

		/* update the home object to reflect the changes */
		/* homePage->header.freeBytes += (sizeof(OID) - sizeof(PID)); */
		goto Success;
	} 

    /*
     * Dealing with a small object.
     * First set up fields for logging,
     * then make sure there is enough free space for the new data.
     */

	/* prepare some of the info to log insert to forwarded part */
	if (GET_LOGLEVEL(LOG_SPACE)) {
		forwardOpInfo.action = LOG_ACTION_INSERT_SMALL_OBJ;
		forwardOpInfo.flags =0 ;
		forwardOpInfo.oid = forwardOid;
		forwardOpInfo.start = start;
		forwardOpInfo.size = length;
		forwardOpInfo.oldData = NULL;
		forwardOpInfo.newData = data;
		forwardOpInfo.actionPid.volid = forwardOid.diskAddr.volid;
		forwardOpInfo.actionPid.page  = forwardOid.diskAddr.page;
		forwardOpInfo.currentLRC = forwardPage->header.lrc;
	}

	sm_AuditSlotPage(forwardPage);

    switch (SM_CheckFree(forwardPage, forwardLink, 
		forwardSlot, forwardObject, length))	{

		case PAGE_NO_SPACE: {
			TRPRINT(TR_SM, TR_LEVEL_2, ("PAGE_NO_SPACE"));
			objectSize = (int)forwardObject->objectHdr.size;
			/* SM_CreateInsert does SM_AbortIfError() */
			if (SM_CreateInsert(bufGroup, &(homePage->header.fid), homeOid,
								length + objectSize, forwardLink, start,
								forwardObject->data, data, objectSize,
								&newOid, (OBJHDR *) forwardObject))	{
				FAIL
			}
			forwardLinkIsDirty = TRUE;

			/* Log the change to the forwarding OID in the home object*/
			if (GET_LOGLEVEL(LOG_SPACE)) {
				opInfo.action = LOG_ACTION_WRITE_SMALL_OBJ;
				opInfo.flags = 0;
				opInfo.oid = *homeOid;
				opInfo.start = 0;
				opInfo.size = sizeof(OID);
				opInfo.oldData = (char *) (homeObject->data);
				opInfo.newData = (char *) &newOid;
				opInfo.actionPid.volid = homeOid->diskAddr.volid;
				opInfo.actionPid.page  = homeOid->diskAddr.page;
				opInfo.currentLRC = homePage->header.lrc;

				if(clog_LogWriteObj(&opInfo)) {
					FAIL
				}
			}

			(*((OID *) homeObject->data)) = newOid;

			oldSize = (int)homeObject->objectHdr.size;
			homeObject->objectHdr.size = newSize;

			/* Log the update to the original header */
			if (GET_LOGLEVEL(LOG_SPACE)) {
				if(clog_LogChangeSmSize(homeOid,&(homeObject->objectHdr),
													oldSize,newSize)) {
					FAIL
				}
			}

			/* Log the delete of the currently forwarded object as a 
				regular delete */

	        /* fill in the log info - using some values filled in above */
			if (GET_LOGLEVEL(LOG_SPACE)) {
				forwardOpInfo.flags = 0;
				forwardOpInfo.start = 0;
				forwardOpInfo.size = objectSize;
				forwardOpInfo.newData = NULL;

				if (GET_LOGLEVEL(LOG_DATA)) {
					forwardOpInfo.action = LOG_ACTION_DESTROY_SMALL_OBJ;
					forwardOpInfo.oldData = (char *) forwardObject;
					/* Log the destroy operation */
					if(clog_LogDestroyObj(&forwardOpInfo)) {
						FAIL
					}
				} else {
					forwardOpInfo.action = LOG_ACTION_DESTROY_SMALL_OBJ_ND;
					forwardOpInfo.oldData = NULL;
					/* Log the destroy operation */
					if(clog_LogDestroyObjND(&forwardOpInfo)) {
						FAIL
					}
				}
			}

			/* give back old forwarded space (i.e., delete the obj) */
			SM_ReturnSlot(forwardLink, forwardSlot, origAlignLength);
			forwardLink = NULL; /* unfixed by SM_ReturnSlot */
			goto Success;
			} /* end    case PAGE_NO_SPACE */

		case PAGE_LAST_SLOT: {
			TRPRINT(TR_SM, TR_LEVEL_2, ("PAGE_LAST_SLOT"));

			/* Reestablish pointer as it may have moved */
			forwardObject = (SMALLOBJ *) GETOBJECTPTR(forwardPage, forwardSlotPtr);
			/* copy the data after the insertion point */
			bcopy(forwardObject->data + start,
				  forwardObject->data + (start + length),
				  (int) (forwardObject->objectHdr.size - start));

			forwardLinkIsDirty = TRUE;

			/* now copy in the data if necessary */
			if (data != NULL)	{
				bcopy(data, forwardObject->data + start, (int) length);
			} else {
				bzero(forwardObject->data + start, (int) length);
			}

			/*
			 *	adjust the free start and the free count
			 */
			forwardPage->header.freeBytes -= (newAlignLength - origAlignLength);
			forwardPage->header.freeStart += (newAlignLength - origAlignLength);

			/* 
			 * Update the object's header, and adjust the 
			 * user pointers into the object.
			 */
			forwardObject->objectHdr.size = newSize;
			oldSize = (int)homeObject->objectHdr.size;
			homeObject->objectHdr.size = newSize;

			/* Log the update to the original header */
			if (GET_LOGLEVEL(LOG_SPACE)) {
				if(clog_LogChangeSmSize(homeOid,&(homeObject->objectHdr),
													oldSize,newSize)){
					FAIL
				}
			}

			if (GET_LOGLEVEL(LOG_SPACE)) {
				if (GET_LOGLEVEL(LOG_DATA) && (data != NULL)) {
					/* Log the insert operation */
					if(clog_LogModifyObj(&forwardOpInfo)) {
						FAIL
					}
				} else {
					forwardOpInfo.action = LOG_ACTION_INSERT_SMALL_OBJ_ND;
					forwardOpInfo.newData = NULL;
					/* Log the insert operation */
					if(clog_LogModifyObjND(&forwardOpInfo)) {
						FAIL
					}
				}
			}
			} /* end   case PAGE_LAST_SLOT */
			break;

		case PAGE_INT_SLOT: {
            char        *source;
            char        *dest;
            int         bytesToMove;
            char        *newLocation;

			TRPRINT(TR_SM, TR_LEVEL_2, ("PAGE_INT_SLOT"));

			/*
			 *	Reestablish pointer as it may have moved
			 */
			forwardObject = (SMALLOBJ *) GETOBJECTPTR(forwardPage, forwardSlotPtr);
            /*
             *  copy the exisiting data up to the insertion point
             *  to the end of the page
             */
            newLocation = ((char*) forwardPage) + forwardPage->header.freeStart;
            bytesToMove = sizeof(OBJHDR) + start;
            bcopy((char*) forwardObject, newLocation, bytesToMove);

			forwardLinkIsDirty = TRUE;

            /*
             *  Create the hole in the data
             */
            source = forwardObject->data + start;
            dest = newLocation + bytesToMove + length;
            bytesToMove = (int)forwardObject->objectHdr.size - start;
            bcopy(source, dest, bytesToMove);

#if OBJECT_MAGIC IS_ENABLED
            /*  Null out the old object magic */
            forwardObject->objectHdr.magic = 0;
#endif OBJECT_MAGIC IS_ENABLED

            forwardObject = (SMALLOBJ *) newLocation;
            /*
             *  copy in the user data if necessary
            */
			dest = ((char *) forwardObject->data) + start;
            if (data != NULL)   {
                bcopy(data, dest, length);
            } else {
                bzero(dest, length);
			}

            /*
             *  adjust the offset in the slot
             */
            forwardSlotPtr->offset = forwardPage->header.freeStart;
			SM_ASSERT(LEVEL_1, forwardSlotPtr->offset % ALIGNON == 0);

            forwardPage->header.freeStart += sizeof(OBJHDR) + newAlignLength;

            /*
             *  Update the size in the object's header
             */
            forwardObject->objectHdr.size = newSize;
			oldSize = (int)homeObject->objectHdr.size;
            homeObject->objectHdr.size = newSize;

            /*
             *  adjust the free space to account for the extra data
             */
            forwardPage->header.freeBytes -= (newAlignLength - origAlignLength);
			TRPRINT(TR_SM, TR_LEVEL_2, ("freeBytes:%d freeStart:%d",
				forwardPage->header.freeBytes, forwardPage->header.freeStart));

			/* Log the update to the original header */
			if (GET_LOGLEVEL(LOG_SPACE)) {
				if(clog_LogChangeSmSize(homeOid,&(homeObject->objectHdr), oldSize,newSize)) {
					FAIL
				}
			}

			if (GET_LOGLEVEL(LOG_SPACE)) {
				if (GET_LOGLEVEL(LOG_DATA) && (data != NULL)) {
					/* Log the insert operation */
					if(clog_LogModifyObj(&forwardOpInfo)) {
						FAIL
					}
				} else {
					forwardOpInfo.action = LOG_ACTION_INSERT_SMALL_OBJ_ND;
					forwardOpInfo.newData = NULL;
					/* Log the insert operation */
					if(clog_LogModifyObjND(&forwardOpInfo)) {
						FAIL
					}
				}
			}
			} /* end     case PAGE_INT_SLOT */
			break;
		
		default:

			/* this cannot happen */
			SM_ERROR(TYPE_FATAL, esmINTERNAL);
			break;
    }

Success:
	if(forwardPage) {
		sm_AuditSlotPage(forwardPage);
	}
	if(forwardLink) {
		bf_UnfixPage(forwardLink, BF_DEFAULT, TRUE);
	}
    return esmNOERROR;

FailureReturn:
	if(forwardLink) {
		bf_UnfixPage(forwardLink, BF_DEFAULT, forwardLinkIsDirty);
	}
    return esmFAILURE;
}


 int
SM_InsertInLarge (
	FID				*fid,
	BUFGROUP		*bufGroup,
	OID				*oid,
	PID				*homePid,
	GROUPLINK		*homeLink,
	LARGEOBJHDR		*largeObject,
	int				start,
	int				length,
	char			*data,
	SLOTINDEX		rootSlot 
)
{
	PID				rootPid;
	int				oldSize;
	OPERATIONINFO		opInfo;

	TRPRINT(TR_SM, TR_LEVEL_1, ("start:%d length:%d", start, length));

	/*
	 *	Check to see if this is a slotted root
	 */
	if (largeObject->objectHdr.properties & P_HEADERONPAGE)	{

		LARGEOBJROOT		*largeRoot;

		TRPRINT(TR_SM, TR_LEVEL_2, ("slotted root"));

		/*
		 *	Do the actual large append
		 */
		if (lg_InsertInObject(fid, homePid, bufGroup, start, length, data, rootSlot, largeObject->objectHdr.properties & P_VERSIONED))	{

			return esmFAILURE;
		}
		 
		/*
		 *	get a pointer to the root
		 */
		largeRoot = (LARGEOBJROOT *) largeObject;

		/*
		 *	check the magic number
		CHECK_SLOTROOT_MAGIC( &(largeRoot->rootPage) );
		 */

		/*
		 *	check to see if the object has moved from its
		 *	slot on the page
		 */
		if (largeRoot->rootPage.header.numSlots == NIL)	{

			SLOTTEDPAGE		*pagePtr;

			TRPRINT(TR_SM, TR_LEVEL_2, ("slotted root removed"));

		    /* copy the part of the slotted root for logging purposes */
		    bcopy((char *) largeRoot,OldData, sizeof(LARGEOBJROOT));

			/*
		 	 *	Update the size and the root page
		 	 */
			largeObject->objectHdr.size += length;

			/*
			 *	the header can no longer fit on the page, so 
			 *	remove the header-on-page designation.
			 */
			largeObject->objectHdr.properties &= ~P_HEADERONPAGE; 

			/*
			 *	store the fid of the root
			 */
			largeObject->rootPid = homePid->page;

            /* Log the conversion from a Slotted root to a large obj header*/
     		/* note, this call will update the size field of the header */
			if (GET_LOGLEVEL(LOG_SPACE)) {
				opInfo.action = LOG_ACTION_CONVERT_TO_LARGEHDR;
				opInfo.flags = 0;
				opInfo.oid = *oid;
				opInfo.start = 0;
				opInfo.size = (int) sizeof(LARGEOBJROOT);
				opInfo.oldData = OldData;
				opInfo.newData = (char *) largeObject;
				opInfo.actionPid.volid = oid->diskAddr.volid;
				opInfo.actionPid.page  = oid->diskAddr.page;
				if(clog_LogConvertToLarge(&opInfo)) {
					FAIL
				}
			}

			/* get a pointer to the home page */
			pagePtr = (SLOTTEDPAGE *) homeLink->bufFrame;

			/* mark the page as having reclaimed the free space */
			pagePtr->header.freeBytes +=
				(ALIGNSIZE(sizeof(LGSLOTROOT)) - ALIGNSIZE(sizeof(SHORTPID)));
		} else {

		    /*
		     *	Update the size and the root page
		     */
 			oldSize = (int)largeObject->objectHdr.size;
			largeObject->objectHdr.size += length;

			/* object is still a slotted root just log the size change*/
			if (GET_LOGLEVEL(LOG_SPACE)) {
				if(clog_LogChangeSmSize(oid,&(largeObject->objectHdr), oldSize,(int)largeObject->objectHdr.size)){
					FAIL
				}
			}
	    }


	} else {

		/*
		 * Dealing with a large object.
		 * Dig out the root pid, then use large object append.
		 */
		rootPid.volid = homePid->volid;
		rootPid.page  = largeObject->rootPid;

		/*
		 *	Do the actual large append
		 */
		if (lg_InsertInObject(fid, &rootPid, bufGroup, start, length, 
							 data, NIL, largeObject->objectHdr.properties & P_VERSIONED))	{

			return esmFAILURE;
		}

		/*
		 *	Update the size
		 */
 		oldSize = (int)largeObject->objectHdr.size;
		largeObject->objectHdr.size += length;

		/* object is still a large object header just log the size change*/
		if (GET_LOGLEVEL(LOG_SPACE)) {
			if(clog_LogChangeSmSize(oid,&(largeObject->objectHdr), oldSize,(int)largeObject->objectHdr.size)) {
				FAIL
			}

			/* If the root page has changed, log it in the home object*/
			if (largeObject->rootPid !=  rootPid.page) {
				opInfo.action = LOG_ACTION_WRITE_SMALL_OBJ;
				opInfo.flags = 0;
				opInfo.oid = *oid;
				opInfo.start = 0;
				opInfo.size = sizeof(SHORTPID);
				opInfo.oldData = (char *) &(largeObject->rootPid);
				opInfo.newData = (char *) &(rootPid.page);
				opInfo.actionPid.volid = oid->diskAddr.volid;
				opInfo.actionPid.page  = oid->diskAddr.page;
				if(clog_LogWriteObj(&opInfo)) {
					FAIL
				}
			}
		}
		/* update the root page PID */
		largeObject->rootPid = rootPid.page;
	}
Success:
	return esmNOERROR;

FailureReturn:
	return esmFAILURE;
}


 int
SM_DecideNear (
	int				nearType,
	OID				*nearObject,
	PID				*nearPid,
	FID				*fid,
	BOOL			*tryNearPage 
)
{
	FLAGS 			firstOrLast=0;
	FLAGS			logOrPhys = 0;
	FILEINFO		*fileInfo;

#ifdef __cplusplus
	register int	(*nearProcedure)( FID*, PID*);
#else
	register PFI    nearProcedure;
#endif

	TRPRINT(TR_SM, TR_LEVEL_1, ("nearType:%d nearPid:%d", nearType, nearPid->page));

	/*
	 *	No procedure for finding the near page has been selected yet.
	 */
	nearProcedure = NULL;

	if ((fileInfo = fi_FindFileInfo(fid, TRUE)) == NULL) {
		return esmFAILURE;
	}

    /*
     *	Switch on near to determine where we should
     *	try and put the new object.
     */
    switch (nearType) {

		case NEAR_DISTINGUISHED:

			/*
			 *	Put the new object near the file's distinguished object.
			 */
			SM_ERROR(TYPE_USER, esmBADNEARTYPE);
			return esmFAILURE;

		case NEAR_OBJ:

			TRPRINT(TR_SM, TR_LEVEL_2, ("near specified object"));

			/*
			 *	Can put the object on the page
			 */
			*tryNearPage = TRUE;
			nearPid->volid = nearObject->diskAddr.volid;
			nearPid->page  = nearObject->diskAddr.page;
			return esmNOERROR;
		
		case NEAR_FIRST:
			firstOrLast = FI_FIRST;
			logOrPhys = fileInfo->tempVolume ? LOGICAL_SCAN : PHYSICAL_SCAN;
			break;

		case NEAR_LAST:
			firstOrLast = FI_LAST;
			logOrPhys = fileInfo->tempVolume ? LOGICAL_SCAN : PHYSICAL_SCAN;
			break;

		default:

			/*
			 *	the near type is not defined
			 */
			SM_ERROR(TYPE_USER, esmBADNEARTYPE);
			return esmFAILURE;

		} /* switch */

		nearPid->volid = fid->pid.volid;

		/*
		 *	Switch on what happened in the fi search.
		 */
		switch (fi_Page(fid, nearPid, firstOrLast | logOrPhys))	{

			case esmNOERROR:

				/*
				 *	No error, try the home page of the first object
				 *	in the file.
				 */
				*tryNearPage = TRUE;
				TRPRINT(TR_SM, TR_LEVEL_2, ("first page exists"));
				return esmNOERROR;

			case esmFAILURE:

				/*
				 *	check to see if the file is empty
				 */
				if (sm_errno == esmEMPTYFILE)	{

					/*
					 *	The file is empty.
					 *	Put the new object near the file's root.
					 */
					*tryNearPage = FALSE;
					nearPid->volid = fid->pid.volid;
					nearPid->page = fid->pid.page;
					TRPRINT(TR_SM, TR_LEVEL_2, ("file is empty"));
					SM_ERROR(TYPE_USER, esmEMPTYFILE);
					return esmNOERROR;

				} else {

					return esmFAILURE;
				}
				break;

			default:

				/*
				 *	There was some sort of problem in the file search.
				 */
				SM_ERROR(TYPE_FATAL, esmINTERNAL);
		}
	SM_ERROR(TYPE_FATAL, esmINTERNAL);
	return esmNOERROR;
}


 int
SM_TryNeighbor (
	BUFGROUP		*bufGroup,
	int				spaceNeeded,
	GROUPLINK		*nearLink,
	BOOL			*allocatePage,
	PID				*nearPid,
	GROUPLINK		**retLink 
)
{
	/* failure: doesn't unfix or fix anything 
	 * retLink can be set to nearLink:
	 * 		no neighbor, no space on neighbor
	 * or
	 * retLink can be a new (fixed) page
	 *
	 * Assumes the nearLink is fixed, and leaves it fixed
	 */
	SLOTTEDPAGE		*nearPage;
	SLOTTEDPAGE		*neighborPage;
	GROUPLINK		*neighborLink = NULL;
	PID				neighborPid;

	TRPRINT(TR_SM, TR_LEVEL_1, ("attempting to create on neighbor page"));

	nearPage = (SLOTTEDPAGE *) nearLink->bufFrame;

	/*
	 * No room for the new object.
	 * Try the neighboring page if there is one.
	 */
	neighborPid = nearPage->header.neighborPid;

	/*
	 *	check to see if the neighbor pid is valid
	 */
	if (neighborPid.volid != NIL) {

		TRPRINT(TR_SM, TR_LEVEL_2, ("valid neighbor link"));

		/*
		 *	check to see if the neighbor pid is slotted
		 */
		if (bf_CheckBitmap(&neighborPid, PAGE_SLOTTED))	{

			TRPRINT(TR_SM, TR_LEVEL_2, ("no neighbor page"));

			/*
			 * In this case, there is no neighbor page.
			 * We'll have to allocate one.
			 */
			*allocatePage = TRUE;
			*retLink = nearLink;
			/* neighborLink is NULL */
			goto Success;
		}

		/*
		 *	Read in the neighbor page
		 */
		if ((neighborLink = bf_ReadPage(bufGroup, &neighborPid, H_SLOTTED,
										SLOTTED_PAGE2SIZE, EX)) == NULL) { 
			FAIL
		}

		/*
		 *	Get a pointer to the near page
		 */
		neighborPage = (SLOTTEDPAGE *) neighborLink->bufFrame;
		TRPRINT(TR_SM, TR_LEVEL_2, ("space on neighbor page:%d",
				neighborPage->header.freeBytes));

		/*
		 *	check to see if the fids are equal
		 */
		if (!FIDEQ(neighborPage->header.fid, nearPage->header.fid))	{

			TRPRINT(TR_SM, TR_LEVEL_2, ("near page not same fid"));

			*retLink = nearLink;
			*allocatePage = TRUE;
		} else if ((CONTIGSPACE(neighborPage) >= spaceNeeded)) {
		/*
		 *	Only use the neighboring page if the neighbor pid link is
		 *	valid and if the neighboring page has enough free space.
		 *	The neighbor pid link is valid if the page specified
		 *	by the link is part the file that the new object will go in.
		 */

			TRPRINT(TR_SM, TR_LEVEL_2, ("space on neighbor"));

			/*
			 * We're in luck; the neighbor pid link is valid
			 * and the neighbor page contains enough free space.
			 */
			*nearPid  = neighborPid;
			*retLink = neighborLink;
			neighborLink = NULL;
			*allocatePage = FALSE;
		} else {

			TRPRINT(TR_SM, TR_LEVEL_2, ("no space on neighbor"));

			/*
			 *	Looks like the neighboring page doesn't cut it.
			 *	We'll have to allocate a new neighboring page
			 *	in this case.  Release the neighbor page but
			 *	keep the near page fixed.
			 */
			*retLink = nearLink;
			*allocatePage = TRUE;
		}
	} else {
		TRPRINT(TR_SM, TR_LEVEL_2, ("no neighbor page"));
		/*
		 * In this case, there is no neighbor page.
		 * We'll have to allocate one.
		 */
		*allocatePage = TRUE;
		*retLink = nearLink;
	}
Success:
	if(neighborLink) {
		bf_UnfixPage(neighborLink, BF_DEFAULT, FALSE);
	}
	return esmNOERROR;
FailureReturn:
	if(neighborLink) {
		bf_UnfixPage(neighborLink, BF_DEFAULT, FALSE);
	}
	*retLink = NULL;
	return esmFAILURE;
}


 int
SM_TryNearPage (
	BUFGROUP		*bufGroup,
	FID				*fid,
	int				spaceNeeded,
	BOOL			*allocatePage, /* OUT: TRUE if for any
									* reason the page given by nearPid
									* won't work
									*/
	PID				*nearPid,
	GROUPLINK		**retLink,	 /* OUT: NULL or a fixed page on exit */
	BOOL			*tryNearPage /* OUT: set to FALSE if retLink is NULL */
)
{
	SLOTTEDPAGE		*nearPage;
	GROUPLINK		*nearLink = NULL;
	PID				neighborPid;
	GROUPLINK		*neighborLink = NULL;

	TRPRINT(TR_SM, TR_LEVEL_1, ("attempting to create on near page"));

	SM_ASSERT(LEVEL_1, (*tryNearPage == TRUE));
	/* this function might set it to false, otherwise it
	 * leaves *tryNearPage alone 
	 */

	*retLink = NULL;

	if ((nearLink = bf_ReadPage(bufGroup, nearPid, H_SLOTTED,
								SLOTTED_PAGE2SIZE, EX)) == NULL)	{
		FAIL
	}

	nearPage = (SLOTTEDPAGE *) nearLink->bufFrame;
	TRPRINT(TR_SM, TR_LEVEL_2, 
		("space on near page:%d", nearPage->header.freeBytes));

	if (CHECK_SLOTTED_MAGIC(nearPage))	{
		SM_ERROR(TYPE_FATAL, esmBADSLOTTEDMAGIC);
	}

	if (!FIDEQ(*fid, nearPage->header.fid))	{
		TRPRINT(TR_SM, TR_LEVEL_2, ("near page not same fid"));

		*tryNearPage = FALSE;
		*allocatePage = TRUE;
		*retLink = NULL;

		/* caller is not going to use nearLink so unfix it */
		bf_UnfixPage(nearLink, BF_DEFAULT, FALSE);

		/*
		 *	release the page lock
		unlockPage(nearPid, nearLink);
		 */

		goto Success;
	}

	/*
	 *	Check to see if there is enough room on the page
	 */
	if (CONTIGSPACE(nearPage) < spaceNeeded) {
		TRPRINT(TR_SM, TR_LEVEL_2, ("no room on near page"));
		/*
		 * Try the neighboring page if there is one.
		 */
		neighborPid = nearPage->header.neighborPid;

		if (neighborPid.volid != NIL) {
			TRPRINT(TR_SM, TR_LEVEL_2, ("valid neighbor link"));

			/*
			 *	check to see if the near page is slotted
			 */
			if (bf_CheckBitmap(&neighborPid, PAGE_SLOTTED))	{
				TRPRINT(TR_SM, TR_LEVEL_2, ("no neighbor page"));

				/*
				 *	In this case, there is no neighbor page.
				 *	We'll have to allocate one.
				 */
				*allocatePage = TRUE;
				*retLink = nearLink;
				/* neighborLink is still NULL */
			} else {
				if ((neighborLink = bf_ReadPage(bufGroup, 
					&neighborPid, H_SLOTTED, SLOTTED_PAGE2SIZE, EX)) == NULL) { 
					FAIL
				}

				nearPage = (SLOTTEDPAGE *) neighborLink->bufFrame;
				TRPRINT(TR_SM, TR_LEVEL_2, ("space on neighbor page:%d",
						nearPage->header.freeBytes));

				/*
				 *	Only use the neighboring page if the neighbor pid link is
				 *	valid and if the neighboring page has enough free space.
				 *	The neighbor pid link is valid if the page specified
				 *	by the link is part the file that the new object will go in.
				 */
				if (FIDEQ(nearPage->header.fid, *fid) && (CONTIGSPACE(nearPage) >= spaceNeeded)) {

					TRPRINT(TR_SM, TR_LEVEL_2, ("space on neighbor"));

					/*
					 * We're in luck; the neighbor pid link is valid
					 * and the neighbor page contains enough free space.
					 * Release the near page.
					 */
					bf_UnfixPage(nearLink, BF_DEFAULT, FALSE);
					*nearPid  = neighborPid;
					*retLink = neighborLink;
					neighborLink = NULL;
					*allocatePage = FALSE;

				} else {
					TRPRINT(TR_SM, TR_LEVEL_2, ("no space on neighbor"));

					/*
					 *	Looks like the neighboring page doesn't cut it.
					 *	We'll have to allocate a new neighboring page
					 *	in this case.  Release the neighbor page but
					 *	keep the near page fixed.
					unlockPage( &neighborPid, neighborLink );
					 */
					*retLink = nearLink;
					*allocatePage = TRUE;
				}
			}
		} else {
			TRPRINT(TR_SM, TR_LEVEL_2, ("no neighbor page"));

			/*
			 * In this case, there is no neighbor page.
			 * We'll have to allocate one.
			 */
			*allocatePage = TRUE;
			*retLink = nearLink;
			/* neighborLink is still NULL */
		}
	} else {
		TRPRINT(TR_SM, TR_LEVEL_2, ("space on near page"));
		/*
		 *	In this case, there's enough room on the near page and
		 *	we don't even have to look at the neighbor page.  We
		 *	must keep the near page fixed in memory
		 */
		*retLink = nearLink;
		/* neighborLink is still NULL */
		*allocatePage = FALSE;
	}
Success:
	if(neighborLink) {
		bf_UnfixPage(neighborLink, BF_DEFAULT, FALSE);
	}
	return esmNOERROR;

FailureReturn:
	if(nearLink) {
		bf_UnfixPage(nearLink, BF_DEFAULT, FALSE);
	}
	if(neighborLink) {
		bf_UnfixPage(neighborLink, BF_DEFAULT, FALSE);
	}
	*retLink  = NULL;
	return esmFAILURE;
}



 int
sm_AppendToObject (
	int						groupIndex,
	OID						*oid,
	int						length,
	void					*dataPtr 
)
{
	char			*data = (char*)dataPtr;  /* rename dataPtr */
    SLOTINDEX		objectSlot;		/* the object's slot			*/
    SLOTTEDPAGE		*homePage = NULL;/* object's home page			*/
    int				objectSize;		/* the object's data size		*/
    int				newSize;		/* size of the object after 	*/
	int				origAlignLength;
	int				newAlignLength;
    SMALLOBJ		*objectPtr;		/* ptr to generic object		*/
    PID				homePid;		/* home page of object			*/
	PAGESLOT		*slotPtr;
	GROUPLINK		*homeLink = NULL;
	BOOL			homeLinkIsDirty = FALSE;
	BUFGROUP		*bufGroup;
	LIST			*suspendList= NULL;
	OPERATIONINFO	opInfo;


	TRPRINT(TR_SM, TR_LEVEL_1, ("length:%d", length));
	
	INIT_SMSTATE(&SM_State);

	if (checkTransState(oid->diskAddr.volid))	{
		return esmFAILURE;
	}

    /*
     *	Construct the pid from the oid
     */
    homePid.volid = oid->diskAddr.volid;
    homePid.page = oid->diskAddr.page;

	bufGroup = &(BufGroupPtr[groupIndex]);
	if (CHECK_BUFGROUP_VALID(bufGroup))	{
		SM_ERROR(TYPE_USER, esmBADGROUP);
		FAIL
	}

    if ((homeLink = bf_ReadPage(bufGroup, &homePid, H_SLOTTED,
								SLOTTED_PAGE2SIZE, EX)) == NULL)	{
		FAIL
	}
	if (SM_SetLogServerState(homeLink, oid) != esmNOERROR) {
		FAIL
	}

	/*
	 *	Get a pointer to the data page
	 */
	homePage = (SLOTTEDPAGE *) homeLink->bufFrame;
	sm_AuditSlotPage(homePage);
	if (CHECK_SLOTTED_MAGIC(homePage))	{
		SM_ERROR(TYPE_FATAL, esmBADSLOTTEDMAGIC); /* internal error */
	}

	/*
	 *	get the slot position and a pointer to the slot
	 */
	objectSlot = oid->diskAddr.slot;
	slotPtr = GETSLOTPTR(homePage, objectSlot);

    /*
     * Make sure we have a valid oid.
     */
    if (CHECKOID(objectSlot, homePage, slotPtr, oid->diskAddr.unique))	{
		SM_ERROR(TYPE_USER, esmBADOID);
		FAIL
	}
	TRPRINT(TR_SM, TR_LEVEL_2, ("valid oid"));

    /*
     *	Get a pointer to the existing object
     */
    objectPtr = (SMALLOBJ *) GETOBJECTPTR(homePage, slotPtr);

    /*
     * Make sure object is either not versioned or not frozen
     */
    if ( (objectPtr->objectHdr.properties & (P_VERSIONED | P_FROZEN)) ==
         (P_VERSIONED | P_FROZEN) ) {
        SM_ERROR(TYPE_USER, esmFROZEN);
		FAIL
    }

	if (length == 0) {
		goto Success;
	}

	/*
	 *	suspend any outstanding pointers if necessary
	 */
	suspendList = SM_SuspendDesc(oid, NULL);

	sm_SetAbortIfError();

    /*
     * Determine whether we are dealing with a large or small object.
     */
    if (objectPtr->objectHdr.properties & P_LARGEOBJ) {

		/* Verify that there is enough space to complete operation */
		if (bf_CheckForSpace(bufGroup, length, TRUE) != esmNOERROR) {
			FAIL
		}

		/*
		 *	Do the append and update the sizes
		 */
		if (SM_AppendToLarge( &(homePage->header.fid), bufGroup, oid,
		 	  homeLink, (LARGEOBJHDR *) objectPtr, length, data, objectSlot))	{
			FAIL
		}

		goto Success;

    } else if (objectPtr->objectHdr.properties & P_MOVED)	{

		/* Verify that there is enough space to complete operation */
		if (bf_CheckForSpace(bufGroup, length, TRUE) != esmNOERROR) {
			FAIL
		}

		if (SM_AppendForward(bufGroup, homeLink, oid, length, data))	{
			FAIL;
		}
		homeLinkIsDirty = TRUE;
		goto Success;

	} else if (objectPtr->objectHdr.properties & P_FORWARD)	{
		SM_ERROR(TYPE_FATAL, esmINTERNAL);
	}

	/*
	 *	Calculate its total final size
	 */
    newSize = (int)objectPtr->objectHdr.size + length;
	TRPRINT(TR_SM, TR_LEVEL_2, ("properties:%x newSize:%d",
			objectPtr->objectHdr.properties, newSize));

	/*
	 *	calculate original size
	 */
	origAlignLength = ALIGNSIZE((int)objectPtr->objectHdr.size);
	TRPRINT(TR_SM, TR_LEVEL_2, ("origAlignLength:%d", origAlignLength));

	/*
	 *	round the size up to the nearest alignment value
	 */
	newAlignLength = ALIGNSIZE(newSize);
	TRPRINT(TR_SM, TR_LEVEL_2, ("newAlignLength:%d", newAlignLength));

	/*
	 *	check to see if the object can fit on a small page
	 */
	if (newSize > LARGETHRESHOLD)	{

		/* Verify that there is enough space to complete operation */
		if (bf_CheckForSpace(bufGroup, length, TRUE) != esmNOERROR) {
			FAIL
		}

		/*
		 *	convert the object to a large
		 */
		if (SM_ConvertAppend(bufGroup, oid, &homePid, homeLink, (int)objectPtr->objectHdr.size,
							 length, objectPtr->data, data))	{

			FAIL
		}
		homeLinkIsDirty = TRUE;
		goto Success;
	}

    /*
     * Dealing with a small object.
     * First set up fields for logging,
     * then make sure there is enough free space for the new data.
     */

	/* NOTE: this assumes no forwarding */
    opInfo.action = LOG_ACTION_APPEND_SMALL_OBJ;
    opInfo.flags = 0;
    opInfo.oid = *oid;
    opInfo.start = (int) objectPtr->objectHdr.size;
    opInfo.size = length;
    opInfo.oldData = NULL;
    opInfo.newData = data;
    opInfo.actionPid.volid = oid->diskAddr.volid;
    opInfo.actionPid.page  = oid->diskAddr.page;
    opInfo.currentLRC = homePage->header.lrc;


    switch (SM_CheckFree(homePage, homeLink, objectSlot, objectPtr, length)){

		case PAGE_NO_SPACE: {
			TRPRINT(TR_SM, TR_LEVEL_2, ("PAGE_NO_SPACE"));

			/*
			 *	check to see if there is room to store the forwarding oid
			 */
			if (objectPtr->objectHdr.size < (int)sizeof(OID))	{

				/*
				 *	convert that object to a large object
				 */
				if (SM_ConvertAppend(bufGroup, oid, &homePid, homeLink,
									 (int)objectPtr->objectHdr.size, length,
							  		 objectPtr->data, data))	{
					FAIL;
				}
				homeLinkIsDirty = TRUE;
			} else {
				OID			newOid;

				objectSize = (int)objectPtr->objectHdr.size;

				/*
				 *	attempt to forward the object to the neighbor page
				 */
				if (SM_CreateAppend(bufGroup, &(homePage->header.fid), oid,
						length + objectSize, homeLink, objectPtr->data, data,
						objectSize, &newOid, (OBJHDR *) objectPtr))	{
					homeLinkIsDirty = TRUE;
					FAIL
				}
				homeLinkIsDirty = TRUE;

				if (GET_LOGLEVEL(LOG_SPACE)) {

					/* delete the data portion of the original object after
						the oid that will be inserted by the forwarding*/
					opInfo.flags = 0;
					opInfo.oid = *oid;
					opInfo.start = (int)sizeof(OID);
					opInfo.size = (int)objectPtr->objectHdr.size - (int)sizeof(OID);
					opInfo.newData = NULL;
					opInfo.actionPid.volid = oid->diskAddr.volid;
					opInfo.actionPid.page  = oid->diskAddr.page;
					opInfo.currentLRC = homePage->header.lrc;

					if (GET_LOGLEVEL(LOG_DATA)) {
						opInfo.action = LOG_ACTION_DELETE_SMALL_OBJ;
						opInfo.oldData = (char *) (objectPtr->data) + (int)sizeof(OID);
						/* log as a regular delete */
						if(clog_LogModifyObj(&opInfo)) {
							FAIL
						}
					} else {
						opInfo.action = LOG_ACTION_DELETE_SMALL_OBJ_ND;
						opInfo.oldData = NULL;
						/* log as a regular delete */
						if(clog_LogModifyObjND(&opInfo)) {
							FAIL
						}
					}
				}

                /* update the object header */
                objectPtr->objectHdr.size = (int)sizeof(OID);

                /* free up space on the page */
                homePage->header.freeBytes += 
                            (origAlignLength - ALIGNSIZE((int)sizeof(OID)));

                /* copy the old object header and the bytes of data that
                   will be overwritten by the OID */
                bcopy((char *) objectPtr,OldData,
                   (int) (ALIGNSIZE((int)sizeof(OBJHDR)) + ALIGNSIZE((int)sizeof(OID))));

				objectPtr->objectHdr.properties |= P_MOVED;
				(*((OID *) objectPtr->data)) = newOid;
				objectPtr->objectHdr.size = newSize;

				if (GET_LOGLEVEL(LOG_SPACE)) {
					/* Log the forwarding info for the home (original) object */

					opInfo.flags = 0;
					opInfo.oid = *oid;
					opInfo.start = 0;
					opInfo.size = 0;
					opInfo.oldData = (char *) OldData;
					opInfo.newData = (char *) objectPtr;

					opInfo.action = LOG_ACTION_FORWARD_SMALL_OBJ;
					/* write the log record*/
					if(clog_LogForwardObj(&opInfo)) {
						FAIL
					}
				}
			}
			goto Success;
			} /* end  case PAGE_NO_SPACE */

		
		case PAGE_LAST_SLOT: {
			TRPRINT(TR_SM, TR_LEVEL_2, ("PAGE_LAST_SLOT"));

			/*
			 *	Reestablish pointer as it may have moved
			 */
			objectPtr = (SMALLOBJ *) GETOBJECTPTR(homePage, slotPtr);

			/*
			 *	we can just copy in the data
			 */
			if (data != NULL)	{
				bcopy(data, ((char *) objectPtr->data +
									 objectPtr->objectHdr.size), (int) length);
			} else {
				bzero(((char *) objectPtr->data + objectPtr->objectHdr.size), (int) length);
			}
			homeLinkIsDirty = TRUE;

			/*
			 *	adjust the free start and the free count
			 */
			homePage->header.freeBytes -= (newAlignLength - origAlignLength);
			homePage->header.freeStart += (newAlignLength - origAlignLength);

			/* 
			 * Update the object's header, and adjust the 
			 * user pointers into the object.
			 */
			objectPtr->objectHdr.size += length;

			if (GET_LOGLEVEL(LOG_SPACE)) {
				if (GET_LOGLEVEL(LOG_DATA) && (data != NULL)) {
					/* Log the append operation */
					if(clog_LogModifyObj(&opInfo)) {
						FAIL
					}
				} else {
					opInfo.action = LOG_ACTION_APPEND_SMALL_OBJ_ND;
					/* Log the append operation */
					if(clog_LogModifyObjND(&opInfo)) {
						FAIL
					}
				}
			} 
			} /* end  case PAGE_LAST_SLOT */
			break;

		
		case PAGE_INT_SLOT: {
			TRPRINT(TR_SM, TR_LEVEL_2, ("PAGE_INT_SLOT"));

			/*
			 *	Reestablish pointer as it may have moved
			 */
			objectPtr = (SMALLOBJ *) GETOBJECTPTR(homePage, slotPtr);

			bcopy((char *) objectPtr, ((char *) homePage) + homePage->header.freeStart,
				  (int) ((int)sizeof(OBJHDR) + objectPtr->objectHdr.size));

			homeLinkIsDirty = TRUE;

			slotPtr->offset = homePage->header.freeStart;
			SM_ASSERT(LEVEL_1, slotPtr->offset % ALIGNON == 0);
			objectPtr = (SMALLOBJ *) GETOBJECTPTR(homePage, slotPtr);
			/* 
			 *	Update the object's header, and adjust the 
			 *	user pointers into the object.
			 */

			if (data != NULL)	{
				bcopy(data, &(objectPtr->data[objectPtr->objectHdr.size]), (int) length);
			} else {
				bzero(&(objectPtr->data[objectPtr->objectHdr.size]), (int) length);
			}
			objectPtr->objectHdr.size += length;

			/*
			 *	adjust the free start to account for the extra data
			 */
			homePage->header.freeStart += (int)sizeof(OBJHDR) + newAlignLength;

			/*
			 *	adjust the free space to account for the extra data
			 */
			homePage->header.freeBytes -= (newAlignLength - origAlignLength);
			TRPRINT(TR_SM, TR_LEVEL_2, ("freeBytes:%d freeStart:%d",
					homePage->header.freeBytes, homePage->header.freeStart));

			if (GET_LOGLEVEL(LOG_SPACE)) {
				if (GET_LOGLEVEL(LOG_DATA) && (data != NULL)) {
					/* Log the append operation */
					if(clog_LogModifyObj(&opInfo)) {
						FAIL
					}
				} else {
					opInfo.action = LOG_ACTION_APPEND_SMALL_OBJ_ND;
					/* Log the append operation */
					if(clog_LogModifyObjND(&opInfo)) {
						FAIL
					}
				}
			} 
			} /* end  case PAGE_INT_SLOT */
			break;

		
		default:
			/*
			 *	this cannot happen
			 */
			SM_ERROR(TYPE_FATAL, esmINTERNAL);
			break;
    }


Success:
	if(homePage) {
		sm_AuditSlotPage(homePage);
	}
	if(homeLink) {
		bf_UnfixPage(homeLink, BF_DEFAULT, TRUE);
	}
	/* restore suspended user descriptors if necessary */
	if (suspendList != NULL)	{
		SM_RestoreDesc(suspendList);
	}

	/* End the action and return */
	BF_AuditFreeList();
    return esmNOERROR;

FailureReturn:
	if(homeLink) {
		bf_UnfixPage(homeLink, BF_DEFAULT, homeLinkIsDirty);
	}
	/* restore suspended user descriptors if necessary */
	if (suspendList != NULL)	{
		SM_RestoreDesc(suspendList);
	}
	SM_AbortIfRequired();
	return esmFAILURE;
}



 int
sm_CreateObject (
	int						groupIndex,
	FID						*fid,
	int						nearType,
	OID						*nearObj,
	OBJHDR					*objectHdr,
	int						length,
	void					*dataPtr,
	OID						*oid 
)
{
	char			*data = (char*)dataPtr;	/* rename dataPtr 		*/
    PAGESLOT		*slotPtr;		/* slot pointer					*/
    LARGEOBJHDR		*largeObject;	/* ptr to large object header	*/
    SMALLOBJ		*smallObject;	/* small object pointer			*/
    SLOTTEDPAGE		*homePage;		/* ptr to home page				*/
    SLOTINDEX		emptySlot;		/* slot used for the new obj	*/
    int				spaceNeeded;	/* space needed for the new obj */
	int				alignLength;
    BOOL			creatingLargeObj; /* whether creating large obj	*/
    BOOL			tryNearPage;	/* whether to look at near page */
    BOOL			allocatePage;	/* whether to allocate new page */
    BOOL		    setNeighborLink; /* " set neighbor pid link */
    BOOL			allocateSlot; /* whether we need to allocate slot */
	BOOL			rootObject;
    PID				nearPid;		/* create the obj near this page */
    SLOTTEDPAGE		*nearPage;		/* indirect ptr to near page	*/
    PID				homePid;		/* the new obj's home page		*/
    PID				rootPid;		/* large object root page		*/
	GROUPLINK		*nearLink = NULL;
	GROUPLINK		*homeLink = NULL;
	BOOL			homeLinkIsDirty = FALSE;
	BUFGROUP		*bufGroup;
	UNIQUE			unique;
	OPERATIONINFO   opInfo;  /* operation info used for logging */
    int				smDataSize;		/*data portion sz for lg obj hdr */


	TRPRINT(TR_SM, TR_LEVEL_1, ("length:%d", length));

	INIT_SMSTATE(&SM_State);

	/*
	 *	see if a transaction is active locally and on the server
	 *  that serves the volume in question.
	 */
	if (checkTransState(fid->pid.volid))	{
		return esmFAILURE;
	}

	bufGroup = &BufGroupPtr[groupIndex];
	if (CHECK_BUFGROUP_VALID(bufGroup))	{
		SM_ERROR(TYPE_USER, esmBADGROUP);
		return esmFAILURE;
	}

	/*
	 *	round the size up to the nearest alignment value
	 */
	alignLength = ALIGNSIZE(length);
	TRPRINT(TR_SM, TR_LEVEL_2, ("alignLength:%d", alignLength));

    /*
     *	Are we creating a large or small object?
     *	In calculating how much space we'll need, waste a little space by 
     *	assuming that we'll need to allocate a new slot; this will save 
     *	time in the calculations below and isn't much of a price to pay.
     */
    if (alignLength > LARGETHRESHOLD) {
		creatingLargeObj = TRUE;

		if (length <= (LSLOTTHRESHOLD))	{
			spaceNeeded = ALIGNSIZE(sizeof(LARGEOBJROOT)) + sizeof(PAGESLOT);
			rootObject  = TRUE;
			TRPRINT(TR_SM, TR_LEVEL_2, ("creating slotted large"));
		} else {
			/*
			 *	allocate enough space to hold the header only
			 */
			spaceNeeded = sizeof(LARGEOBJHDR) + sizeof(PAGESLOT);
			rootObject  = FALSE;
			TRPRINT(TR_SM, TR_LEVEL_2, ("creating large object on seperate"));
		}
    } else {
		creatingLargeObj = FALSE;
		spaceNeeded		 = sizeof(OBJHDR) + alignLength + sizeof(PAGESLOT);
		TRPRINT(TR_SM, TR_LEVEL_2, ("creating small object"));
    }
	TRPRINT(TR_SM, TR_LEVEL_2, ("spaced needed:%d", spaceNeeded));

	/*
	 *	Verify that there is sufficient space to complete the operation
	 */
	if (bf_CheckForSpace(bufGroup, alignLength, creatingLargeObj) != esmNOERROR) {
		FAIL
	}

    /*
     *	Switch on near to determine where we should
     *	try to put the new object.
     */
	if (SM_DecideNear(nearType, nearObj, &nearPid, fid, &tryNearPage))	{
		/* it's a user error; SM_DecideNear already set sm_errno */
		FAIL
	}

    /*
     *	Try and create the new object on nearObj's home page.
     */
    if (tryNearPage) {
		TRPRINT(TR_SM, TR_LEVEL_2, ("attempting to create on near page"));

		if (SM_TryNearPage(bufGroup, fid, spaceNeeded, &allocatePage,
						   &nearPid, &nearLink, &tryNearPage))	{
			SM_ASSERT(LEVEL_3, (nearLink == NULL));
			FAIL
		}

		/*
		 *	check to see if we have a valid page
		 */
		if (!allocatePage)	{
			homePid = nearPid;
			homeLink = nearLink;
			homePage = (SLOTTEDPAGE *) homeLink->bufFrame;
			nearLink = NULL; /* sanity:so we don't doubly unfix the page */
		} else if (tryNearPage)	{
			/*
			 *	set nearPage for future use (we update its neighborPid)
			 */
			nearPage = (SLOTTEDPAGE *) nearLink->bufFrame;
		}
    } else {
		/* ! tryNearPage  */
		SM_ASSERT(LEVEL_3, (homeLink == NULL));
		SM_ASSERT(LEVEL_3, (nearLink == NULL));

		TRPRINT(TR_SM, TR_LEVEL_2, ("not using near page"));

		/*
		 * In this case, there is no nearObj page to look at and
		 * we have to allocate a page.
		 */
		allocatePage = TRUE;
	}

    /*
     *	Figure out whether we have to set the neighbor pid link.
	 *  If the nearLink is defined and legit (tryNearPage==TRUE) and
	 *  we are allocating a new page, we'll update the neighbor pid link.
     */
    setNeighborLink = tryNearPage && allocatePage;

    /*
     *	Allocate a page if necessary.
     *	If setNeighborLink is TRUE, then nearPage is valid
     *	and points to the near page whose neighbor pid link has to be set.
     */
    if (allocatePage == TRUE) {
		SM_ASSERT(LEVEL_3, (homeLink == NULL));
		SM_ASSERT(LEVEL_3,  (
		((setNeighborLink==FALSE) && (nearLink == NULL))
			|| 
		((setNeighborLink==TRUE) && (nearLink != NULL))
			));
		/*
		 * Do the low-level allocation.
		 */
		if (io_AllocSlottedPages(1, &nearPid, &homePid, fid, creatingLargeObj,
								 &unique, bufGroup))	{
			FAIL
		}
		TRPRINT(TR_SM, TR_LEVEL_2, ("allocated new page:%d", homePid.page));

		if ((homeLink = bf_FakeReadPage(bufGroup, &homePid, H_SLOTTED,
								   SLOTTED_PAGE2SIZE)) == NULL)	{
			FAIL
		}

		homePage = (SLOTTEDPAGE *) homeLink->bufFrame;
		INITPAGE(homePage, homePid, *fid, creatingLargeObj, unique);

		homeLinkIsDirty = TRUE;
		sm_SetAbortIfError();

		if (SM_SetLogServerState(homeLink, NULL) != esmNOERROR) {
			FAIL
		}
		homePage->header.lrc = *(GET_LOGINFO->initLRC);

		if (setNeighborLink == TRUE) {
			TRPRINT(TR_SM, TR_LEVEL_2, ("setting neighbor link"));
			/*
			 *	set the link but don't dirty the page since it
			 *	is just a hint
			 */
			nearPage->header.neighborPid = homePid;
			TRPRINT(TR_SM, TR_LEVEL_2, ("setting neighbor link"));
		}
    } else {
		SM_ASSERT(LEVEL_3, (nearLink == NULL));
		SM_ASSERT(LEVEL_3, (homeLink != NULL));

		if (SM_SetLogServerState(homeLink, NULL) != esmNOERROR) {
			FAIL
		}
		if (creatingLargeObj == TRUE)	{
			if (SM_LargeObjOnPage(homePage)) {
				FAIL
			}
			/* set homeLinkIsDirty = TRUE below */
		}
    }
	homeLinkIsDirty = TRUE;
	sm_SetAbortIfError();

	/*
	 *	check to see if there are any stray free slots
	 */
	if (homePage->header.freeSlot == NIL)	{
		TRPRINT(TR_SM, TR_LEVEL_2, ("slot is new"));

		/*
		 *	Just grab the next slot after current last
		 */
		emptySlot = homePage->header.slotCount;
		TRPRINT(TR_SM, TR_LEVEL_2, ("using free slot:%d", emptySlot));

		slotPtr = GETSLOTPTR(homePage, emptySlot);
		homePage->header.slotCount++;
		homeLinkIsDirty = TRUE;
		allocateSlot = TRUE;
	} else {
		emptySlot = homePage->header.freeSlot;
		TRPRINT(TR_SM, TR_LEVEL_2, ("using free slot:%d", emptySlot));

		slotPtr = GETSLOTPTR(homePage, emptySlot);
		homePage->header.freeSlot = (SLOTINDEX) slotPtr->unique;
		TRPRINT(TR_SM, TR_LEVEL_2, ("new free slot:%d", homePage->header.freeSlot));
		allocateSlot = FALSE;
	}

     if (creatingLargeObj == TRUE) {
		/*
		 *	fill in the root page in case we have the 
		 *	large object header on the slotted page
		 */
		rootPid = homePid;
		slotPtr->offset = homePage->header.freeStart;
		SM_ASSERT(LEVEL_1, slotPtr->offset % ALIGNON == 0);
		largeObject = (LARGEOBJHDR *)
					 ((char *) homePage + homePage->header.freeStart);

		/* mark the object as create in progress so the lg code won't
			try to log any changes to the slotted root before its create
			is logged */
		largeObject->objectHdr.properties |= P_CREATEINPROGRESS;

        /*
         * Creating a large object.
         * Large object create will fill in the rootPid.
         * For root objects, will call lg_Append after the create of an empty
         *    lg object is logged - to avoid create-in-progress logging probs.
         */
        if (lg_CreateObject(fid, &homePid, bufGroup, (rootObject ? 0 : length),
                           data, &rootPid, (rootObject ? emptySlot : NIL))) {

			slotPtr->offset = NIL;
			if (allocateSlot) {
				homePage->header.slotCount--;
			} else {
				slotPtr->unique = homePage->header.freeSlot;
				homePage->header.freeSlot = emptySlot;
			}
			FAIL;
		}

		/* mark that the create is done */
		largeObject->objectHdr.properties &= ~P_CREATEINPROGRESS;
		largeObject->objectHdr.size	 = length;

#if OBJECT_MAGIC IS_ENABLED
		largeObject->objectHdr.magic = OBJECT_HDR_MAGIC;
#endif OBJECT_MAGIC IS_ENABLED

		/*
		 *	check to see if we can use the unique on the page
		 */
		if (homePage->header.uniqueCount == 0)	{
			if (io_GetUnique(&homePid, &unique, UNIQUE_ALLOC_COUNT))	{
				FAIL
			}
			homePage->header.unique = unique;
			homePage->header.uniqueCount = UNIQUE_ALLOC_COUNT;
		}

		slotPtr->unique = homePage->header.unique++;
		homePage->header.uniqueCount--;
		TRPRINT(TR_SM, TR_LEVEL_2, ("unique:%x", slotPtr->unique));

		if (objectHdr != NULL)	{
			largeObject->objectHdr.tag = objectHdr->tag;

			/*
			 *	See if the new object is a version history graph, the
			 *	call must be made from within an sm function.
			 */
			if ( (objectHdr->properties & P_HISTORYGRAPH) && 
				 InternalCall) {
				/*
				 *	Mark the obj as a VHG
				 */
				largeObject->objectHdr.properties = P_HISTORYGRAPH;
			} else {
				/*
				 *	No properties caller, so initialize the
				 *	properties field so that additional properties
				 *	can be OR'd on.
				 */
				 largeObject->objectHdr.properties = 0;
			}
		} else {
			/*
			 *	No header provided by caller, so initialize the
			 *	properties field so that additional properties
			 *	can be OR'd on.
			 */
			 largeObject->objectHdr.properties = 0;
		}

		/*
		 *	Check to see if we created a root object
		 */
		if (rootObject)	{

			largeObject->objectHdr.properties |= P_LARGEOBJ|P_HEADERONPAGE;
			/*
			 *	Make sure object agrees with P_HEADERONPAGE
			 */
			CHECK_SLOTROOT_MAGIC( (&((LARGEOBJROOT*)largeObject)->rootPage) );

			homePage->header.freeStart += ALIGNSIZE(sizeof(LARGEOBJROOT));
			homePage->header.freeBytes -= ALIGNSIZE(sizeof(LARGEOBJROOT));

			/* calc the size of the data portion of the root - for logging */
		    smDataSize = sizeof(LARGEOBJROOT) - sizeof(OBJHDR);

		} else {
			
			largeObject->objectHdr.properties |= P_LARGEOBJ;

			homePage->header.freeStart += sizeof(LARGEOBJHDR);
			homePage->header.freeBytes -= sizeof(LARGEOBJHDR);

			/* calc the size of the data portion of the header - for logging */
		    smDataSize = sizeof(LARGEOBJHDR) - sizeof(OBJHDR);

			largeObject->rootPid = rootPid.page;
		}

		/*
		 *	if we are allocating a slot then subtract that space
		 */
		if (allocateSlot == TRUE) {
			homePage->header.freeBytes -= sizeof(PAGESLOT);
		}

		oid->diskAddr.page	 = homePid.page;
		oid->diskAddr.slot	 = emptySlot;
		oid->diskAddr.unique = slotPtr->unique;
		oid->diskAddr.volid  = homePid.volid;
		TRPRINT(TR_SM, TR_LEVEL_2, ("returning oid pid:%d slot:%d unique:%x",
				oid->diskAddr.page, emptySlot, slotPtr->unique));

		if (GET_LOGLEVEL(LOG_SPACE)) {

			/* 
			 *	log the create of the large object header as a small
			 *	object.  This must be logged if space logging is on
			 */
			opInfo.action = LOG_ACTION_CREATE_SMALL_OBJ;
			opInfo.flags = 0;
			opInfo.oid = *oid;
			opInfo.start = 0;
			opInfo.size = smDataSize;
			opInfo.oldData = NULL;
			opInfo.newData = (char *) largeObject;
			opInfo.actionPid.volid = oid->diskAddr.volid;
			opInfo.actionPid.page  = oid->diskAddr.page;
			if(clog_LogCreateObj(&opInfo)) {
				FAIL
			}
		}

        /* now append data if this is a slotted root object */
        if (rootObject) {
            if (lg_AppendToObject(fid, &rootPid, bufGroup, length,
                            data,  emptySlot, FALSE)) {

				FAIL
            }
        }
		goto Success;
    }
    /*
     * If we got here, we're creating a small object.
     * Copy the object to the home page.
     */
    smallObject = (SMALLOBJ *) (((char *) homePage) + homePage->header.freeStart);
    smallObject->objectHdr.properties = 0;
    smallObject->objectHdr.size		  = length;
	TRPRINT(TR_SM, TR_LEVEL_2, ("freestart:%d length:%d",
			homePage->header.freeStart, length));

#if OBJECT_MAGIC IS_ENABLED
	smallObject->objectHdr.magic      = OBJECT_HDR_MAGIC;
#endif OBJECT_MAGIC IS_ENABLED

	if (objectHdr != NULL)	{
		smallObject->objectHdr.tag = objectHdr->tag;

		/*
		 *	See if the new object is a version history graph, the
		 *	call must be made from within an sm function.
		 */
		if ( (objectHdr->properties & P_HISTORYGRAPH) && InternalCall) {
			/*
			 *	Mark the obj as a VHG
			 */
			smallObject->objectHdr.properties = P_HISTORYGRAPH;
		}
	}

    /*
     *	Only copy the data in if the pointer is valid
     */
    if (data != NULL)	{
    	bcopy(data, smallObject->data, (int) length);
    } else {
    	bzero(smallObject->data, (int) length);
	}

    slotPtr->offset = homePage->header.freeStart;
	SM_ASSERT(LEVEL_1, slotPtr->offset % ALIGNON == 0);

	/*
	 *	check to see if we can use the unique on the page
	 */
	if (homePage->header.uniqueCount == 0)	{

		if (io_GetUnique(&homePid, &unique, UNIQUE_ALLOC_COUNT))	{
			FAIL
		}

		homePage->header.unique = unique;
		homePage->header.uniqueCount = UNIQUE_ALLOC_COUNT;
	}

	slotPtr->unique = homePage->header.unique++;
	homePage->header.uniqueCount--;
	TRPRINT(TR_SM, TR_LEVEL_2, ("unique:%x", slotPtr->unique));

	/* 
	 *	There must always be room for a page id in an object
	 *	so that in can be enlarged into a large object.
	 */
	SM_ASSERT(LEVEL_3, alignLength >= sizeof(SHORTPID));

    homePage->header.freeStart += sizeof(OBJHDR) + alignLength;
    homePage->header.freeBytes -= sizeof(OBJHDR) + alignLength;

    SM_ASSERT(LEVEL_3, !CHECK_SLOTTED_MAGIC(homePage));

    if (allocateSlot == TRUE) {
		homePage->header.freeBytes -= sizeof(PAGESLOT);
    }
	TRPRINT(TR_SM, TR_LEVEL_2, ("freeStart:%d freeBytes:%d",
			homePage->header.freeStart, homePage->header.freeBytes));

    oid->diskAddr.page	 = homePid.page;
    oid->diskAddr.slot	 = emptySlot;
	oid->diskAddr.unique = slotPtr->unique;
	oid->diskAddr.volid  = homePid.volid;
	TRPRINT(TR_SM, TR_LEVEL_2, ("oid page:%d slot:%d unique:%d",
			oid->diskAddr.page, oid->diskAddr.slot, oid->diskAddr.unique));

	if (GET_LOGLEVEL(LOG_SPACE)) {
		opInfo.flags = 0;
		opInfo.oid = *oid;
		opInfo.start = 0;
		opInfo.size = (int)smallObject->objectHdr.size;
		opInfo.oldData = NULL;
		opInfo.newData = (char *) smallObject;
		opInfo.actionPid.volid = oid->diskAddr.volid;
		opInfo.actionPid.page  = oid->diskAddr.page;

		if (GET_LOGLEVEL(LOG_DATA) && (data != NULL)) {
			opInfo.action = LOG_ACTION_CREATE_SMALL_OBJ;
			/* Log the create operation */
			if(clog_LogCreateObj(&opInfo)) {
				FAIL
			}
		} else {
			opInfo.action = LOG_ACTION_CREATE_SMALL_OBJ_ND;
			/* Log the create operation */
			if(clog_LogCreateObjND(&opInfo)) {
				FAIL
			}
		}
	}

Success:
    SM_ASSERT(LEVEL_3, !CHECK_SLOTTED_MAGIC(homePage));
	if (nearLink != NULL) {
		bf_UnfixPage(nearLink, BF_DEFAULT, FALSE); /* a lie */
	}
	sm_AuditSlotPage(homePage);
    bf_UnfixPage(homeLink, BF_DEFAULT, TRUE);

	BF_AuditFreeList();
    return esmNOERROR;

FailureReturn:
	if (nearLink != NULL) {
		bf_UnfixPage(nearLink, BF_DEFAULT, FALSE);
	}
	if (homeLink != NULL) {
		bf_UnfixPage(homeLink, BF_DEFAULT, homeLinkIsDirty);
	}
	SM_AbortIfRequired();
	return esmFAILURE;
}

 int
SM_ConvertDelete(
	BUFGROUP		*bufGroup,
	OID				*oid,
	PID				*homePid,
	GROUPLINK		*homeLink,
	OBJECT		 	*objectPtr,
	int			 	length,
	PID				*rootPid,
	SLOTINDEX		rootSlot
)
{
	OBJHDR			tmpHdr;
	OPERATIONINFO	opInfo;
	int				origAlignLength;
	int				newAlignLength;
	SLOTTEDPAGE		*homePage = (SLOTTEDPAGE*) homeLink->bufFrame;

	/*
	 *	This function is only implemented for objects where all
	 *	of the data is being deleted.  If only enough data is deleted
	 *	to make the object small, the this function needs to be
	 *	extended.
	 */
	SM_ASSERT(LEVEL_3, length == objectPtr->objectHdr.size);
	SM_ASSERT(LEVEL_3, objectPtr->objectHdr.properties & P_LARGEOBJ);

	/* first, destroy the large object tree */
	/* this lg_ function calls  sm_SetAbortIfError(); */
	if (lg_DestroyObject(rootPid, bufGroup, rootSlot, FALSE,
					WORKING_VERSION(objectPtr->objectHdr.properties))) {
		FAIL
	}

	newAlignLength = ALIGNSIZE(0);

	/* delete lg btree root or root pid */
	tmpHdr = objectPtr->objectHdr;
	tmpHdr.size = 0;
	if (objectPtr->objectHdr.properties & P_HEADERONPAGE) {
	
		origAlignLength = ALIGNSIZE((int)sizeof(LGSLOTROOT));
		
		/*	change the object's properties to small (ie. not large). */
		tmpHdr.properties &= ~(P_LARGEOBJ|P_HEADERONPAGE);

		/* size of the old data */
		opInfo.size = (int)sizeof(OBJHDR) + origAlignLength;

	} else {

		origAlignLength = ALIGNSIZE((int)sizeof(SHORTPID));

		/* 
		 *	properties to small (ie. not large).
		 */
		tmpHdr.properties &= ~(P_LARGEOBJ);

		/* size of the old data */
		opInfo.size = (int)sizeof(OBJHDR) + origAlignLength;
	}

	/* log it */
	opInfo.action = LOG_ACTION_CONVERT_TO_SMALL;
	opInfo.flags =0 ;
	opInfo.oid = *oid;
	opInfo.start = 0;
	opInfo.newData = (char*) &tmpHdr;
	opInfo.oldData = (char*) &(objectPtr->objectHdr);
	opInfo.actionPid.volid = oid->diskAddr.volid;
	opInfo.actionPid.page  = oid->diskAddr.page;
	/* peform the logging as a generic header update */
	if(clog_LogConvertToSmall(&opInfo)) {
		FAIL
	}

	/*
	 *	Install the new object header.
	 *	This cannot be done earlier because log record generation
	 *	requires that the old header contiguous with the old data.
	 */
	objectPtr->objectHdr = tmpHdr;

    /*  Update the page info */
    homePage->header.freeBytes += (origAlignLength - newAlignLength);
    homePage->header.largeCount--;

Success:
    return esmNOERROR;

FailureReturn:
    return esmFAILURE;
}



 int
sm_DestroyObject (
	int						groupIndex,
	OID						*oid 
)
{
    PAGESLOT       *slotPtr;		/* slot pointer					*/
    SLOTTEDPAGE    *homePage = NULL;		/* object's home page			*/
    OBJECT			*objectPtr;		/* ptr to generic object		*/
    register LARGEOBJHDR	*lgobjectPtr;	/* ptr to large object			*/
    SLOTINDEX				objectSlot;		/* slot of obj being destroyed	*/
    PID						homePid;		/* home page of object			*/
    PID						rootPid;		/* pid of large object root		*/
	GROUPLINK				*homeLink = NULL;	/* grouplink of home page	*/
	BUFGROUP				*bufGroup;
	int						freeSize;

	/*
	 *	The following variables deal with versioned objects.  When 
	 *	a version is deleted, it is possible that one which had a
	 *	delete pending can also be deleted, in this case the 
	 *	destroyReturned variables will be set so that at the end of 
	 *	this routine, a call an be make to delete the object.  It is 
	 *	also possible that when a version is destroyed, the version 
	 *	history graph (VHG) will be empty.  If so, destroyVHG will be 
	 *	set so that the VHG can be destroyed at the end.
	 */
	BOOL		destroyReturned = FALSE;	
		/* indicates if a destroy pending obj should be destroyed*/

	OID			destroyReturnedObj;	
		/* oid of destroy pending object to be destroyed		*/

	BOOL		destroyVHG = FALSE;
		/* indicates if the version history graph should be destroyed */

	VHGNODEPTR	vhgNodePtr;
		/* oid and node id of the VHG node for the object		*/


	TRPRINT(TR_SM, TR_LEVEL_1, ("oid page:%d slot:%d unique:%d",
			oid->diskAddr.page, oid->diskAddr.slot, oid->diskAddr.unique));

	INIT_SMSTATE(&SM_State);

	bufGroup = &BufGroupPtr[groupIndex];
	if (CHECK_BUFGROUP_VALID(bufGroup))	{
		SM_ERROR(TYPE_USER, esmBADGROUP);
		return esmFAILURE;
	}

	if (checkTransState(oid->diskAddr.volid))	{
		FAIL
	}

	/*
	 *	Get the page out of the object id
	 */
    homePid.volid = oid->diskAddr.volid;
    homePid.page  = oid->diskAddr.page;


    /*
     * Read in the page containing the object's header and
     * determine whether we are dealing with a large or small object.
     */
    if ((homeLink = bf_ReadPage(bufGroup, &homePid, H_SLOTTED,
								SLOTTED_PAGE2SIZE, EX)) == NULL)	{
		FAIL
	}

	if (SM_SetLogServerState(homeLink, oid) != esmNOERROR) {
		FAIL
	}

	homePage = (SLOTTEDPAGE *) homeLink->bufFrame;
	if (CHECK_SLOTTED_MAGIC(homePage))	{
		SM_ERROR(TYPE_FATAL, esmBADSLOTTEDMAGIC); /* internal error */
	}

    objectSlot = oid->diskAddr.slot;
	slotPtr = GETSLOTPTR(homePage, objectSlot);

	/* Check for a valid oid */
    if (CHECKOID(objectSlot, homePage, slotPtr, oid->diskAddr.unique)) {
		SM_ERROR(TYPE_USER, esmBADOID);
		FAIL
	}

	/* Release any other fixed copies of the object */
	SM_ReleaseList(oid);

    objectPtr = GETOBJECTPTR(homePage, slotPtr);
	sm_SetAbortIfError();

	/* Determine if we are dealing with a large or small object */
    if (objectPtr->objectHdr.properties & P_LARGEOBJ) {
		VERSIONGRAPH	*versionGraph = NULL;
		BOOL			destroyPending;		
							/* indicates if destruction was delayed	*/

		TRPRINT(TR_SM, TR_LEVEL_2, ("large object"));

		if (objectPtr->objectHdr.properties & P_VERSIONED) {
			TRPRINT(TR_SM, TR_LEVEL_2, ("getting the VHG"));

			/* Get the VHG oid */
			if (lg_GetVHGPtr(bufGroup, 
							 (LARGEOBJHDR*)&objectPtr->objectHdr,
							 oid->diskAddr.volid, &vhgNodePtr)) {
				FAIL
			}
				
			/* Get the version history graph */
			if (SM_GetVersionGraph(bufGroup, &vhgNodePtr.oid, 
									&versionGraph, FALSE, EX)) {
				FAIL
			}
		}

		/* 
		 *	Check to see if this is a slotted root
		 */
		if (objectPtr->objectHdr.properties & P_HEADERONPAGE)	{

			TRPRINT(TR_SM, TR_LEVEL_2, ("slotted root"));

			/* initialize the rootpid */
			rootPid = homePid;

			if (objectPtr->objectHdr.properties & P_VERSIONED) {
				/* Destroy the version */
				if (lg_DestroyVersion( bufGroup, &rootPid, objectSlot, 
					  FALSE, 
					  WORKING_VERSION(objectPtr->objectHdr.properties),
					  versionGraph, &destroyPending, &destroyReturned, 
					  &destroyReturnedObj, &destroyVHG)) {

					FAIL
				}

				/* determine the size to free */
				if (destroyPending) {
					/* Object not destroyed, so don't free any space */
					freeSize = 0; 
	
					/*
					 *	The object is now waiting for all its working 
					 *	version to be frozen before it can be destroyed.
					 *	Record this fact.
					 */
					objectPtr->objectHdr.properties |= P_DESTROYPENDING;

				} else {

					/* Object was destroyed */
					freeSize = (int)sizeof(LGSLOTROOT);
				}
			} else {
				/* Destroy the object */
				if (lg_DestroyObject( &rootPid, bufGroup, objectSlot, FALSE, FALSE) ) {

					FAIL
				}
				/* determine the size to free */
				freeSize = (int)sizeof(LGSLOTROOT);
			}
		} else {

			TRPRINT(TR_SM, TR_LEVEL_2, ("page root"));

			/*
			 * Dealing with a large object.
			 * Dig out the root pid, then use large object destroy.
			 */
			rootPid.volid = homePid.volid;
			lgobjectPtr    = (LARGEOBJHDR *) objectPtr;
			rootPid.page  = lgobjectPtr->rootPid;

			if (objectPtr->objectHdr.properties & P_VERSIONED) {
				if (lg_DestroyVersion( bufGroup, &rootPid, NIL, FALSE, 
					  WORKING_VERSION(objectPtr->objectHdr.properties),
					  versionGraph, &destroyPending, &destroyReturned, 
					  &destroyReturnedObj, &destroyVHG)) {
					FAIL
				}

				/* determine the size to free */
				if (destroyPending) {
					/* Object not destroyed, so don't free any space */
					freeSize = 0;
				} else {
					/* Object was destroyed */
					freeSize = (int)sizeof(SHORTPID);
				}
			} else {

				if (lg_DestroyObject( &rootPid, bufGroup, NIL, FALSE, 
									  WORKING_VERSION(objectPtr->objectHdr.properties))	) {
					FAIL
				}
				/* determine the size to free */
				freeSize = (int)sizeof(SHORTPID);
			}
		}

		/* decrement the count of the large objects on the page */
		if (freeSize > 0) {
			homePage->header.largeCount--;
		}

		if (objectPtr->objectHdr.properties & P_VERSIONED) {

			TRPRINT(TR_SM, TR_LEVEL_2, ("setting the VHG"));

			/* Set the version history graph */
			if (SM_SetVersionGraph(bufGroup, &vhgNodePtr.oid, 
									versionGraph)) {
				FAIL
			}
		}

    } else if (objectPtr->objectHdr.properties & P_MOVED)	{
		/* Free up the forwarded portion */
		if (SM_DestroyForward(bufGroup, (OID *) objectPtr->data))	{
			FAIL
		}

		/* determine the size to free */
		freeSize = (int)sizeof(OID);
    } else {
		/* We are freeing up the object */
		freeSize = ALIGNSIZE(objectPtr->objectHdr.size);
	}

	/*
	 *	free up the slot, if something was destroyed
	 */
	if (freeSize > 0) {
		if (GET_LOGLEVEL(LOG_SPACE)) {
			OPERATIONINFO		    opInfo;
			/* fill in the log info*/
			opInfo.flags = 0;
			opInfo.oid = *oid;
			opInfo.start = 0;
			opInfo.size = freeSize;
			opInfo.oldData = (char *) objectPtr;
			opInfo.newData = NULL;
			opInfo.actionPid.volid = oid->diskAddr.volid;
			opInfo.actionPid.page  = oid->diskAddr.page;
			opInfo.currentLRC = homePage->header.lrc;

			if (GET_LOGLEVEL(LOG_DATA) || objectPtr->objectHdr.properties == NOFLAGS) {

				/* 
				 *  Log everything if LOG_DATA or if object is not 
				 *	a regular small object 
				 */
				opInfo.action = LOG_ACTION_DESTROY_SMALL_OBJ;
				/* Log the destroy operation */
				if(clog_LogDestroyObj(&opInfo)) {
					FAIL
				}
			} else {
				opInfo.action = LOG_ACTION_DESTROY_SMALL_OBJ_ND;
				/* Log the destroy operation */
				if(clog_LogDestroyObjND(&opInfo)) {
					FAIL
				}
			}
		}
		SM_ReturnSlot(homeLink, objectSlot, freeSize);
		homeLink = NULL; /* unfixed by SM_ReturnSlot */
		homePage = NULL; /* audited by SM_ReturnSlot */
	} else {
		/*
		 *	the object was only marked destroy pending, so just 
		 *	unfix and dirty the page
		 */

		bf_UnfixPage(homeLink, BF_DEFAULT, TRUE);
		homeLink = NULL;
	}

	/*
	 *	See if any pending destruction can be performed
	 */
	if (destroyReturned) {

		TRPRINT(TR_SM, TR_LEVEL_2, ("destroying pending"));

		/*
		 *	The original object destroyed, must have been a working
		 *	version.
		 */
		SM_ASSERT(LEVEL_3, WORKING_VERSION(objectPtr->objectHdr.properties));

		if (sm_DestroyObject(groupIndex, &destroyReturnedObj)) {
			FAIL
		}
	}

	/*
	 *	See if a version history graph needs to be destroyed
	 */
	if (destroyVHG) {
		TRPRINT(TR_SM, TR_LEVEL_2, ("destroying VHG"));
		
		/* destroy the version history graph */
		if (sm_DestroyObject(groupIndex, &vhgNodePtr.oid)) {
			FAIL
		}
	}

Success:
    return esmNOERROR;

FailureReturn:
	if (homeLink != NULL) {
		bf_UnfixPage(homeLink, BF_DEFAULT, FALSE);
	}
	SM_AbortIfRequired();
	return esmFAILURE;
}

 int
sm_InsertInObject (
	int						groupIndex,
	OID						*oid,
	int						start,
	int						length,
	void					*dataPtr
)
{
	char			*data = (char*)dataPtr;	/* rename dataPtr		*/
    SLOTINDEX		objectSlot;		/* the object's slot			*/
    SLOTTEDPAGE		*homePage = NULL;	/* object's home page			*/
    int				objectSize;		/* the object's data size		*/
	int				origAlignLength;
    int				newSize;		/* size of the object after 	*/
	int				newAlignLength;
    SMALLOBJ		*objectPtr;		/* ptr to generic object		*/
    PID				homePid;		/* home page of object			*/
	PAGESLOT		*slotPtr;
	GROUPLINK		*homeLink = NULL;
	BOOL			homeLinkIsDirty = FALSE;
	BUFGROUP		*bufGroup;
	LIST			*suspendList = NULL;
	OPERATIONINFO   opInfo;


	TRPRINT(TR_SM, TR_LEVEL_1, ("start:%d length:%d", start, length));

	INIT_SMSTATE(&SM_State);

	bufGroup = &(BufGroupPtr[groupIndex]);
	if (CHECK_BUFGROUP_VALID(bufGroup))	{
		SM_ERROR(TYPE_USER, esmBADGROUP);
		return esmFAILURE;
	}

	if (checkTransState(oid->diskAddr.volid))	{
		FAIL
	}

    homePid.volid = oid->diskAddr.volid;
    homePid.page  = oid->diskAddr.page;

    if ((homeLink = bf_ReadPage(bufGroup, &homePid, H_SLOTTED,
								SLOTTED_PAGE2SIZE, EX)) == NULL)	{
		FAIL
	}

	if (SM_SetLogServerState(homeLink, oid) != esmNOERROR) {
		FAIL
	}

	homePage = (SLOTTEDPAGE *) homeLink->bufFrame;
	if (CHECK_SLOTTED_MAGIC(homePage))	{
		SM_ERROR(TYPE_FATAL, esmBADSLOTTEDMAGIC);
	}
	sm_AuditSlotPage(homePage);

	objectSlot = oid->diskAddr.slot;
	slotPtr = GETSLOTPTR(homePage, objectSlot);

    if (CHECKOID(objectSlot, homePage, slotPtr, oid->diskAddr.unique))	{
		SM_ERROR(TYPE_USER, esmBADOID);
		FAIL
	}
	TRPRINT(TR_SM, TR_LEVEL_2, ("valid oid"));

    /*
     *  Get a pointer to the existing object and make sure
     *  the insertion point is valid
     */
    objectPtr = (SMALLOBJ *) GETOBJECTPTR(homePage, slotPtr);
    if (start > objectPtr->objectHdr.size) {
        SM_ERROR(TYPE_USER, esmBADPARAMS);
		FAIL
    }
    /*
     * Make sure object is not both versioned and frozen
     */
    if ( (objectPtr->objectHdr.properties & (P_VERSIONED | P_FROZEN)) ==
         (P_VERSIONED | P_FROZEN) ) {
        SM_ERROR(TYPE_USER, esmFROZEN);
		FAIL
    }
	if (length == 0) {
		goto Success;
	}
	/*
	 *	suspend any pointers if necessary
	 */
	suspendList = SM_SuspendDesc(oid, NULL);
	sm_AuditSlotPage(homePage);

    /*
     * Determine whether we are dealing with a large or small object.
     */
    if (objectPtr->objectHdr.properties & P_LARGEOBJ) {

		/* Verify that there is enough space to complete operation */
		if (bf_CheckForSpace(bufGroup, length, TRUE) != esmNOERROR) {
			FAIL
		}

		if (start < objectPtr->objectHdr.size) {
			sm_AuditSlotPage(homePage);
			/*
			 *	Do the insert and update the sizes
			 */
			if (SM_InsertInLarge( &(homePage->header.fid), bufGroup, oid, &homePid,
				homeLink, (LARGEOBJHDR *) objectPtr, start, length, data,
																objectSlot)){

				FAIL
			}
		} else {
			/*
			 *	We can get away with just an append
			 */
			if (SM_AppendToLarge( &(homePage->header.fid), bufGroup, oid,
				  homeLink, (LARGEOBJHDR *) objectPtr, length, data, objectSlot))	{
				FAIL
			}
		}
		homeLinkIsDirty = TRUE;
		goto Success;

    } else if (objectPtr->objectHdr.properties & P_MOVED)	{
		sm_AuditSlotPage(homePage);

		/* Verify that there is enough space to complete operation */
		if (bf_CheckForSpace(bufGroup, length, TRUE) != esmNOERROR) {
			FAIL
		}

		/*
		 *	we must read deal with a forwarded object
		 */
		if (SM_InsertForward(bufGroup, homeLink, oid, start, length, data)){
			FAIL
		}
		homeLinkIsDirty = TRUE;
		sm_SetAbortIfError();
		goto Success;
	} else if (objectPtr->objectHdr.properties & P_FORWARD)	{
		SM_ERROR(TYPE_FATAL, esmBADOBJECTFLAGS); /* internal error */
	}

	/*
	 *	Calculate its total final size
	 */
    newSize = (int)objectPtr->objectHdr.size + length;
	TRPRINT(TR_SM, TR_LEVEL_2, ("properties:%x newSize:%d",
			objectPtr->objectHdr.properties, newSize));

	/*
	 *	calculate original size
	 */
	origAlignLength = ALIGNSIZE((int)objectPtr->objectHdr.size);
	TRPRINT(TR_SM, TR_LEVEL_2, ("origAlignLength:%d", origAlignLength));

	/*
	 *	round the size up to the nearest alignment value
	 */
	newAlignLength = ALIGNSIZE(newSize);
	TRPRINT(TR_SM, TR_LEVEL_2, ("newAlignLength:%d", newAlignLength));

	/*
	 *	check to see if the object can fit on a small page
	 */
	if (newSize > LARGETHRESHOLD)	{
		sm_AuditSlotPage(homePage);

		/* Verify that there is enough space to complete operation */
		if (bf_CheckForSpace(bufGroup, length, TRUE) != esmNOERROR) {
			FAIL
		}

		if (SM_ConvertInsert(bufGroup, oid, &homePid, homeLink, start,
			  		(int)objectPtr->objectHdr.size, length, 
					objectPtr->data, data))	{
			FAIL
		}
		homeLinkIsDirty = TRUE;
		/* sm_SetAbortIfError(): not necessary
		 * because we're done 
		 */
		goto Success;
	}

    /*
     * Dealing with a small object.
     * First set up fields for logging,
     * then make sure there is enough free space for the new data.
     */
    opInfo.flags = 0;
    opInfo.oid = *oid;
    opInfo.start = start;
    opInfo.size = length;
    opInfo.oldData = NULL;
    opInfo.actionPid.volid = oid->diskAddr.volid;
    opInfo.actionPid.page  = oid->diskAddr.page;
    opInfo.currentLRC = homePage->header.lrc;

    switch (SM_CheckFree(homePage, homeLink, objectSlot, objectPtr, length)){

		case PAGE_NO_SPACE:

			TRPRINT(TR_SM, TR_LEVEL_2, ("PAGE_NO_SPACE"));

			/*
			 *	check to see if there is room to store the forwarding oid
			 */
			if (objectPtr->objectHdr.size < (int)sizeof(OID))	{

				/*
				 *	convert that object to a large object
				 */
				if (SM_ConvertInsert(bufGroup, oid, &homePid, homeLink, start,
						  		     (int)objectPtr->objectHdr.size, length,
								  	 objectPtr->data, data))	{
					FAIL
				}
				homeLinkIsDirty = TRUE;
				sm_SetAbortIfError();

			} else {
				OID			newOid;

				objectSize = (int)objectPtr->objectHdr.size;

				/*
				 *	attempt to forward the object to the neighbor page
				 */
				/* SM_CreateInsert does sm_SetAbortIfError(); */
				if (SM_CreateInsert(bufGroup, &(homePage->header.fid), oid,
						length + objectSize, homeLink, start, objectPtr->data,
						data, objectSize, &newOid, (OBJHDR *) objectPtr))	{
					FAIL
				}
				homeLinkIsDirty = TRUE;

				if (GET_LOGLEVEL(LOG_SPACE)) {
					/* delete the data portion of the original object 
						except for enough space to store the OID of the
						forwarded part of the object*/
					opInfo.action = LOG_ACTION_DELETE_SMALL_OBJ;
					opInfo.flags = 0;
					opInfo.oid = *oid;
					opInfo.start = (int)sizeof(OID);
					opInfo.size = (int)objectPtr->objectHdr.size - (int)sizeof(OID);
					opInfo.newData = NULL;
					opInfo.currentLRC = homePage->header.lrc;
					
					if (GET_LOGLEVEL(LOG_DATA) ) {
						opInfo.action = LOG_ACTION_DELETE_SMALL_OBJ;
						opInfo.oldData = (char *) (objectPtr->data) + (int)sizeof(OID);
						/* log as a regular delete */
						if(clog_LogModifyObj(&opInfo)) {
							FAIL
						}
					} else {
						opInfo.action = LOG_ACTION_DELETE_SMALL_OBJ_ND;
						opInfo.oldData = NULL;
						/* log as a regular delete */
						if(clog_LogModifyObjND(&opInfo)) {
							FAIL
						}
					}
				}
				objectPtr->objectHdr.size = (int)sizeof(OID);

				/* free up space on the page */
				homePage->header.freeBytes += 
							(origAlignLength - ALIGNSIZE((int)sizeof(OID)));
				 
				/* copy the old object header and the bytes of data that
				   will be overwritten by the OID */
			    bcopy((char *) objectPtr,OldData,
				   (int) (ALIGNSIZE((int)sizeof(OBJHDR)) + ALIGNSIZE(sizeof(OID))));

				/*
				 *	update the contents of this record
				 */
				objectPtr->objectHdr.properties |= P_MOVED;

				objectPtr->objectHdr.size = newSize;

				(*((OID *) objectPtr->data)) = newOid;

				/* Log the forwarding info for the home (original) object */
				if (GET_LOGLEVEL(LOG_SPACE)) {
					opInfo.action = LOG_ACTION_FORWARD_SMALL_OBJ;
					opInfo.flags = 0;
					opInfo.oid = *oid;
					opInfo.start = 0;
					opInfo.size = 0;
					opInfo.oldData = (char *) OldData;
					opInfo.newData = (char *) objectPtr;

					/* write the log record*/
					if(clog_LogForwardObj(&opInfo)) {
						FAIL
					}

				}
			}
			goto Success;

		case PAGE_LAST_SLOT:	{

			TRPRINT(TR_SM, TR_LEVEL_2, ("PAGE_LAST_SLOT"));

			/*
			 *	Reestablish pointer as it may have moved
			 */
			objectPtr = (SMALLOBJ *) GETOBJECTPTR(homePage, slotPtr);

			/* copy data after insertion point */
			bcopy(objectPtr->data + start,
				  objectPtr->data + (start + length),
				  (int) (objectPtr->objectHdr.size - start));

			/* now copy in the data if necessary */
			if (data != NULL)	{
				bcopy(data, objectPtr->data + start, (int) length);
			} else {
				bzero(objectPtr->data + start, (int) length);
			}

			homePage->header.freeBytes -= (newAlignLength - origAlignLength);
			homePage->header.freeStart += (newAlignLength - origAlignLength);

			/* 
			 * Update the object's header, and adjust the 
			 * user pointers into the object.
			 */
			objectPtr->objectHdr.size = newSize;
	
			if (GET_LOGLEVEL(LOG_SPACE)) {
				if (GET_LOGLEVEL(LOG_DATA) && (data != NULL)) {
					opInfo.action = LOG_ACTION_INSERT_SMALL_OBJ;
					opInfo.newData = data;
					/* Log the insert operation */
					if(clog_LogModifyObj(&opInfo)) {
						FAIL
					}
				} else {
					opInfo.action = LOG_ACTION_INSERT_SMALL_OBJ_ND;
					opInfo.newData = NULL;
					/* Log the insert operation */
					if(clog_LogModifyObjND(&opInfo)) {
						FAIL
					}
				}
			}
			break;
		}

		
		case PAGE_INT_SLOT:	{

			char		*source;
			char		*dest;
			int			bytesToMove;
			char		*newLocation;

			TRPRINT(TR_SM, TR_LEVEL_2, ("PAGE_INT_SLOT"));

			/*
			 *	Reestablish pointer as it may have moved
			 */
			objectPtr = (SMALLOBJ *) GETOBJECTPTR(homePage, slotPtr);

			newLocation = ((char *) homePage) + homePage->header.freeStart;

			/*
			 *	copy the exisiting data up to the insertion point
			 *	to the end of the page
			 */
			bytesToMove = (int)sizeof(OBJHDR) + start;
			bcopy((char *) objectPtr, newLocation, (int) bytesToMove);
			TRPRINT(TR_SM, TR_LEVEL_2, ("new offset:%d bytesToMove:%d",
					homePage->header.freeStart, bytesToMove));

			/*
			 *	create the hole in the data
			 */
			source		= ((char *) objectPtr->data) + start;
			dest		= newLocation + bytesToMove + length;
			bytesToMove = (int)objectPtr->objectHdr.size - start;
			bcopy(source, dest, (int) bytesToMove);
			TRPRINT(TR_SM, TR_LEVEL_2, ("bytesToMove:%d", bytesToMove));

#if OBJECT_MAGIC IS_ENABLED
			/*
			 *	Null out the old object magic
			 */
			objectPtr->objectHdr.magic = 0;

#endif

			objectPtr = (SMALLOBJ *) newLocation;

			/*
			 *	copy in the user data if necessary
			 */
			dest = ((char *) objectPtr->data) + start;
			if (data != NULL)	{
				bcopy(data, dest, (int) length);
			} else {
				bzero(dest, (int) length);
			}

			slotPtr->offset = homePage->header.freeStart;
			SM_ASSERT(LEVEL_1, slotPtr->offset % ALIGNON == 0);

			homePage->header.freeStart += (int)sizeof(OBJHDR) + newAlignLength;
			homePage->header.freeBytes -= (newAlignLength - origAlignLength);

			TRPRINT(TR_SM, TR_LEVEL_2, ("freeBytes:%d freeStart:%d",
					homePage->header.freeBytes, homePage->header.freeStart));

			/* 
			 * Update the object's header, and adjust the 
			 * user pointers into the object.
			 */
			objectPtr->objectHdr.size = newSize;

			if (GET_LOGLEVEL(LOG_SPACE)) {
				if (GET_LOGLEVEL(LOG_DATA) && (data != NULL)) {
					opInfo.action = LOG_ACTION_INSERT_SMALL_OBJ;
					opInfo.newData = data;
					/* Log the insert operation */
					if(clog_LogModifyObj(&opInfo)) {
						FAIL
					}
				} else {
					opInfo.action = LOG_ACTION_INSERT_SMALL_OBJ_ND;
					opInfo.newData = NULL;
					/* Log the insert operation */
					if(clog_LogModifyObjND(&opInfo)) {
						FAIL
					}
				}
			}
			break;
		}
		
		default:
			SM_ERROR(TYPE_FATAL, esmINTERNAL);
			break;
    }

Success:
	if(homeLink != NULL) {
		bf_UnfixPage(homeLink, BF_DEFAULT, TRUE);
	}

	if(homePage) {
		sm_AuditSlotPage(homePage);
	}
	if (suspendList != NULL)	{
		SM_RestoreDesc(suspendList);
		suspendList = NULL;
	}
    return esmNOERROR ;

FailureReturn:
	if (homeLink != NULL) {
		bf_UnfixPage(homeLink, BF_DEFAULT, homeLinkIsDirty);
	}
	if (suspendList != NULL)	{
		SM_RestoreDesc(suspendList);
		suspendList = NULL;
	}
	SM_AbortIfRequired();
	return esmFAILURE;
}

 int
sm_LockObject (
	int			groupIndex,
	OID			*oid,
	LOCKMODE	lockMode
)
{
	BUFGROUP	*bufGroup;
	GROUPLINK	*groupLink = NULL;
	PID			pid;

    TRPRINT(TR_SM, TR_LEVEL_1, ("oid:V%d:P%d:S%d:U%d",
			oid->diskAddr.volid, oid->diskAddr.page, oid->diskAddr.slot,
			oid->diskAddr.unique));

	INIT_SMSTATE(&SM_State);

	bufGroup = &(BufGroupPtr[groupIndex]);
	if (CHECK_BUFGROUP_VALID(bufGroup))	{
		SM_ERROR(TYPE_USER, esmBADGROUP);
		return esmFAILURE;
	}

	if (checkTransState(oid->diskAddr.volid))	{
		FAIL
	}

	switch (lockMode) {
	case NL:
		/* no lock needed */
		return esmNOERROR;
	case SH: case EX:
		break;
	default:
		SM_ERROR(TYPE_USER, esmBADLOCKMODE);
		return esmFAILURE;
	}

	pid.volid = oid->diskAddr.volid;
	pid.page  = oid->diskAddr.page;

	if ((groupLink = bf_ReadPage(bufGroup, &pid, H_SLOTTED, 
		SLOTTED_PAGE2SIZE, lockMode)) == NULL) {
		FAIL
	}
Success:
	bf_UnfixPage(groupLink, BF_DEFAULT, FALSE);
    return esmNOERROR;

FailureReturn:
	if (groupLink != NULL) {
		bf_UnfixPage(groupLink, BF_DEFAULT, FALSE);
	}
	SM_AbortIfRequired();
	return esmFAILURE;
}

 int
sm_ReadObjectHeader (
	int						groupIndex,
	OID						*oid,
	OBJHDR					*objectHdr 
)
{
    SLOTTEDPAGE    		*homePage;		/* object's home page		*/
    OBJECT				*objectPtr;		/* ptr to generic object	*/
	GROUPLINK			*homeLink = NULL;
	BUFGROUP			*bufGroup;		/* pointer to buffer group	*/
    PID					homePid;		/* home page of object		*/
	PAGESLOT			*slotPtr;
    SLOTINDEX			objectSlot;		/* the object's slot		*/


	TRPRINT(TR_SM, TR_LEVEL_1, ("oid page:%d slot:%d unique:%d",
			oid->diskAddr.page, oid->diskAddr.slot, oid->diskAddr.unique));

	bufGroup = &(BufGroupPtr[groupIndex]);
	if (CHECK_BUFGROUP_VALID(bufGroup))	{
		SM_ERROR(TYPE_USER, esmBADGROUP);
		return esmFAILURE;
	}

	if (checkTransState(oid->diskAddr.volid))	{
		FAIL
	}

    homePid.volid = oid->diskAddr.volid;
    homePid.page = oid->diskAddr.page;

	if ((homeLink = bf_ReadPage(bufGroup, &homePid, H_SLOTTED,
								SLOTTED_PAGE2SIZE, SH)) == NULL)	{
		FAIL
	}

	homePage = (SLOTTEDPAGE *) homeLink->bufFrame;
	if (CHECK_SLOTTED_MAGIC(homePage))	{
		SM_ERROR(TYPE_FATAL, esmBADSLOTTEDMAGIC);
	}
    objectSlot = oid->diskAddr.slot;
	slotPtr = GETSLOTPTR(homePage, objectSlot);

	if (CHECKOID(objectSlot, homePage, slotPtr, oid->diskAddr.unique))	{
		SM_ERROR(TYPE_USER, esmBADOID);
		FAIL
	}
    objectPtr = GETOBJECTPTR(homePage, slotPtr);
    *objectHdr = objectPtr->objectHdr;

Success:
    bf_UnfixPage(homeLink, BF_DEFAULT, FALSE);
    return esmNOERROR;

FailureReturn:
	if (homeLink != NULL) {
		bf_UnfixPage(homeLink, BF_DEFAULT, FALSE);
	}
	SM_AbortIfRequired();
	return esmFAILURE;
}


 int
sm_OidToFid (
	int						groupIndex,
	OID						*oid,
	FID						*fid 
)
{
    SLOTTEDPAGE    	*homePage;		/* object's home page		*/
	GROUPLINK		*homeLink = NULL;
	BUFGROUP		*bufGroup;		/* pointer to buffer group	*/
    PID				homePid;		/* home page of object		*/

	TRPRINT(TR_SM, TR_LEVEL_1, ("oid page:%d slot:%d unique:%d",
			oid->diskAddr.page, oid->diskAddr.slot, oid->diskAddr.unique));

	INIT_SMSTATE(&SM_State);

	bufGroup = &(BufGroupPtr[groupIndex]);
	if (CHECK_BUFGROUP_VALID(bufGroup))	{
		SM_ERROR(TYPE_USER, esmBADGROUP);
		return esmFAILURE;
	}

	if (checkTransState(oid->diskAddr.volid))	{
		FAIL
	}

    /* Construct the pid of the home page */
    homePid.volid = oid->diskAddr.volid;
    homePid.page = oid->diskAddr.page;

	if ((homeLink = bf_ReadPage(bufGroup, &homePid, H_SLOTTED,
								SLOTTED_PAGE2SIZE, SH)) == NULL)	{
		FAIL
	}

	homePage = (SLOTTEDPAGE *) homeLink->bufFrame;
	if (CHECK_SLOTTED_MAGIC(homePage))	{
		SM_ERROR(TYPE_FATAL, esmBADSLOTTEDMAGIC);
	}

	*fid = homePage->header.fid;

Success:
    bf_UnfixPage(homeLink, BF_DEFAULT, FALSE);
    return esmNOERROR;

FailureReturn:
	if (homeLink != NULL) {
		bf_UnfixPage(homeLink, BF_DEFAULT, FALSE);
	}
	SM_AbortIfRequired();
	return esmFAILURE;
}

 int
sm_SetObjectHeader (
	int						groupIndex,
	OID						*oid,
	OBJHDR					*objectHdr 
)
{
    register SLOTTEDPAGE    *homePage;	/* object's home page		*/
    register OBJECT			*objectPtr;	/* ptr to generic object	*/
	GROUPLINK				*homeLink = NULL;
	BUFGROUP				*bufGroup;	/* pointer to buffer group	*/
    PID						homePid;	/* home page of object		*/
	PAGESLOT				*slotPtr;
    SLOTINDEX				objectSlot;	/* the object's slot		*/
	TWO						oldTag;


	TRPRINT(TR_SM, TR_LEVEL_1, ("oid page:%d slot:%d unique:%d",
			 oid->diskAddr.page, oid->diskAddr.slot, oid->diskAddr.unique));

	INIT_SMSTATE(&SM_State);

	bufGroup = &(BufGroupPtr[groupIndex]);
	if (CHECK_BUFGROUP_VALID(bufGroup))	{
		SM_ERROR(TYPE_USER, esmBADGROUP);
		return esmFAILURE;
	}

	if (checkTransState(oid->diskAddr.volid))	{
		return esmFAILURE;
	}

    /* Construct the pid of the home page */
    homePid.volid = oid->diskAddr.volid;
    homePid.page = oid->diskAddr.page;

	/*
	 *	read the page into a buffer
	 */
	if ((homeLink = bf_ReadPage(bufGroup, &homePid, H_SLOTTED,
								SLOTTED_PAGE2SIZE, EX)) == NULL)	{
		FAIL
	}

	if (SM_SetLogServerState(homeLink, oid) != esmNOERROR) {
		FAIL
	}
	homePage = (SLOTTEDPAGE *) homeLink->bufFrame;

	if (CHECK_SLOTTED_MAGIC(homePage))	{
		SM_ERROR(TYPE_FATAL, esmBADSLOTTEDMAGIC);
	}

    /* Make sure we have a valid slot in the oid.  */
    objectSlot = oid->diskAddr.slot;

	/* get a pointer to the slot */
	slotPtr = GETSLOTPTR(homePage, objectSlot);

	/* check the validity of the slot & unique */
	if (CHECKOID(objectSlot, homePage, slotPtr, oid->diskAddr.unique))	{
		SM_ERROR(TYPE_USER, esmBADOID);
		FAIL
	}

    objectPtr = GETOBJECTPTR(homePage, slotPtr);

	/* copy out the header */
	oldTag = objectPtr->objectHdr.tag;
    objectPtr->objectHdr.tag = objectHdr->tag;
	/* homeLink is now dirty! */
	sm_SetAbortIfError();

	if (GET_LOGLEVEL(LOG_SPACE)) {
		if(clog_LogChangeSmTag(oid, &(objectPtr->objectHdr), 
			oldTag, objectHdr->tag) != esmNOERROR) {
			FAIL;
		}
	}

Success:
    bf_UnfixPage(homeLink, BF_DEFAULT, TRUE);
    return esmNOERROR;

FailureReturn: /* FAIL goes here */
	if(homeLink != NULL) {
		bf_UnfixPage(homeLink, BF_DEFAULT, FALSE);
	}
	SM_AbortIfRequired();
	return esmFAILURE;
}


 int
sm_WriteObject (
    int		groupIndex,	/* the buffer group to use		*/
    int		start,		/* starting offset of write		*/
    int		length,		/* length of write			*/
    void	*data,		/* data to write in			*/
    USERDESC *userDesc,   /* address of ptr to user descriptor	*/
    BOOL	release 	/* whether to release the object	*/
)
{
	BUFGROUP 	*bufGroup;
	GROUPLINK	*homeLink = NULL;

    TRPRINT(TR_SM, TR_LEVEL_1, ("start:%d length:%d", start, length));

	INIT_SMSTATE(&SM_State);

    CHECK_USERDESC(userDesc);
    if (checkTransState((userDesc)->oid.diskAddr.volid)){ 
		return esmFAILURE;							
    }

	/* This buffer group is not used; the one in the user 
	 * descriptor is the one that is used.
	 * Check it anyway.
	 */
	bufGroup = &(BufGroupPtr[groupIndex]);
	if (CHECK_BUFGROUP_VALID(bufGroup))	{
		SM_ERROR(TYPE_USER, esmBADGROUP);
		return esmFAILURE;
	}

    /*
     * Make sure object is not both versioned and frozen
     */
    if ( (userDesc->userFlags & (P_VERSIONED | P_FROZEN)) == 
		 (P_VERSIONED | P_FROZEN) ) {
		SM_ERROR(TYPE_USER, esmFROZEN);
		return esmFAILURE;
    }
	if (length == 0) {
		/* no work to do  */
		return esmNOERROR;
	}
    /*
     *	check arguments given by user
	 *  data==NULL means to zero-fill the object.
     */
    if ( (start < 0) || (length < 0) || (start + length > userDesc->byteCount))	{
		SM_ERROR(TYPE_USER, esmBADPARAMS);
		return esmFAILURE;
    }

    /*
     * are we dealing with a large or a small object?
     */
    if (userDesc->type == TYPE_CHUNK) {
		/* write is to a large object */
		PID		homePid;

		homePid.volid = userDesc->oid.diskAddr.volid;
		homePid.page  = userDesc->oid.diskAddr.page;

		/*
		 *	Read the home page from disk
		 */
		if ((homeLink = bf_ReadPage(userDesc->bufGroup, &homePid, H_SLOTTED,
						SLOTTED_PAGE2SIZE, EX)) == NULL) {
			FAIL
		}
		if (SM_SetLogServerState(homeLink, &(userDesc->oid)) != esmNOERROR) {
			FAIL
		}
		/*
		 *  Write it. This will also log the write.
		 *  Do not release it here; we'll do that later.
		 */
		if (lg_WriteObject((CHUNKDESC *) userDesc->objInfo, userDesc->basePtr,
									start, length, FALSE, (char*)data))	{
			FAIL
		}

		/* homeLink is not dirty because only the
		 * object header is on that page, and the
		 * header doesn't change (the obj doesn't change
		 * size with a write)
		 */
		goto Success;

    } else if (userDesc->type == TYPE_PAGE)	{
		register PAGEHASH	*pageHash;

		if (SM_SetLogServerState(userDesc->groupLink, &(userDesc->oid)) != esmNOERROR) {
			FAIL
		}
		pageHash = userDesc->groupLink->pageHash;
		CHECK_PAGEHASH_MAGIC(pageHash);

		/*
		 *	see if the page is locked in IX mode
		 */
		if (pageHash->writeLock != TransRec.clientTid)	{
			if (lm_LockPage( &(pageHash->hashList.hashPid), pageHash, EX, 
							&(userDesc->fid) )) {
				FAIL
			} else {
				pageHash->writeLock = TransRec.clientTid;
			}
		}

		/* log the write operation if logging is on */

		if (GET_LOGLEVEL(LOG_DATA)) {
			OPERATIONINFO opInfo;
			opInfo.flags = 0;

			/* if this is a forwarded object, log the oid of the
			   forwarded part, NOT the home part */

			if (userDesc->userFlags & P_MOVED) {
				opInfo.oid = userDesc->forwardOid;
			} else {
				opInfo.oid = userDesc->oid;
			}	
			opInfo.start = userDesc->offset + start;
			opInfo.size = length;
			opInfo.oldData = userDesc->basePtr + start;
			opInfo.actionPid.volid = opInfo.oid.diskAddr.volid;
			opInfo.actionPid.page  = opInfo.oid.diskAddr.page;
			opInfo.currentLRC = ((SLOTTEDPAGE *) pageHash->pageInfo->bufFrame)->header.lrc;

			if (data == NULL) {
				opInfo.action = LOG_ACTION_WRITE_SMALL_OBJ_ND; 
				opInfo.newData = NULL;
				if(clog_LogWriteObjND(&opInfo)) {
					FAIL
				}
			} else {
				opInfo.action = LOG_ACTION_WRITE_SMALL_OBJ; 
				opInfo.newData = (char*)data;
				if(clog_LogWriteObj(&opInfo)) {
					FAIL
				}
			}
		}
		sm_SetAbortIfError();

		/* copy in the new data */
		if (data == NULL) {
			bzero(userDesc->basePtr + start, (int) length);
		} else {
			bcopy((char*)data, userDesc->basePtr + start, (int) length);
		}	

		bf_DirtyPage(userDesc->groupLink->pageHash);
    } else {
		/* Return an error to the caller */
		SM_ERROR(TYPE_USER, esmBADUSERDESC);
		return esmFAILURE;
    }
Success:
	if (release) {
		if( SM_ReleaseObject(userDesc) ) {
			FAIL;
		}
	}

	if(homeLink!= NULL) {
		bf_UnfixPage(homeLink, BF_DEFAULT, FALSE);
	}
    return esmNOERROR;

FailureReturn:
	if(homeLink != NULL) {
		bf_UnfixPage(homeLink, BF_DEFAULT, FALSE);
	}
	SM_AbortIfRequired();
	return esmFAILURE;
}

 int
sm_WroteObject (
    int		groupIndex,	/* the buffer group to use				*/
    int		start,		/* starting offset of write				*/
    int		length,		/* length of write						*/
    void	*data,		/* old value of the data written		*/
    USERDESC *userDesc,   /* ptr to user descriptor				*/
    BOOL	release 	/* whether to release the object		*/
)
{
	BUFGROUP 	*bufGroup;
	GROUPLINK	*homeLink = NULL;

    TRPRINT(TR_SM, TR_LEVEL_1, ("start:%d length:%d", start, length));

	INIT_SMSTATE(&SM_State);

    CHECK_USERDESC(userDesc);
    if (checkTransState((userDesc)->oid.diskAddr.volid)){ 
		return esmFAILURE;							
    }

	/* This buffer group is not used; the one in the user 
	 * descriptor is the one that is used.
	 * Check it anyway.
	 */
	bufGroup = &(BufGroupPtr[groupIndex]);
	if (CHECK_BUFGROUP_VALID(bufGroup))	{
		SM_ERROR(TYPE_USER, esmBADGROUP);
		return esmFAILURE;
	}

    /* Make sure object is not (versioned and frozen) */
    if ( (userDesc->userFlags & (P_VERSIONED | P_FROZEN)) == 
		 (P_VERSIONED | P_FROZEN) ) {
		SM_ERROR(TYPE_USER, esmFROZEN);
		return esmFAILURE;
    }
	if (length == 0) {
		/* no work to do */
		return esmNOERROR;
	}
    if ((data == NULL) || (start < 0) || 
		(length < 0) || (start + length > userDesc->byteCount))	{
		SM_ERROR(TYPE_USER, esmBADPARAMS);
		return esmFAILURE;
    }

    /*
     * Examine the userDesc to see if we are dealing with a large or
     * small object.
     */
    if (userDesc->type == TYPE_CHUNK) {
		/* write is to a large object */

		PID		homePid;

		/* construct the pid of the slotted page */
		homePid.volid = userDesc->oid.diskAddr.volid;
		homePid.page  = userDesc->oid.diskAddr.page;

		/* Read the target page from disk */
		if ((homeLink = bf_ReadPage(userDesc->bufGroup, &homePid, H_SLOTTED,
						SLOTTED_PAGE2SIZE, EX)) == NULL) {
			FAIL
		}
		if (SM_SetLogServerState(homeLink, &(userDesc->oid)) != esmNOERROR) {
			FAIL
		}
		/*
		 *  Write it and release it if necessary.  This will also log the write.
		 */
		if (lg_WroteObject((CHUNKDESC *) userDesc->objInfo, userDesc->basePtr,
									start, length, release, (char*)data))	{
			FAIL
		}
		if (release)	{
			SM_ReleaseObject(userDesc);
		}
		bf_UnfixPage(homeLink, BF_DEFAULT, FALSE);
		homeLink = NULL;
		return esmNOERROR;

    } else if (userDesc->type == TYPE_PAGE)	{
		register PAGEHASH	*pageHash;

		if (SM_SetLogServerState(userDesc->groupLink, &(userDesc->oid)) != esmNOERROR) {
			FAIL
		}
		pageHash = userDesc->groupLink->pageHash;
		CHECK_PAGEHASH_MAGIC(pageHash);

		/*
		 *	Get an exclusive lock on the page.
		 */
		if (pageHash->writeLock != TransRec.clientTid)	{
			if (lm_LockPage( &(pageHash->hashList.hashPid), 
					pageHash, EX, &(userDesc->fid) )) {
				FAIL
			} else {
				pageHash->writeLock = TransRec.clientTid;
			}
		}
		/* 
		 * Log the write operation if logging is on 
	     */

		if (GET_LOGLEVEL(LOG_DATA)) {
			OPERATIONINFO opInfo;

			opInfo.flags = 0;
			/* if this is a forwarded object, log the oid of the
			   forwarded part, NOT the home part */

			if (userDesc->userFlags & P_MOVED) {
				opInfo.oid = userDesc->forwardOid;
			} else {
				opInfo.oid = userDesc->oid;
			}	
			opInfo.start = userDesc->offset + start;
			opInfo.size = length;
			opInfo.oldData = (char*) data;
			opInfo.actionPid.volid = opInfo.oid.diskAddr.volid;
			opInfo.actionPid.page  = opInfo.oid.diskAddr.page;
			opInfo.currentLRC = ((SLOTTEDPAGE *) pageHash->pageInfo->bufFrame)->header.lrc;

			if (data == NULL) {
				SM_ASSERT(LEVEL_3, esmNOTIMPLEMENTED);
				opInfo.action = LOG_ACTION_WRITE_SMALL_OBJ_ND; 
				opInfo.newData = NULL;
				if(clog_LogWriteObjND(&opInfo)) {
					FAIL
				}
			} else {
				opInfo.action = LOG_ACTION_WRITE_SMALL_OBJ; 
				opInfo.newData = userDesc->basePtr + start;
				if(clog_LogWriteObj(&opInfo)) {
					FAIL
				}
			}
		}

		if (release) {
			bf_UnfixPage(userDesc->groupLink, BF_DEFAULT, TRUE);
			SM_ReleaseObject(userDesc);
		} else {
			bf_DirtyPage(userDesc->groupLink->pageHash);
		}
    } else {
		/* Return an error to the caller */
		SM_ERROR(TYPE_USER, esmBADUSERDESC);
		return esmFAILURE;
    }
Success:
    return esmNOERROR;

FailureReturn:
	if(homeLink != NULL) {
		bf_UnfixPage(homeLink, BF_DEFAULT, FALSE);
	}
	SM_AbortIfRequired();
	return esmFAILURE;
}


 int
sm_DeleteFromObject (
	int						groupIndex,
	OID						*oid,
	int						start,
	int						length 
)
{
    register SLOTINDEX		objectSlot;	/* the object's slot		*/
    SLOTTEDPAGE	*homePage	= NULL;		/* object's home page		*/
    register OBJECT			*objectPtr;	/* ptr to generic object	*/
    register SMALLOBJ		*smallObject;/* pointer to the object	*/
    int						oldSize;	/* size of the object before delete	*/
    int						newSize;	/* size of the object after delete	*/
	int						origAlignLength;
	int						newAlignLength;
    int						bytesToMove;/* bytes to move for delete	*/
    char					*source;	/* the source for the byte move*/
    char					*deletePoint;/* the deletion point		*/
    PID						homePid;	/* home page of object		*/
    PID						rootPid;	/* pid of large object root	*/
	GROUPLINK				*homeLink	= NULL;
	BOOL					homeLinkIsDirty = FALSE;
	PAGESLOT				*slotPtr;
	BUFGROUP				*bufGroup;
	LIST					*suspendList	= NULL;
	OPERATIONINFO			opInfo;


	TRPRINT(TR_SM, TR_LEVEL_1, ("start:%d length:%d", start, length));

	INIT_SMSTATE(&SM_State);

	bufGroup = &(BufGroupPtr[groupIndex]);
	if (CHECK_BUFGROUP_VALID(bufGroup))	{
		SM_ERROR(TYPE_USER, esmBADGROUP);
		return esmFAILURE;
	}

	if (checkTransState(oid->diskAddr.volid))	{
		FAIL
	}

	homePid.page  = oid->diskAddr.page;
	homePid.volid = oid->diskAddr.volid;

    if ((homeLink = bf_ReadPage(bufGroup, &homePid, H_SLOTTED,
								SLOTTED_PAGE2SIZE, EX)) == NULL)	{
		FAIL
	}

	if (SM_SetLogServerState(homeLink, oid) != esmNOERROR) {
		FAIL
	}

	homePage = (SLOTTEDPAGE *) homeLink->bufFrame;

	if (CHECK_SLOTTED_MAGIC(homePage))	{
		SM_ERROR(TYPE_FATAL, esmBADSLOTTEDMAGIC); /* internal error */
	}

    objectSlot = oid->diskAddr.slot;
	slotPtr = GETSLOTPTR(homePage, objectSlot);

    if (CHECKOID(objectSlot, homePage, slotPtr, oid->diskAddr.unique))	{
		SM_ERROR(TYPE_USER, esmBADOID);
		FAIL
	}

    objectPtr = GETOBJECTPTR(homePage, slotPtr);

    if ( (objectPtr->objectHdr.properties & (P_VERSIONED | P_FROZEN)) ==
         (P_VERSIONED | P_FROZEN) ) {
        SM_ERROR(TYPE_USER, esmFROZEN);
		FAIL
    }

	if (length == 0) {
		goto Success;
	}

	/*
	 *	check to make sure the start location is valid
	 */
	if (start > objectPtr->objectHdr.size)	{
		SM_ERROR(TYPE_USER, esmBADSTART);
		FAIL
	}

    if (length < 0) {
		length = (int)objectPtr->objectHdr.size - start;
    } else {
		/*
		 *	Check to make sure the caller is not deleting
		 *	more data than the object holds
		 */
		if (start + length > objectPtr->objectHdr.size)	{
			SM_ERROR(TYPE_USER, esmBADLENGTH);
			FAIL
		}
    }

    /*
     * Calculate the data size of the object after the delete.
     */
    newSize = (int)objectPtr->objectHdr.size - length;

	suspendList = SM_SuspendDesc(oid, NULL);

    /*
     * Determine whether we are dealing with a large or small object.
     */
    if (objectPtr->objectHdr.properties & P_LARGEOBJ) {
		TRPRINT(TR_SM, TR_LEVEL_2, ("large object"));	

		/* Verify that there is enough space to complete operation */
		if (bf_CheckForSpace(bufGroup, length, TRUE) != esmNOERROR) {
			FAIL
		}

		/*
		 *	Check to see if this is a slotted root
		 */
		if (objectPtr->objectHdr.properties & P_HEADERONPAGE)	{
			LARGEOBJROOT		*largeRoot;

			TRPRINT(TR_SM, TR_LEVEL_2, ("slotted root"));

			rootPid = homePid;

			/* 
			 *	See if we should convert the obj to small or just
			 *	do the delete.  Currently, we only do the conversion
			 *	if the new length of the object will be zero
			 */
			if (length == objectPtr->objectHdr.size) {
				TRPRINT(TR_SM, TR_LEVEL_2, ("deleting all bytes"));
				if (SM_ConvertDelete(bufGroup, oid, &homePid, homeLink,
									 objectPtr, length, &rootPid,
									 objectSlot)) {
					FAIL
				}
			} else {
				/* Delete the bytes */
				if (lg_DeleteFromObject( &(homePage->header.fid),
						&rootPid, bufGroup, start, length, objectSlot,
						objectPtr->objectHdr.properties & P_VERSIONED))	{
					FAIL
				}

				/*
				 *	Update the size, log it
				 */
				oldSize = (int)objectPtr->objectHdr.size;
				objectPtr->objectHdr.size = newSize;
				if (GET_LOGLEVEL(LOG_SPACE)) {
					if(clog_LogChangeSmSize(oid, &(objectPtr->objectHdr), oldSize, newSize)) {
						FAIL
					}
				}

				largeRoot = (LARGEOBJROOT *) objectPtr;
				/* check the magic number */
				CHECK_SLOTROOT_MAGIC( &(largeRoot->rootPage) );
			}
		} else {
    		register LARGEOBJHDR	*largeHdr;

			/*
			 *	Dealing with a large object.
			 *	Dig out the root pid
			 */
			largeHdr = (LARGEOBJHDR *) objectPtr;
			rootPid.volid = homePid.volid;
			rootPid.page = largeHdr->rootPid;

			/* 
			 *	See if we should convert the obj to small for just
			 *	do the delete.  Currently, we only do the conversion
			 *	if the new length of the object will be zero
			 */
			if (length == objectPtr->objectHdr.size) {
				TRPRINT(TR_SM, TR_LEVEL_2, ("deleting all bytes"));
				if (SM_ConvertDelete(bufGroup, oid, &homePid, homeLink,
									 objectPtr, length, &rootPid, NIL)) {
					FAIL
				}
			} else {
				/* Delete the bytes */
				if (lg_DeleteFromObject( &(homePage->header.fid), &rootPid,
										 bufGroup, start, length, NIL,
										 objectPtr->objectHdr.properties & P_VERSIONED))	{
					FAIL
				}

				/*
				 *	Update the size and the root pid in the slot
				 */
				oldSize = (int)largeHdr->objectHdr.size;
				largeHdr->objectHdr.size = newSize;
				SM_ASSERT(LEVEL_3, largeHdr->rootPid == rootPid.page);
				TRPRINT(TR_SM, TR_LEVEL_2, ("root pid:%d", largeHdr->rootPid));
				if (GET_LOGLEVEL(LOG_SPACE)) {
					if(clog_LogChangeSmSize(oid, &(largeHdr->objectHdr), oldSize, newSize)) {
						FAIL
					}
				}
			}
		}
		homeLinkIsDirty = TRUE;
		goto Success;

    } else if (objectPtr->objectHdr.properties & P_MOVED)	{

		/*
		 *	delete the forward object
		 */
		if (SM_DeleteForward(bufGroup, (OID *) objectPtr->data,start,length)) {
			FAIL
		}

		/*
		 *	Update home page size 
		 */
    	oldSize = (int)objectPtr->objectHdr.size;
    	objectPtr->objectHdr.size = (FOUR)newSize;

        /* Log the update to the original header */
		if (GET_LOGLEVEL(LOG_SPACE)) {
			if(clog_LogChangeSmSize(oid,&(objectPtr->objectHdr),
				oldSize,newSize)) {
				FAIL
			}
		}
		homeLinkIsDirty = TRUE;
		goto Success;

	} else if (objectPtr->objectHdr.properties & P_FORWARD)	{
		SM_ERROR(TYPE_FATAL, esmINTERNAL);
	}

    /*
     * Dealing with a small object.
     * Delete the bytes using bcopy, then update object's header info.
     */

    smallObject		= (SMALLOBJ *) objectPtr;
	origAlignLength = ALIGNSIZE((int)smallObject->objectHdr.size);
	TRPRINT(TR_SM, TR_LEVEL_2, ("origAlignLength:%d", origAlignLength));
    deletePoint		= smallObject->data + start;
    source			= deletePoint + length;
    bytesToMove		= ((int)smallObject->objectHdr.size - start) - length;

	sm_SetAbortIfError();

	if (GET_LOGLEVEL(LOG_SPACE)) {

		opInfo.flags = 0;
		opInfo.oid = *oid;
		opInfo.start = start;
		opInfo.size = length;
		opInfo.newData = NULL;
		opInfo.actionPid.volid = oid->diskAddr.volid;
		opInfo.actionPid.page  = oid->diskAddr.page;
		opInfo.currentLRC = homePage->header.lrc;

		if (GET_LOGLEVEL(LOG_DATA)) {
			opInfo.action = LOG_ACTION_DELETE_SMALL_OBJ;
			opInfo.oldData = deletePoint;
			/* Log the delete operation */
			if(clog_LogModifyObj(&opInfo)) {
				FAIL
			}
		} else {
			opInfo.action = LOG_ACTION_DELETE_SMALL_OBJ_ND;
			opInfo.oldData = NULL;
			/* Log the delete operation */
			if(clog_LogModifyObjND(&opInfo)) {
				FAIL
			}
		}
	}

    bcopy(source, deletePoint, (int) bytesToMove);
    smallObject->objectHdr.size = newSize;

	/*
	 *	calculate the new aligned length
	 */
	newAlignLength = ALIGNSIZE(newSize);
	TRPRINT(TR_SM, TR_LEVEL_2, ("newAlignLength:%d", newAlignLength));

    homePage->header.freeBytes += (origAlignLength - newAlignLength);
	homeLinkIsDirty = TRUE;

Success:
	if(homePage != NULL) {
		sm_AuditSlotPage(homePage);
	}
	if(homeLink != NULL) {
		bf_UnfixPage(homeLink, BF_DEFAULT, TRUE);
	}
	if (suspendList != NULL)	{
		SM_RestoreDesc(suspendList);
		suspendList = NULL;
	}
	return esmNOERROR;

FailureReturn:
	if (homeLink != NULL) {
		bf_UnfixPage(homeLink, BF_DEFAULT, homeLinkIsDirty);
	}
	if (suspendList != NULL)	{
		SM_RestoreDesc(suspendList);
		suspendList = NULL;
	}
	SM_AbortIfRequired();
	return esmFAILURE;
}
